@@ -1,5 +1,5 @@
 /*
- * Copyright 2016 Andrey Gusev
+ * Copyright 2016 Matteo Bruni for CodeWeavers
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,8 +17,7 @@
  */
 
 #define COBJMACROS
-#include "d3d10_1.h"
-#include "d3dx10.h"
+#include "d3dx11.h"
 #include "d3dcompiler.h"
 #include "dxhelpers.h"
 #include "winternl.h"
@@ -30,7 +29,7 @@ WINE_DEFAULT_DEBUG_CHANNEL(d3dx);
 
 struct asyncdataloader
 {
-    ID3DX10DataLoader ID3DX10DataLoader_iface;
+    ID3DX11DataLoader ID3DX11DataLoader_iface;
 
     union
     {
@@ -48,20 +47,20 @@ struct asyncdataloader
     DWORD size;
 };
 
-static inline struct asyncdataloader *impl_from_ID3DX10DataLoader(ID3DX10DataLoader *iface)
+static inline struct asyncdataloader *impl_from_ID3DX11DataLoader(ID3DX11DataLoader *iface)
 {
-    return CONTAINING_RECORD(iface, struct asyncdataloader, ID3DX10DataLoader_iface);
+    return CONTAINING_RECORD(iface, struct asyncdataloader, ID3DX11DataLoader_iface);
 }
 
-static HRESULT WINAPI memorydataloader_Load(ID3DX10DataLoader *iface)
+static HRESULT WINAPI memorydataloader_Load(ID3DX11DataLoader *iface)
 {
     TRACE("iface %p.\n", iface);
     return S_OK;
 }
 
-static HRESULT WINAPI memorydataloader_Decompress(ID3DX10DataLoader *iface, void **data, SIZE_T *size)
+static HRESULT WINAPI memorydataloader_Decompress(ID3DX11DataLoader *iface, void **data, SIZE_T *size)
 {
-    struct asyncdataloader *loader = impl_from_ID3DX10DataLoader(iface);
+    struct asyncdataloader *loader = impl_from_ID3DX11DataLoader(iface);
 
     TRACE("iface %p, data %p, size %p.\n", iface, data, size);
 
@@ -71,9 +70,9 @@ static HRESULT WINAPI memorydataloader_Decompress(ID3DX10DataLoader *iface, void
     return S_OK;
 }
 
-static HRESULT WINAPI memorydataloader_Destroy(ID3DX10DataLoader *iface)
+static HRESULT WINAPI memorydataloader_Destroy(ID3DX11DataLoader *iface)
 {
-    struct asyncdataloader *loader = impl_from_ID3DX10DataLoader(iface);
+    struct asyncdataloader *loader = impl_from_ID3DX11DataLoader(iface);
 
     TRACE("iface %p.\n", iface);
 
@@ -81,24 +80,16 @@ static HRESULT WINAPI memorydataloader_Destroy(ID3DX10DataLoader *iface)
     return S_OK;
 }
 
-static const ID3DX10DataLoaderVtbl memorydataloadervtbl =
+static const ID3DX11DataLoaderVtbl memorydataloadervtbl =
 {
     memorydataloader_Load,
     memorydataloader_Decompress,
     memorydataloader_Destroy
 };
 
-HRESULT load_file(const WCHAR *path, void **data, DWORD *size)
+static HRESULT WINAPI filedataloader_Load(ID3DX11DataLoader *iface)
 {
-    HRESULT hr;
-
-    hr = d3dx_load_file(path, data, size);
-    return (hr == D3DX_HELPER_ERR_FILE_NOT_FOUND) ? D3D10_ERROR_FILE_NOT_FOUND : hr;
-}
-
-static HRESULT WINAPI filedataloader_Load(ID3DX10DataLoader *iface)
-{
-    struct asyncdataloader *loader = impl_from_ID3DX10DataLoader(iface);
+    struct asyncdataloader *loader = impl_from_ID3DX11DataLoader(iface);
     void *data;
     DWORD size;
     HRESULT hr;
@@ -116,9 +107,9 @@ static HRESULT WINAPI filedataloader_Load(ID3DX10DataLoader *iface)
     return S_OK;
 }
 
-static HRESULT WINAPI filedataloader_Decompress(ID3DX10DataLoader *iface, void **data, SIZE_T *size)
+static HRESULT WINAPI filedataloader_Decompress(ID3DX11DataLoader *iface, void **data, SIZE_T *size)
 {
-    struct asyncdataloader *loader = impl_from_ID3DX10DataLoader(iface);
+    struct asyncdataloader *loader = impl_from_ID3DX11DataLoader(iface);
 
     TRACE("iface %p, data %p, size %p.\n", iface, data, size);
 
@@ -131,9 +122,9 @@ static HRESULT WINAPI filedataloader_Decompress(ID3DX10DataLoader *iface, void *
     return S_OK;
 }
 
-static HRESULT WINAPI filedataloader_Destroy(ID3DX10DataLoader *iface)
+static HRESULT WINAPI filedataloader_Destroy(ID3DX11DataLoader *iface)
 {
-    struct asyncdataloader *loader = impl_from_ID3DX10DataLoader(iface);
+    struct asyncdataloader *loader = impl_from_ID3DX11DataLoader(iface);
 
     TRACE("iface %p.\n", iface);
 
@@ -144,54 +135,16 @@ static HRESULT WINAPI filedataloader_Destroy(ID3DX10DataLoader *iface)
     return S_OK;
 }
 
-static const ID3DX10DataLoaderVtbl filedataloadervtbl =
+static const ID3DX11DataLoaderVtbl filedataloadervtbl =
 {
     filedataloader_Load,
     filedataloader_Decompress,
     filedataloader_Destroy
 };
 
-static HRESULT load_resource_initA(HMODULE module, const char *resource, HRSRC *rsrc)
+static HRESULT WINAPI resourcedataloader_Load(ID3DX11DataLoader *iface)
 {
-    HRESULT hr = d3dx_load_resource_initA(module, resource, rsrc);
-    return (hr == D3DX_HELPER_ERR_INVALID_DATA) ? D3DX10_ERR_INVALID_DATA : hr;
-}
-
-static HRESULT load_resource_initW(HMODULE module, const WCHAR *resource, HRSRC *rsrc)
-{
-    HRESULT hr = d3dx_load_resource_initW(module, resource, rsrc);
-    return (hr == D3DX_HELPER_ERR_INVALID_DATA) ? D3DX10_ERR_INVALID_DATA : hr;
-}
-
-static HRESULT load_resource(HMODULE module, HRSRC rsrc, void **data, DWORD *size)
-{
-    HRESULT hr = d3dx_load_resource(module, rsrc, data, size);
-    return (hr == D3DX_HELPER_ERR_INVALID_DATA) ? D3DX10_ERR_INVALID_DATA : hr;
-}
-
-HRESULT load_resourceA(HMODULE module, const char *resource, void **data, DWORD *size)
-{
-    HRESULT hr;
-    HRSRC rsrc;
-
-    if (FAILED((hr = load_resource_initA(module, resource, &rsrc))))
-        return hr;
-    return load_resource(module, rsrc, data, size);
-}
-
-HRESULT load_resourceW(HMODULE module, const WCHAR *resource, void **data, DWORD *size)
-{
-    HRESULT hr;
-    HRSRC rsrc;
-
-    if ((FAILED(hr = load_resource_initW(module, resource, &rsrc))))
-        return hr;
-    return load_resource(module, rsrc, data, size);
-}
-
-static HRESULT WINAPI resourcedataloader_Load(ID3DX10DataLoader *iface)
-{
-    struct asyncdataloader *loader = impl_from_ID3DX10DataLoader(iface);
+    struct asyncdataloader *loader = impl_from_ID3DX11DataLoader(iface);
 
     TRACE("iface %p.\n", iface);
 
@@ -202,9 +155,9 @@ static HRESULT WINAPI resourcedataloader_Load(ID3DX10DataLoader *iface)
             &loader->data, &loader->size);
 }
 
-static HRESULT WINAPI resourcedataloader_Decompress(ID3DX10DataLoader *iface, void **data, SIZE_T *size)
+static HRESULT WINAPI resourcedataloader_Decompress(ID3DX11DataLoader *iface, void **data, SIZE_T *size)
 {
-    struct asyncdataloader *loader = impl_from_ID3DX10DataLoader(iface);
+    struct asyncdataloader *loader = impl_from_ID3DX11DataLoader(iface);
 
     TRACE("iface %p, data %p, size %p.\n", iface, data, size);
 
@@ -217,9 +170,9 @@ static HRESULT WINAPI resourcedataloader_Decompress(ID3DX10DataLoader *iface, vo
     return S_OK;
 }
 
-static HRESULT WINAPI resourcedataloader_Destroy(ID3DX10DataLoader *iface)
+static HRESULT WINAPI resourcedataloader_Destroy(ID3DX11DataLoader *iface)
 {
-    struct asyncdataloader *loader = impl_from_ID3DX10DataLoader(iface);
+    struct asyncdataloader *loader = impl_from_ID3DX11DataLoader(iface);
 
     TRACE("iface %p.\n", iface);
 
@@ -228,7 +181,7 @@ static HRESULT WINAPI resourcedataloader_Destroy(ID3DX10DataLoader *iface)
     return S_OK;
 }
 
-static const ID3DX10DataLoaderVtbl resourcedataloadervtbl =
+static const ID3DX11DataLoaderVtbl resourcedataloadervtbl =
 {
     resourcedataloader_Load,
     resourcedataloader_Decompress,
@@ -237,32 +190,32 @@ static const ID3DX10DataLoaderVtbl resourcedataloadervtbl =
 
 struct texture_info_processor
 {
-    ID3DX10DataProcessor ID3DX10DataProcessor_iface;
-    D3DX10_IMAGE_INFO *info;
+    ID3DX11DataProcessor ID3DX11DataProcessor_iface;
+    D3DX11_IMAGE_INFO *info;
 };
 
-static inline struct texture_info_processor *impl_from_ID3DX10DataProcessor(ID3DX10DataProcessor *iface)
+static inline struct texture_info_processor *impl_from_ID3DX11DataProcessor(ID3DX11DataProcessor *iface)
 {
-    return CONTAINING_RECORD(iface, struct texture_info_processor, ID3DX10DataProcessor_iface);
+    return CONTAINING_RECORD(iface, struct texture_info_processor, ID3DX11DataProcessor_iface);
 }
 
-static HRESULT WINAPI texture_info_processor_Process(ID3DX10DataProcessor *iface, void *data, SIZE_T size)
+static HRESULT WINAPI texture_info_processor_Process(ID3DX11DataProcessor *iface, void *data, SIZE_T size)
 {
-    struct texture_info_processor *processor = impl_from_ID3DX10DataProcessor(iface);
+    struct texture_info_processor *processor = impl_from_ID3DX11DataProcessor(iface);
 
     TRACE("iface %p, data %p, size %Iu.\n", iface, data, size);
     return get_image_info(data, size, processor->info);
 }
 
-static HRESULT WINAPI texture_info_processor_CreateDeviceObject(ID3DX10DataProcessor *iface, void **object)
+static HRESULT WINAPI texture_info_processor_CreateDeviceObject(ID3DX11DataProcessor *iface, void **object)
 {
     TRACE("iface %p, object %p.\n", iface, object);
     return S_OK;
 }
 
-static HRESULT WINAPI texture_info_processor_Destroy(ID3DX10DataProcessor *iface)
+static HRESULT WINAPI texture_info_processor_Destroy(ID3DX11DataProcessor *iface)
 {
-    struct texture_info_processor *processor = impl_from_ID3DX10DataProcessor(iface);
+    struct texture_info_processor *processor = impl_from_ID3DX11DataProcessor(iface);
 
     TRACE("iface %p.\n", iface);
 
@@ -270,7 +223,7 @@ static HRESULT WINAPI texture_info_processor_Destroy(ID3DX10DataProcessor *iface
     return S_OK;
 }
 
-static ID3DX10DataProcessorVtbl texture_info_processor_vtbl =
+static ID3DX11DataProcessorVtbl texture_info_processor_vtbl =
 {
     texture_info_processor_Process,
     texture_info_processor_CreateDeviceObject,
@@ -279,21 +232,21 @@ static ID3DX10DataProcessorVtbl texture_info_processor_vtbl =
 
 struct texture_processor
 {
-    ID3DX10DataProcessor ID3DX10DataProcessor_iface;
-    ID3D10Device *device;
-    D3DX10_IMAGE_INFO img_info;
-    D3DX10_IMAGE_LOAD_INFO load_info;
-    D3D10_SUBRESOURCE_DATA *resource_data;
+    ID3DX11DataProcessor ID3DX11DataProcessor_iface;
+    ID3D11Device *device;
+    D3DX11_IMAGE_INFO img_info;
+    D3DX11_IMAGE_LOAD_INFO load_info;
+    D3D11_SUBRESOURCE_DATA *resource_data;
 };
 
-static inline struct texture_processor *texture_processor_from_ID3DX10DataProcessor(ID3DX10DataProcessor *iface)
+static inline struct texture_processor *texture_processor_from_ID3DX11DataProcessor(ID3DX11DataProcessor *iface)
 {
-    return CONTAINING_RECORD(iface, struct texture_processor, ID3DX10DataProcessor_iface);
+    return CONTAINING_RECORD(iface, struct texture_processor, ID3DX11DataProcessor_iface);
 }
 
-static HRESULT WINAPI texture_processor_Process(ID3DX10DataProcessor *iface, void *data, SIZE_T size)
+static HRESULT WINAPI texture_processor_Process(ID3DX11DataProcessor *iface, void *data, SIZE_T size)
 {
-    struct texture_processor *processor = texture_processor_from_ID3DX10DataProcessor(iface);
+    struct texture_processor *processor = texture_processor_from_ID3DX11DataProcessor(iface);
 
     TRACE("iface %p, data %p, size %Iu.\n", iface, data, size);
 
@@ -306,9 +259,9 @@ static HRESULT WINAPI texture_processor_Process(ID3DX10DataProcessor *iface, voi
     return load_texture_data(data, size, &processor->load_info, &processor->resource_data);
 }
 
-static HRESULT WINAPI texture_processor_CreateDeviceObject(ID3DX10DataProcessor *iface, void **object)
+static HRESULT WINAPI texture_processor_CreateDeviceObject(ID3DX11DataProcessor *iface, void **object)
 {
-    struct texture_processor *processor = texture_processor_from_ID3DX10DataProcessor(iface);
+    struct texture_processor *processor = texture_processor_from_ID3DX11DataProcessor(iface);
 
     TRACE("iface %p, object %p.\n", iface, object);
 
@@ -316,22 +269,22 @@ static HRESULT WINAPI texture_processor_CreateDeviceObject(ID3DX10DataProcessor
         return E_FAIL;
 
     return create_d3d_texture(processor->device, &processor->load_info,
-            processor->resource_data, (ID3D10Resource **)object);
+            processor->resource_data, (ID3D11Resource **)object);
 }
 
-static HRESULT WINAPI texture_processor_Destroy(ID3DX10DataProcessor *iface)
+static HRESULT WINAPI texture_processor_Destroy(ID3DX11DataProcessor *iface)
 {
-    struct texture_processor *processor = texture_processor_from_ID3DX10DataProcessor(iface);
+    struct texture_processor *processor = texture_processor_from_ID3DX11DataProcessor(iface);
 
     TRACE("iface %p.\n", iface);
 
-    ID3D10Device_Release(processor->device);
+    ID3D11Device_Release(processor->device);
     free(processor->resource_data);
     free(processor);
     return S_OK;
 }
 
-static ID3DX10DataProcessorVtbl texture_processor_vtbl =
+static ID3DX11DataProcessorVtbl texture_processor_vtbl =
 {
     texture_processor_Process,
     texture_processor_CreateDeviceObject,
@@ -340,64 +293,64 @@ static ID3DX10DataProcessorVtbl texture_processor_vtbl =
 
 struct srv_processor
 {
-    ID3DX10DataProcessor ID3DX10DataProcessor_iface;
-    ID3DX10DataProcessor *texture_processor;
+    ID3DX11DataProcessor ID3DX11DataProcessor_iface;
+    ID3DX11DataProcessor *texture_processor;
 };
 
-static inline struct srv_processor *srv_processor_from_ID3DX10DataProcessor(ID3DX10DataProcessor *iface)
+static inline struct srv_processor *srv_processor_from_ID3DX11DataProcessor(ID3DX11DataProcessor *iface)
 {
-    return CONTAINING_RECORD(iface, struct srv_processor, ID3DX10DataProcessor_iface);
+    return CONTAINING_RECORD(iface, struct srv_processor, ID3DX11DataProcessor_iface);
 }
 
-static HRESULT WINAPI srv_processor_Process(ID3DX10DataProcessor *iface, void *data, SIZE_T size)
+static HRESULT WINAPI srv_processor_Process(ID3DX11DataProcessor *iface, void *data, SIZE_T size)
 {
-    struct srv_processor *processor = srv_processor_from_ID3DX10DataProcessor(iface);
+    struct srv_processor *processor = srv_processor_from_ID3DX11DataProcessor(iface);
 
     TRACE("iface %p, data %p, size %Iu.\n", iface, data, size);
 
-    return ID3DX10DataProcessor_Process(processor->texture_processor, data, size);
+    return ID3DX11DataProcessor_Process(processor->texture_processor, data, size);
 }
 
-static HRESULT WINAPI srv_processor_CreateDeviceObject(ID3DX10DataProcessor *iface, void **object)
+static HRESULT WINAPI srv_processor_CreateDeviceObject(ID3DX11DataProcessor *iface, void **object)
 {
-    struct srv_processor *processor = srv_processor_from_ID3DX10DataProcessor(iface);
-    struct texture_processor *tex_processor = texture_processor_from_ID3DX10DataProcessor(processor->texture_processor);
-    ID3D10Resource *texture_resource;
+    struct srv_processor *processor = srv_processor_from_ID3DX11DataProcessor(iface);
+    struct texture_processor *tex_processor = texture_processor_from_ID3DX11DataProcessor(processor->texture_processor);
+    ID3D11Resource *texture_resource;
     HRESULT hr;
 
     TRACE("iface %p, object %p.\n", iface, object);
 
-    hr = ID3DX10DataProcessor_CreateDeviceObject(processor->texture_processor, (void **)&texture_resource);
+    hr = ID3DX11DataProcessor_CreateDeviceObject(processor->texture_processor, (void **)&texture_resource);
     if (FAILED(hr))
         return hr;
 
-    hr = ID3D10Device_CreateShaderResourceView(tex_processor->device, texture_resource, NULL,
-            (ID3D10ShaderResourceView **)object);
-    ID3D10Resource_Release(texture_resource);
+    hr = ID3D11Device_CreateShaderResourceView(tex_processor->device, texture_resource, NULL,
+            (ID3D11ShaderResourceView **)object);
+    ID3D11Resource_Release(texture_resource);
     return hr;
 }
 
-static HRESULT WINAPI srv_processor_Destroy(ID3DX10DataProcessor *iface)
+static HRESULT WINAPI srv_processor_Destroy(ID3DX11DataProcessor *iface)
 {
-    struct srv_processor *processor = srv_processor_from_ID3DX10DataProcessor(iface);
+    struct srv_processor *processor = srv_processor_from_ID3DX11DataProcessor(iface);
 
     TRACE("iface %p.\n", iface);
 
-    ID3DX10DataProcessor_Destroy(processor->texture_processor);
+    ID3DX11DataProcessor_Destroy(processor->texture_processor);
     free(processor);
     return S_OK;
 }
 
-static ID3DX10DataProcessorVtbl srv_processor_vtbl =
+static ID3DX11DataProcessorVtbl srv_processor_vtbl =
 {
     srv_processor_Process,
     srv_processor_CreateDeviceObject,
     srv_processor_Destroy
 };
 
-HRESULT WINAPI D3DX10CompileFromMemory(const char *data, SIZE_T data_size, const char *filename,
+HRESULT WINAPI D3DX11CompileFromMemory(const char *data, SIZE_T data_size, const char *filename,
         const D3D10_SHADER_MACRO *defines, ID3D10Include *include, const char *entry_point,
-        const char *target, UINT sflags, UINT eflags, ID3DX10ThreadPump *pump, ID3D10Blob **shader,
+        const char *target, UINT sflags, UINT eflags, ID3DX11ThreadPump *pump, ID3D10Blob **shader,
         ID3D10Blob **error_messages, HRESULT *hresult)
 {
     TRACE("data %s, data_size %Iu, filename %s, defines %p, include %p, entry_point %s, target %s, "
@@ -407,37 +360,54 @@ HRESULT WINAPI D3DX10CompileFromMemory(const char *data, SIZE_T data_size, const
             error_messages, hresult);
 
     if (pump)
-        FIXME("Unimplemented ID3DX10ThreadPump handling.\n");
+        FIXME("Unimplemented ID3DX11ThreadPump handling.\n");
 
     return D3DCompile(data, data_size, filename, defines, include, entry_point, target,
             sflags, eflags, shader, error_messages);
 }
 
-HRESULT WINAPI D3DX10CreateEffectPoolFromFileA(const char *filename, const D3D10_SHADER_MACRO *defines,
-        ID3D10Include *include, const char *profile, UINT hlslflags, UINT fxflags, ID3D10Device *device,
-        ID3DX10ThreadPump *pump, ID3D10EffectPool **effectpool, ID3D10Blob **errors, HRESULT *hresult)
+HRESULT WINAPI D3DX11CompileFromFileA(const char *filename, const D3D10_SHADER_MACRO *defines,
+        ID3D10Include *include, const char *entry_point, const char *target, UINT sflags, UINT eflags,
+        ID3DX11ThreadPump *pump, ID3D10Blob **shader, ID3D10Blob **error_messages, HRESULT *hresult)
 {
-    FIXME("filename %s, defines %p, include %p, profile %s, hlslflags %#x, fxflags %#x, device %p, "
-            "pump %p, effectpool %p, errors %p, hresult %p, stub!\n",
-            debugstr_a(filename), defines, include, debugstr_a(profile), hlslflags, fxflags, device,
-            pump, effectpool, errors, hresult);
+    WCHAR filename_w[MAX_PATH];
 
-    return E_NOTIMPL;
+    TRACE("filename %s, defines %p, include %p, entry_point %s, target %s, sflags %#x, "
+            "eflags %#x, pump %p, shader %p, error_messages %p, hresult %p.\n",
+            debugstr_a(filename), defines, include, debugstr_a(entry_point), debugstr_a(target),
+            sflags, eflags, pump, shader, error_messages, hresult);
+
+    MultiByteToWideChar(CP_ACP, 0, filename, -1, filename_w, ARRAY_SIZE(filename_w));
+
+    return D3DX11CompileFromFileW(filename_w, defines, include, entry_point, target,
+            sflags, eflags, pump, shader, error_messages, hresult);
 }
 
-HRESULT WINAPI D3DX10CreateEffectPoolFromFileW(const WCHAR *filename, const D3D10_SHADER_MACRO *defines,
-        ID3D10Include *include, const char *profile, UINT hlslflags, UINT fxflags, ID3D10Device *device,
-        ID3DX10ThreadPump *pump, ID3D10EffectPool **effectpool, ID3D10Blob **errors, HRESULT *hresult)
+HRESULT WINAPI D3DX11CompileFromFileW(const WCHAR *filename, const D3D10_SHADER_MACRO *defines,
+        ID3D10Include *include, const char *entry_point, const char *target, UINT sflags, UINT eflags,
+        ID3DX11ThreadPump *pump, ID3D10Blob **shader, ID3D10Blob **error_messages, HRESULT *hresult)
 {
-    FIXME("filename %s, defines %p, include %p, profile %s, hlslflags %#x, fxflags %#x, device %p, "
-            "pump %p, effectpool %p, errors %p, hresult %p, stub!\n",
-            debugstr_w(filename), defines, include, debugstr_a(profile), hlslflags, fxflags, device,
-            pump, effectpool, errors, hresult);
+    HRESULT hr;
 
-    return E_NOTIMPL;
+    TRACE("filename %s, defines %p, include %p, entry_point %s, target %s, sflags %#x, "
+            "eflags %#x, pump %p, shader %p, error_messages %p, hresult %p.\n",
+            debugstr_w(filename), defines, include, debugstr_a(entry_point), debugstr_a(target),
+            sflags, eflags, pump, shader, error_messages, hresult);
+
+    if (pump)
+        FIXME("Unimplemented ID3DX11ThreadPump handling.\n");
+
+    if (!include)
+        include = D3D_COMPILE_STANDARD_FILE_INCLUDE;
+
+    hr = D3DCompileFromFile(filename, defines, include, entry_point, target, sflags, eflags, shader, error_messages);
+    if (hresult)
+        *hresult = hr;
+
+    return hr;
 }
 
-HRESULT WINAPI D3DX10CreateAsyncMemoryLoader(const void *data, SIZE_T data_size, ID3DX10DataLoader **loader)
+HRESULT WINAPI D3DX11CreateAsyncMemoryLoader(const void *data, SIZE_T data_size, ID3DX11DataLoader **loader)
 {
     struct asyncdataloader *object;
 
@@ -450,16 +420,16 @@ HRESULT WINAPI D3DX10CreateAsyncMemoryLoader(const void *data, SIZE_T data_size,
     if (!object)
         return E_OUTOFMEMORY;
 
-    object->ID3DX10DataLoader_iface.lpVtbl = &memorydataloadervtbl;
+    object->ID3DX11DataLoader_iface.lpVtbl = &memorydataloadervtbl;
     object->data = (void *)data;
     object->size = data_size;
 
-    *loader = &object->ID3DX10DataLoader_iface;
+    *loader = &object->ID3DX11DataLoader_iface;
 
     return S_OK;
 }
 
-HRESULT WINAPI D3DX10CreateAsyncFileLoaderA(const char *filename, ID3DX10DataLoader **loader)
+HRESULT WINAPI D3DX11CreateAsyncFileLoaderA(const char *filename, ID3DX11DataLoader **loader)
 {
     WCHAR *filename_w;
     HRESULT hr;
@@ -474,14 +444,14 @@ HRESULT WINAPI D3DX10CreateAsyncFileLoaderA(const char *filename, ID3DX10DataLoa
     filename_w = malloc(len * sizeof(*filename_w));
     MultiByteToWideChar(CP_ACP, 0, filename, -1, filename_w, len);
 
-    hr = D3DX10CreateAsyncFileLoaderW(filename_w, loader);
+    hr = D3DX11CreateAsyncFileLoaderW(filename_w, loader);
 
     free(filename_w);
 
     return hr;
 }
 
-HRESULT WINAPI D3DX10CreateAsyncFileLoaderW(const WCHAR *filename, ID3DX10DataLoader **loader)
+HRESULT WINAPI D3DX11CreateAsyncFileLoaderW(const WCHAR *filename, ID3DX11DataLoader **loader)
 {
     struct asyncdataloader *object;
 
@@ -494,7 +464,7 @@ HRESULT WINAPI D3DX10CreateAsyncFileLoaderW(const WCHAR *filename, ID3DX10DataLo
     if (!object)
         return E_OUTOFMEMORY;
 
-    object->ID3DX10DataLoader_iface.lpVtbl = &filedataloadervtbl;
+    object->ID3DX11DataLoader_iface.lpVtbl = &filedataloadervtbl;
     object->u.file.path = malloc((lstrlenW(filename) + 1) * sizeof(WCHAR));
     if (!object->u.file.path)
     {
@@ -505,12 +475,12 @@ HRESULT WINAPI D3DX10CreateAsyncFileLoaderW(const WCHAR *filename, ID3DX10DataLo
     object->data = NULL;
     object->size = 0;
 
-    *loader = &object->ID3DX10DataLoader_iface;
+    *loader = &object->ID3DX11DataLoader_iface;
 
     return S_OK;
 }
 
-HRESULT WINAPI D3DX10CreateAsyncResourceLoaderA(HMODULE module, const char *resource, ID3DX10DataLoader **loader)
+HRESULT WINAPI D3DX11CreateAsyncResourceLoaderA(HMODULE module, const char *resource, ID3DX11DataLoader **loader)
 {
     struct asyncdataloader *object;
     HRSRC rsrc;
@@ -531,18 +501,18 @@ HRESULT WINAPI D3DX10CreateAsyncResourceLoaderA(HMODULE module, const char *reso
         return hr;
     }
 
-    object->ID3DX10DataLoader_iface.lpVtbl = &resourcedataloadervtbl;
+    object->ID3DX11DataLoader_iface.lpVtbl = &resourcedataloadervtbl;
     object->u.resource.module = module;
     object->u.resource.rsrc = rsrc;
     object->data = NULL;
     object->size = 0;
 
-    *loader = &object->ID3DX10DataLoader_iface;
+    *loader = &object->ID3DX11DataLoader_iface;
 
     return S_OK;
 }
 
-HRESULT WINAPI D3DX10CreateAsyncResourceLoaderW(HMODULE module, const WCHAR *resource, ID3DX10DataLoader **loader)
+HRESULT WINAPI D3DX11CreateAsyncResourceLoaderW(HMODULE module, const WCHAR *resource, ID3DX11DataLoader **loader)
 {
     struct asyncdataloader *object;
     HRSRC rsrc;
@@ -563,18 +533,18 @@ HRESULT WINAPI D3DX10CreateAsyncResourceLoaderW(HMODULE module, const WCHAR *res
         return hr;
     }
 
-    object->ID3DX10DataLoader_iface.lpVtbl = &resourcedataloadervtbl;
+    object->ID3DX11DataLoader_iface.lpVtbl = &resourcedataloadervtbl;
     object->u.resource.module = module;
     object->u.resource.rsrc = rsrc;
     object->data = NULL;
     object->size = 0;
 
-    *loader = &object->ID3DX10DataLoader_iface;
+    *loader = &object->ID3DX11DataLoader_iface;
 
     return S_OK;
 }
 
-HRESULT WINAPI D3DX10CreateAsyncTextureInfoProcessor(D3DX10_IMAGE_INFO *info, ID3DX10DataProcessor **processor)
+HRESULT WINAPI D3DX11CreateAsyncTextureInfoProcessor(D3DX11_IMAGE_INFO *info, ID3DX11DataProcessor **processor)
 {
     struct texture_info_processor *object;
 
@@ -587,15 +557,15 @@ HRESULT WINAPI D3DX10CreateAsyncTextureInfoProcessor(D3DX10_IMAGE_INFO *info, ID
     if (!object)
         return E_OUTOFMEMORY;
 
-    object->ID3DX10DataProcessor_iface.lpVtbl = &texture_info_processor_vtbl;
+    object->ID3DX11DataProcessor_iface.lpVtbl = &texture_info_processor_vtbl;
     object->info = info;
 
-    *processor = &object->ID3DX10DataProcessor_iface;
+    *processor = &object->ID3DX11DataProcessor_iface;
     return S_OK;
 }
 
-HRESULT WINAPI D3DX10CreateAsyncTextureProcessor(ID3D10Device *device,
-        D3DX10_IMAGE_LOAD_INFO *load_info, ID3DX10DataProcessor **processor)
+HRESULT WINAPI D3DX11CreateAsyncTextureProcessor(ID3D11Device *device,
+        D3DX11_IMAGE_LOAD_INFO *load_info, ID3DX11DataProcessor **processor)
 {
     struct texture_processor *object;
 
@@ -608,19 +578,19 @@ HRESULT WINAPI D3DX10CreateAsyncTextureProcessor(ID3D10Device *device,
     if (!object)
         return E_OUTOFMEMORY;
 
-    object->ID3DX10DataProcessor_iface.lpVtbl = &texture_processor_vtbl;
+    object->ID3DX11DataProcessor_iface.lpVtbl = &texture_processor_vtbl;
     object->device = device;
-    ID3D10Device_AddRef(device);
+    ID3D11Device_AddRef(device);
     init_load_info(load_info, &object->load_info);
     if (!object->load_info.pSrcInfo)
         object->load_info.pSrcInfo = &object->img_info;
 
-    *processor = &object->ID3DX10DataProcessor_iface;
+    *processor = &object->ID3DX11DataProcessor_iface;
     return S_OK;
 }
 
-HRESULT WINAPI D3DX10CreateAsyncShaderResourceViewProcessor(ID3D10Device *device,
-        D3DX10_IMAGE_LOAD_INFO *load_info, ID3DX10DataProcessor **processor)
+HRESULT WINAPI D3DX11CreateAsyncShaderResourceViewProcessor(ID3D11Device *device,
+        D3DX11_IMAGE_LOAD_INFO *load_info, ID3DX11DataProcessor **processor)
 {
     struct srv_processor *object;
     HRESULT hr;
@@ -634,14 +604,14 @@ HRESULT WINAPI D3DX10CreateAsyncShaderResourceViewProcessor(ID3D10Device *device
     if (!object)
         return E_OUTOFMEMORY;
 
-    hr = D3DX10CreateAsyncTextureProcessor(device, load_info, &object->texture_processor);
+    hr = D3DX11CreateAsyncTextureProcessor(device, load_info, &object->texture_processor);
     if (FAILED(hr))
     {
         free(object);
         return hr;
     }
-    object->ID3DX10DataProcessor_iface.lpVtbl = &srv_processor_vtbl;
-    *processor = &object->ID3DX10DataProcessor_iface;
+    object->ID3DX11DataProcessor_iface.lpVtbl = &srv_processor_vtbl;
+    *processor = &object->ID3DX11DataProcessor_iface;
     return S_OK;
 }
 
@@ -649,16 +619,16 @@ struct work_item
 {
     struct list entry;
 
-    ID3DX10DataLoader *loader;
-    ID3DX10DataProcessor *processor;
+    ID3DX11DataLoader *loader;
+    ID3DX11DataProcessor *processor;
     HRESULT *result;
     void **object;
 };
 
 static inline void work_item_free(struct work_item *work_item, BOOL cancel)
 {
-    ID3DX10DataLoader_Destroy(work_item->loader);
-    ID3DX10DataProcessor_Destroy(work_item->processor);
+    ID3DX11DataLoader_Destroy(work_item->loader);
+    ID3DX11DataProcessor_Destroy(work_item->processor);
     if (cancel && work_item->result)
         *work_item->result = S_FALSE;
     free(work_item);
@@ -667,7 +637,7 @@ static inline void work_item_free(struct work_item *work_item, BOOL cancel)
 #define THREAD_PUMP_EXITING UINT_MAX
 struct thread_pump
 {
-    ID3DX10ThreadPump ID3DX10ThreadPump_iface;
+    ID3DX11ThreadPump ID3DX11ThreadPump_iface;
     LONG refcount;
 
     LONG processing_count;
@@ -690,19 +660,19 @@ struct thread_pump
     HANDLE threads[1];
 };
 
-static inline struct thread_pump *impl_from_ID3DX10ThreadPump(ID3DX10ThreadPump *iface)
+static inline struct thread_pump *impl_from_ID3DX11ThreadPump(ID3DX11ThreadPump *iface)
 {
-    return CONTAINING_RECORD(iface, struct thread_pump, ID3DX10ThreadPump_iface);
+    return CONTAINING_RECORD(iface, struct thread_pump, ID3DX11ThreadPump_iface);
 }
 
-static HRESULT WINAPI thread_pump_QueryInterface(ID3DX10ThreadPump *iface, REFIID riid, void **out)
+static HRESULT WINAPI thread_pump_QueryInterface(ID3DX11ThreadPump *iface, REFIID riid, void **out)
 {
     TRACE("iface %p, riid %s, out %p.\n", iface, debugstr_guid(riid), out);
 
-    if (IsEqualGUID(riid, &IID_ID3DX10ThreadPump)
+    if (IsEqualGUID(riid, &IID_ID3DX11ThreadPump)
             || IsEqualGUID(riid, &IID_IUnknown))
     {
-        ID3DX10ThreadPump_AddRef(iface);
+        ID3DX11ThreadPump_AddRef(iface);
         *out = iface;
         return S_OK;
     }
@@ -712,9 +682,9 @@ static HRESULT WINAPI thread_pump_QueryInterface(ID3DX10ThreadPump *iface, REFII
     return E_NOINTERFACE;
 }
 
-static ULONG WINAPI thread_pump_AddRef(ID3DX10ThreadPump *iface)
+static ULONG WINAPI thread_pump_AddRef(ID3DX11ThreadPump *iface)
 {
-    struct thread_pump *thread_pump = impl_from_ID3DX10ThreadPump(iface);
+    struct thread_pump *thread_pump = impl_from_ID3DX11ThreadPump(iface);
     ULONG refcount = InterlockedIncrement(&thread_pump->refcount);
 
     TRACE("%p increasing refcount to %lu.\n", iface, refcount);
@@ -722,9 +692,9 @@ static ULONG WINAPI thread_pump_AddRef(ID3DX10ThreadPump *iface)
     return refcount;
 }
 
-static ULONG WINAPI thread_pump_Release(ID3DX10ThreadPump *iface)
+static ULONG WINAPI thread_pump_Release(ID3DX11ThreadPump *iface)
 {
-    struct thread_pump *thread_pump = impl_from_ID3DX10ThreadPump(iface);
+    struct thread_pump *thread_pump = impl_from_ID3DX11ThreadPump(iface);
     ULONG refcount = InterlockedDecrement(&thread_pump->refcount);
     struct work_item *item, *next;
     struct list list;
@@ -773,10 +743,10 @@ static ULONG WINAPI thread_pump_Release(ID3DX10ThreadPump *iface)
     return refcount;
 }
 
-static HRESULT WINAPI thread_pump_AddWorkItem(ID3DX10ThreadPump *iface, ID3DX10DataLoader *loader,
-        ID3DX10DataProcessor *processor, HRESULT *result, void **object)
+static HRESULT WINAPI thread_pump_AddWorkItem(ID3DX11ThreadPump *iface, ID3DX11DataLoader *loader,
+        ID3DX11DataProcessor *processor, HRESULT *result, void **object)
 {
-    struct thread_pump *thread_pump = impl_from_ID3DX10ThreadPump(iface);
+    struct thread_pump *thread_pump = impl_from_ID3DX11ThreadPump(iface);
     struct work_item *work_item;
 
     TRACE("iface %p, loader %p, processor %p, result %p, object %p.\n",
@@ -803,9 +773,9 @@ static HRESULT WINAPI thread_pump_AddWorkItem(ID3DX10ThreadPump *iface, ID3DX10D
     return S_OK;
 }
 
-static UINT WINAPI thread_pump_GetWorkItemCount(ID3DX10ThreadPump *iface)
+static UINT WINAPI thread_pump_GetWorkItemCount(ID3DX11ThreadPump *iface)
 {
-    struct thread_pump *thread_pump = impl_from_ID3DX10ThreadPump(iface);
+    struct thread_pump *thread_pump = impl_from_ID3DX11ThreadPump(iface);
     UINT ret;
 
     TRACE("iface %p.\n", iface);
@@ -816,9 +786,9 @@ static UINT WINAPI thread_pump_GetWorkItemCount(ID3DX10ThreadPump *iface)
     return ret;
 }
 
-static HRESULT WINAPI thread_pump_WaitForAllItems(ID3DX10ThreadPump *iface)
+static HRESULT WINAPI thread_pump_WaitForAllItems(ID3DX11ThreadPump *iface)
 {
-    struct thread_pump *thread_pump = impl_from_ID3DX10ThreadPump(iface);
+    struct thread_pump *thread_pump = impl_from_ID3DX11ThreadPump(iface);
     HRESULT hr;
     LONG v;
 
@@ -826,7 +796,7 @@ static HRESULT WINAPI thread_pump_WaitForAllItems(ID3DX10ThreadPump *iface)
 
     for (;;)
     {
-        if (FAILED((hr = ID3DX10ThreadPump_ProcessDeviceWorkItems(iface, UINT_MAX))))
+        if (FAILED((hr = ID3DX11ThreadPump_ProcessDeviceWorkItems(iface, UINT_MAX))))
             return hr;
 
         AcquireSRWLockExclusive(&thread_pump->device_lock);
@@ -846,9 +816,9 @@ static HRESULT WINAPI thread_pump_WaitForAllItems(ID3DX10ThreadPump *iface)
     return S_OK;
 }
 
-static HRESULT WINAPI thread_pump_ProcessDeviceWorkItems(ID3DX10ThreadPump *iface, UINT count)
+static HRESULT WINAPI thread_pump_ProcessDeviceWorkItems(ID3DX11ThreadPump *iface, UINT count)
 {
-    struct thread_pump *thread_pump = impl_from_ID3DX10ThreadPump(iface);
+    struct thread_pump *thread_pump = impl_from_ID3DX11ThreadPump(iface);
     struct work_item *work_item;
     HRESULT hr;
     UINT i;
@@ -869,7 +839,7 @@ static HRESULT WINAPI thread_pump_ProcessDeviceWorkItems(ID3DX10ThreadPump *ifac
         list_remove(&work_item->entry);
         ReleaseSRWLockExclusive(&thread_pump->device_lock);
 
-        hr = ID3DX10DataProcessor_CreateDeviceObject(work_item->processor, work_item->object);
+        hr = ID3DX11DataProcessor_CreateDeviceObject(work_item->processor, work_item->object);
         if (work_item->result)
             *work_item->result = hr;
         work_item_free(work_item, FALSE);
@@ -893,9 +863,9 @@ static void purge_list(struct list *list, LONG *count)
     }
 }
 
-static HRESULT WINAPI thread_pump_PurgeAllItems(ID3DX10ThreadPump *iface)
+static HRESULT WINAPI thread_pump_PurgeAllItems(ID3DX11ThreadPump *iface)
 {
-    struct thread_pump *thread_pump = impl_from_ID3DX10ThreadPump(iface);
+    struct thread_pump *thread_pump = impl_from_ID3DX11ThreadPump(iface);
     LONG v;
 
     TRACE("iface %p.\n", iface);
@@ -926,10 +896,10 @@ static HRESULT WINAPI thread_pump_PurgeAllItems(ID3DX10ThreadPump *iface)
     return S_OK;
 }
 
-static HRESULT WINAPI thread_pump_GetQueueStatus(ID3DX10ThreadPump *iface,
+static HRESULT WINAPI thread_pump_GetQueueStatus(ID3DX11ThreadPump *iface,
         UINT *io_queue, UINT *process_queue, UINT *device_queue)
 {
-    struct thread_pump *thread_pump = impl_from_ID3DX10ThreadPump(iface);
+    struct thread_pump *thread_pump = impl_from_ID3DX11ThreadPump(iface);
 
     TRACE("iface %p, io_queue %p, process_queue %p, device_queue %p.\n",
             iface, io_queue, process_queue, device_queue);
@@ -940,7 +910,7 @@ static HRESULT WINAPI thread_pump_GetQueueStatus(ID3DX10ThreadPump *iface,
     return S_OK;
 }
 
-static const ID3DX10ThreadPumpVtbl thread_pump_vtbl =
+static const ID3DX11ThreadPumpVtbl thread_pump_vtbl =
 {
     thread_pump_QueryInterface,
     thread_pump_AddRef,
@@ -979,7 +949,7 @@ static DWORD WINAPI io_thread(void *arg)
         list_remove(&work_item->entry);
         ReleaseSRWLockExclusive(&thread_pump->io_lock);
 
-        if (FAILED(hr = ID3DX10DataLoader_Load(work_item->loader)))
+        if (FAILED(hr = ID3DX11DataLoader_Load(work_item->loader)))
         {
             if (work_item->result)
                 *work_item->result = hr;
@@ -1033,7 +1003,7 @@ static DWORD WINAPI proc_thread(void *arg)
         list_remove(&work_item->entry);
         ReleaseSRWLockExclusive(&thread_pump->proc_lock);
 
-        if (FAILED(hr = ID3DX10DataLoader_Decompress(work_item->loader, &data, &size)))
+        if (FAILED(hr = ID3DX11DataLoader_Decompress(work_item->loader, &data, &size)))
         {
             if (work_item->result)
                 *work_item->result = hr;
@@ -1049,7 +1019,7 @@ static DWORD WINAPI proc_thread(void *arg)
             return 0;
         }
 
-        if (FAILED(hr = ID3DX10DataProcessor_Process(work_item->processor, data, size)))
+        if (FAILED(hr = ID3DX11DataProcessor_Process(work_item->processor, data, size)))
         {
             if (work_item->result)
                 *work_item->result = hr;
@@ -1076,7 +1046,7 @@ static DWORD WINAPI proc_thread(void *arg)
     return 0;
 }
 
-HRESULT WINAPI D3DX10CreateThreadPump(UINT io_threads, UINT proc_threads, ID3DX10ThreadPump **pump)
+HRESULT WINAPI D3DX11CreateThreadPump(UINT io_threads, UINT proc_threads, ID3DX11ThreadPump **pump)
 {
     struct thread_pump *object;
     unsigned int i;
@@ -1099,7 +1069,7 @@ HRESULT WINAPI D3DX10CreateThreadPump(UINT io_threads, UINT proc_threads, ID3DX1
     if (!(object = calloc(1, FIELD_OFFSET(struct thread_pump, threads[io_threads + proc_threads]))))
         return E_OUTOFMEMORY;
 
-    object->ID3DX10ThreadPump_iface.lpVtbl = &thread_pump_vtbl;
+    object->ID3DX11ThreadPump_iface.lpVtbl = &thread_pump_vtbl;
     object->refcount = 1;
     InitializeSRWLock(&object->io_lock);
     InitializeConditionVariable(&object->io_cv);
@@ -1116,11 +1086,11 @@ HRESULT WINAPI D3DX10CreateThreadPump(UINT io_threads, UINT proc_threads, ID3DX1
         object->threads[i] = CreateThread(NULL, 0, i < io_threads ? io_thread : proc_thread, object, 0, NULL);
         if (!object->threads[i])
         {
-            ID3DX10ThreadPump_Release(&object->ID3DX10ThreadPump_iface);
+            ID3DX11ThreadPump_Release(&object->ID3DX11ThreadPump_iface);
             return E_FAIL;
         }
     }
 
-    *pump = &object->ID3DX10ThreadPump_iface;
+    *pump = &object->ID3DX11ThreadPump_iface;
     return S_OK;
 }
