@@ -61,15 +61,19 @@ static BOOL fetch_process_info(struct dump_context* dc)
         {
             if (HandleToUlong(spi->UniqueProcessId) == dc->pid)
             {
-                dc->num_threads = spi->dwThreadCount;
+                dc->num_threads = 0;
                 dc->threads = HeapAlloc(GetProcessHeap(), 0,
-                                        dc->num_threads * sizeof(dc->threads[0]));
-                if (!dc->threads) goto failed;
-                for (i = 0; i < dc->num_threads; i++)
+                                        spi->dwThreadCount * sizeof(dc->threads[0]));
+                if (!dc->threads) break;
+                for (i = 0; i < spi->dwThreadCount; i++)
                 {
-                    dc->threads[i].tid        = HandleToULong(spi->ti[i].ClientId.UniqueThread);
-                    dc->threads[i].prio_class = spi->ti[i].dwBasePriority; /* FIXME */
-                    dc->threads[i].curr_prio  = spi->ti[i].dwCurrentPriority;
+                    /* don't include current thread */
+                    if (HandleToULong(spi->ti[i].ClientId.UniqueThread) == GetCurrentThreadId())
+                        continue;
+                    dc->threads[dc->num_threads].tid        = HandleToULong(spi->ti[i].ClientId.UniqueThread);
+                    dc->threads[dc->num_threads].prio_class = spi->ti[i].dwBasePriority; /* FIXME */
+                    dc->threads[dc->num_threads].curr_prio  = spi->ti[i].dwCurrentPriority;
+                    dc->num_threads++;
                 }
                 HeapFree(GetProcessHeap(), 0, pcs_buffer);
                 return TRUE;
@@ -78,7 +82,6 @@ static BOOL fetch_process_info(struct dump_context* dc)
             spi = (SYSTEM_PROCESS_INFORMATION*)((char*)spi + spi->NextEntryOffset);
         }
     }
-failed:
     HeapFree(GetProcessHeap(), 0, pcs_buffer);
     return FALSE;
 }
@@ -116,7 +119,6 @@ static void fetch_thread_stack(struct dump_context* dc, const void* teb_addr,
  * fetches some information about thread of id 'tid'
  */
 static BOOL fetch_thread_info(struct dump_context* dc, int thd_idx,
-                              const MINIDUMP_EXCEPTION_INFORMATION* except,
                               MINIDUMP_THREAD* mdThd, CONTEXT* ctx)
 {
     DWORD                       tid = dc->threads[thd_idx].tid;
@@ -141,44 +143,22 @@ static BOOL fetch_thread_info(struct dump_context* dc, int thd_idx,
         FIXME("Couldn't open thread %lu (%lu)\n", tid, GetLastError());
         return FALSE;
     }
-    
+
     if (NtQueryInformationThread(hThread, ThreadBasicInformation,
                                  &tbi, sizeof(tbi), NULL) == STATUS_SUCCESS)
     {
         mdThd->Teb = (ULONG_PTR)tbi.TebBaseAddress;
         if (tbi.ExitStatus == STILL_ACTIVE)
         {
-            if (tid != GetCurrentThreadId() &&
-                (mdThd->SuspendCount = SuspendThread(hThread)) != (DWORD)-1)
-            {
-                ctx->ContextFlags = CONTEXT_FULL;
-                if (!GetThreadContext(hThread, ctx))
-                    memset(ctx, 0, sizeof(*ctx));
-
-                fetch_thread_stack(dc, tbi.TebBaseAddress, ctx, &mdThd->Stack);
-                ResumeThread(hThread);
-            }
-            else if (tid == GetCurrentThreadId() && except)
-            {
-                CONTEXT lctx, *pctx;
-                mdThd->SuspendCount = 1;
-                if (except->ClientPointers)
-                {
-                    EXCEPTION_POINTERS      ep;
-
-                    ReadProcessMemory(dc->process->handle, except->ExceptionPointers,
-                                      &ep, sizeof(ep), NULL);
-                    ReadProcessMemory(dc->process->handle, ep.ContextRecord,
-                                      &lctx, sizeof(lctx), NULL);
-                    pctx = &lctx;
-                }
-                else pctx = except->ExceptionPointers->ContextRecord;
-
-                *ctx = *pctx;
-                fetch_thread_stack(dc, tbi.TebBaseAddress, pctx, &mdThd->Stack);
-            }
-            else mdThd->SuspendCount = 0;
+            mdThd->SuspendCount = SuspendThread(hThread);
+            ctx->ContextFlags = CONTEXT_ALL;
+            if (!GetThreadContext(hThread, ctx))
+                memset(ctx, 0, sizeof(*ctx));
+            fetch_thread_stack(dc, tbi.TebBaseAddress, ctx, &mdThd->Stack);
+            ResumeThread(hThread);
         }
+        else
+            mdThd->SuspendCount = (DWORD)-1;
     }
     CloseHandle(hThread);
     return TRUE;
@@ -397,22 +377,21 @@ static void append(struct dump_context* dc, const void* data, unsigned size)
  *
  * Write in File the exception information from pcs
  */
-static  unsigned        dump_exception_info(struct dump_context* dc,
-                                            const MINIDUMP_EXCEPTION_INFORMATION* except)
+static  unsigned        dump_exception_info(struct dump_context* dc)
 {
     MINIDUMP_EXCEPTION_STREAM   mdExcpt;
     EXCEPTION_RECORD            rec, *prec;
     CONTEXT                     ctx, *pctx;
     DWORD                       i;
 
-    mdExcpt.ThreadId = except->ThreadId;
+    mdExcpt.ThreadId = dc->except_param->ThreadId;
     mdExcpt.__alignment = 0;
-    if (except->ClientPointers)
+    if (dc->except_param->ClientPointers)
     {
         EXCEPTION_POINTERS      ep;
 
         ReadProcessMemory(dc->process->handle,
-                          except->ExceptionPointers, &ep, sizeof(ep), NULL);
+                          dc->except_param->ExceptionPointers, &ep, sizeof(ep), NULL);
         ReadProcessMemory(dc->process->handle,
                           ep.ExceptionRecord, &rec, sizeof(rec), NULL);
         ReadProcessMemory(dc->process->handle,
@@ -422,8 +401,8 @@ static  unsigned        dump_exception_info(struct dump_context* dc,
     }
     else
     {
-        prec = except->ExceptionPointers->ExceptionRecord;
-        pctx = except->ExceptionPointers->ContextRecord;
+        prec = dc->except_param->ExceptionPointers->ExceptionRecord;
+        pctx = dc->except_param->ExceptionPointers->ContextRecord;
     }
     mdExcpt.ExceptionRecord.ExceptionCode = prec->ExceptionCode;
     mdExcpt.ExceptionRecord.ExceptionFlags = prec->ExceptionFlags;
@@ -715,8 +694,7 @@ static  unsigned        dump_system_info(struct dump_context* dc)
  *
  * Dumps into File the information about running threads
  */
-static  unsigned        dump_threads(struct dump_context* dc,
-                                     const MINIDUMP_EXCEPTION_INFORMATION* except)
+static  unsigned        dump_threads(struct dump_context* dc)
 {
     MINIDUMP_THREAD             mdThd;
     MINIDUMP_THREAD_LIST        mdThdList;
@@ -732,7 +710,7 @@ static  unsigned        dump_threads(struct dump_context* dc,
 
     for (i = 0; i < dc->num_threads; i++)
     {
-        fetch_thread_info(dc, i, except, &mdThd, &ctx);
+        fetch_thread_info(dc, i, &mdThd, &ctx);
 
         flags_out = ThreadWriteThread | ThreadWriteStack | ThreadWriteContext |
             ThreadWriteInstructionWindow;
@@ -793,6 +771,60 @@ static  unsigned        dump_threads(struct dump_context* dc,
     return sz;
 }
 
+/******************************************************************
+ *		dump_threads_names
+ *
+ * Dumps into File the information about threads's name
+ */
+static  unsigned        dump_threads_names(struct dump_context* dc)
+{
+    MINIDUMP_THREAD_NAME        md_thread_name;
+    MINIDUMP_THREAD_NAME_LIST   md_thread_name_list;
+    unsigned                    i, sz;
+    RVA                         rva_base;
+
+    /* FIXME this could be optimized
+     * (we use dc->num_threads disk space, could be optimized to the number of threads with name)
+     */
+
+    rva_base = dc->rva;
+    dc->rva += sz = offsetof(MINIDUMP_THREAD_NAME_LIST, ThreadNames[dc->num_threads]);
+
+    md_thread_name_list.NumberOfThreadNames = 0;
+
+    for (i = 0; i < dc->num_threads; i++)
+    {
+        HANDLE  thread;
+        WCHAR  *thread_name;
+
+        if ((thread = OpenThread(THREAD_ALL_ACCESS, FALSE, dc->threads[i].tid)) != NULL)
+        {
+            if (GetThreadDescription(thread, &thread_name))
+            {
+                MINIDUMP_STRING md_string;
+
+                md_thread_name.ThreadId = dc->threads[i].tid;
+                md_thread_name.RvaOfThreadName = dc->rva;
+
+                md_string.Length = wcslen(thread_name) * sizeof(WCHAR);
+                append(dc, &md_string.Length, sizeof(md_string.Length));
+                append(dc, thread_name, md_string.Length);
+
+                writeat(dc,
+                        rva_base + offsetof(MINIDUMP_THREAD_NAME_LIST, ThreadNames[md_thread_name_list.NumberOfThreadNames]),
+                        &md_thread_name, sizeof(md_thread_name));
+                md_thread_name_list.NumberOfThreadNames++;
+                LocalFree(thread_name);
+            }
+            CloseHandle(thread);
+        }
+    }
+    if (!md_thread_name_list.NumberOfThreadNames) return 0;
+
+    writeat(dc, rva_base, &md_thread_name_list.NumberOfThreadNames, sizeof(md_thread_name_list.NumberOfThreadNames));
+    return sz;
+}
+
 /******************************************************************
  *		dump_memory_info
  *
@@ -906,6 +938,132 @@ static unsigned         dump_misc_info(struct dump_context* dc)
     return sizeof(mmi);
 }
 
+static DWORD CALLBACK write_minidump(void *_args)
+{
+    struct dump_context *dc = _args;
+    static const MINIDUMP_DIRECTORY emptyDir = {UnusedStream, {0, 0}};
+    MINIDUMP_HEADER     mdHead;
+    MINIDUMP_DIRECTORY  mdDir;
+    DWORD               i, nStreams, idx_stream;
+
+    if (!fetch_process_info(dc)) return FALSE;
+    fetch_modules_info(dc);
+
+    /* 1) init */
+    nStreams = 7 + (dc->except_param ? 1 : 0) +
+        (dc->user_stream ? dc->user_stream->UserStreamCount : 0);
+
+    /* pad the directory size to a multiple of 4 for alignment purposes */
+    nStreams = (nStreams + 3) & ~3;
+
+    /* 2) write header */
+    mdHead.Signature = MINIDUMP_SIGNATURE;
+    mdHead.Version = MINIDUMP_VERSION;  /* NOTE: native puts in an 'implementation specific' value in the high order word of this member */
+    mdHead.NumberOfStreams = nStreams;
+    mdHead.CheckSum = 0;                /* native sets a 0 checksum in its files */
+    mdHead.StreamDirectoryRva = sizeof(mdHead);
+    mdHead.TimeDateStamp = time(NULL);
+    mdHead.Flags = dc->type;
+    append(dc, &mdHead, sizeof(mdHead));
+
+    /* 3) write stream directories */
+    dc->rva += nStreams * sizeof(mdDir);
+    idx_stream = 0;
+
+    /* 3.1) write data stream directories */
+
+    /* must be first in minidump */
+    mdDir.StreamType = SystemInfoStream;
+    mdDir.Location.Rva = dc->rva;
+    mdDir.Location.DataSize = dump_system_info(dc);
+    writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+            &mdDir, sizeof(mdDir));
+
+    mdDir.StreamType = ThreadListStream;
+    mdDir.Location.Rva = dc->rva;
+    mdDir.Location.DataSize = dump_threads(dc);
+    writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+            &mdDir, sizeof(mdDir));
+
+    mdDir.StreamType = ThreadNamesStream;
+    mdDir.Location.Rva = dc->rva;
+    if ((mdDir.Location.DataSize = dump_threads_names(dc)))
+        writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+                &mdDir, sizeof(mdDir));
+
+    mdDir.StreamType = ModuleListStream;
+    mdDir.Location.Rva = dc->rva;
+    mdDir.Location.DataSize = dump_modules(dc, FALSE);
+    writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+            &mdDir, sizeof(mdDir));
+
+    mdDir.StreamType = 0xfff0; /* FIXME: this is part of MS reserved streams */
+    mdDir.Location.Rva = dc->rva;
+    mdDir.Location.DataSize = dump_modules(dc, TRUE);
+    writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+            &mdDir, sizeof(mdDir));
+
+
+    if (!(dc->type & MiniDumpWithFullMemory))
+    {
+        mdDir.StreamType = MemoryListStream;
+        mdDir.Location.Rva = dc->rva;
+        mdDir.Location.DataSize = dump_memory_info(dc);
+        writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+                &mdDir, sizeof(mdDir));
+    }
+
+    mdDir.StreamType = MiscInfoStream;
+    mdDir.Location.Rva = dc->rva;
+    mdDir.Location.DataSize = dump_misc_info(dc);
+    writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+            &mdDir, sizeof(mdDir));
+
+    /* 3.2) write exception information (if any) */
+    if (dc->except_param)
+    {
+        mdDir.StreamType = ExceptionStream;
+        mdDir.Location.Rva = dc->rva;
+        mdDir.Location.DataSize = dump_exception_info(dc);
+        writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+                &mdDir, sizeof(mdDir));
+    }
+
+    /* 3.3) write user defined streams (if any) */
+    if (dc->user_stream)
+    {
+        for (i = 0; i < dc->user_stream->UserStreamCount; i++)
+        {
+            mdDir.StreamType = dc->user_stream->UserStreamArray[i].Type;
+            mdDir.Location.DataSize = dc->user_stream->UserStreamArray[i].BufferSize;
+            mdDir.Location.Rva = dc->rva;
+            writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+                    &mdDir, sizeof(mdDir));
+            append(dc, dc->user_stream->UserStreamArray[i].Buffer,
+                   dc->user_stream->UserStreamArray[i].BufferSize);
+        }
+    }
+
+    /* 3.4) write full memory (if requested) */
+    if (dc->type & MiniDumpWithFullMemory)
+    {
+        fetch_memory64_info(dc);
+
+        mdDir.StreamType = Memory64ListStream;
+        mdDir.Location.Rva = dc->rva;
+        mdDir.Location.DataSize = dump_memory64_info(dc);
+        writeat(dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
+                &mdDir, sizeof(mdDir));
+    }
+
+    /* fill the remaining directory entries with 0's (unused stream types) */
+    /* NOTE: this should always come last in the dump! */
+    for (i = idx_stream; i < nStreams; i++)
+        writeat(dc, mdHead.StreamDirectoryRva + i * sizeof(emptyDir), &emptyDir, sizeof(emptyDir));
+
+    return TRUE;
+}
+
 /******************************************************************
  *		MiniDumpWriteDump (DEBUGHLP.@)
  *
@@ -916,12 +1074,12 @@ BOOL WINAPI MiniDumpWriteDump(HANDLE hProcess, DWORD pid, HANDLE hFile,
                               PMINIDUMP_USER_STREAM_INFORMATION UserStreamParam,
                               PMINIDUMP_CALLBACK_INFORMATION CallbackParam)
 {
-    static const MINIDUMP_DIRECTORY emptyDir = {UnusedStream, {0, 0}};
-    MINIDUMP_HEADER     mdHead;
-    MINIDUMP_DIRECTORY  mdDir;
-    DWORD               i, nStreams, idx_stream;
     struct dump_context dc;
     BOOL                sym_initialized = FALSE;
+    BOOL                ret = FALSE;
+
+    TRACE("(%p, %lu, %p, %u, %p, %p, %p)\n",
+          hProcess, pid, hFile, DumpType, ExceptionParam, UserStreamParam, CallbackParam);
 
     if (!(dc.process = process_find_by_handle(hProcess)))
     {
@@ -933,6 +1091,15 @@ BOOL WINAPI MiniDumpWriteDump(HANDLE hProcess, DWORD pid, HANDLE hFile,
         dc.process = process_find_by_handle(hProcess);
     }
 
+    if (DumpType & MiniDumpWithDataSegs)
+        FIXME("NIY MiniDumpWithDataSegs\n");
+    if (DumpType & MiniDumpWithHandleData)
+        FIXME("NIY MiniDumpWithHandleData\n");
+    if (DumpType & MiniDumpFilterMemory)
+        FIXME("NIY MiniDumpFilterMemory\n");
+    if (DumpType & MiniDumpScanMemory)
+        FIXME("NIY MiniDumpScanMemory\n");
+
     dc.hFile = hFile;
     dc.pid = pid;
     dc.modules = NULL;
@@ -949,125 +1116,26 @@ BOOL WINAPI MiniDumpWriteDump(HANDLE hProcess, DWORD pid, HANDLE hFile,
     dc.num_mem64 = 0;
     dc.alloc_mem64 = 0;
     dc.rva = 0;
+    dc.except_param = ExceptionParam;
+    dc.user_stream = UserStreamParam;
 
-    if (!fetch_process_info(&dc)) return FALSE;
-    fetch_modules_info(&dc);
-
-    /* 1) init */
-    nStreams = 6 + (ExceptionParam ? 1 : 0) +
-        (UserStreamParam ? UserStreamParam->UserStreamCount : 0);
-
-    /* pad the directory size to a multiple of 4 for alignment purposes */
-    nStreams = (nStreams + 3) & ~3;
-
-    if (DumpType & MiniDumpWithDataSegs)
-        FIXME("NIY MiniDumpWithDataSegs\n");
-    if (DumpType & MiniDumpWithHandleData)
-        FIXME("NIY MiniDumpWithHandleData\n");
-    if (DumpType & MiniDumpFilterMemory)
-        FIXME("NIY MiniDumpFilterMemory\n");
-    if (DumpType & MiniDumpScanMemory)
-        FIXME("NIY MiniDumpScanMemory\n");
-
-    /* 2) write header */
-    mdHead.Signature = MINIDUMP_SIGNATURE;
-    mdHead.Version = MINIDUMP_VERSION;  /* NOTE: native puts in an 'implementation specific' value in the high order word of this member */
-    mdHead.NumberOfStreams = nStreams;
-    mdHead.CheckSum = 0;                /* native sets a 0 checksum in its files */
-    mdHead.StreamDirectoryRva = sizeof(mdHead);
-    mdHead.TimeDateStamp = time(NULL);
-    mdHead.Flags = DumpType;
-    append(&dc, &mdHead, sizeof(mdHead));
-
-    /* 3) write stream directories */
-    dc.rva += nStreams * sizeof(mdDir);
-    idx_stream = 0;
-
-    /* 3.1) write data stream directories */
-
-    /* must be first in minidump */
-    mdDir.StreamType = SystemInfoStream;
-    mdDir.Location.Rva = dc.rva;
-    mdDir.Location.DataSize = dump_system_info(&dc);
-    writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
-            &mdDir, sizeof(mdDir));
-
-    mdDir.StreamType = ThreadListStream;
-    mdDir.Location.Rva = dc.rva;
-    mdDir.Location.DataSize = dump_threads(&dc, ExceptionParam);
-    writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir), 
-            &mdDir, sizeof(mdDir));
-
-    mdDir.StreamType = ModuleListStream;
-    mdDir.Location.Rva = dc.rva;
-    mdDir.Location.DataSize = dump_modules(&dc, FALSE);
-    writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
-            &mdDir, sizeof(mdDir));
-
-    mdDir.StreamType = 0xfff0; /* FIXME: this is part of MS reserved streams */
-    mdDir.Location.Rva = dc.rva;
-    mdDir.Location.DataSize = dump_modules(&dc, TRUE);
-    writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
-            &mdDir, sizeof(mdDir));
-
-
-    if (!(DumpType & MiniDumpWithFullMemory))
+    /* have a dedicated thread for fetching info on self */
+    if (dc.pid != GetCurrentProcessId())
+        ret = write_minidump(&dc);
+    else
     {
-        mdDir.StreamType = MemoryListStream;
-        mdDir.Location.Rva = dc.rva;
-        mdDir.Location.DataSize = dump_memory_info(&dc);
-        writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
-                &mdDir, sizeof(mdDir));
-    }
-
-    mdDir.StreamType = MiscInfoStream;
-    mdDir.Location.Rva = dc.rva;
-    mdDir.Location.DataSize = dump_misc_info(&dc);
-    writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
-            &mdDir, sizeof(mdDir));
-
-    /* 3.2) write exception information (if any) */
-    if (ExceptionParam)
-    {
-        mdDir.StreamType = ExceptionStream;
-        mdDir.Location.Rva = dc.rva;
-        mdDir.Location.DataSize = dump_exception_info(&dc, ExceptionParam);
-        writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
-                &mdDir, sizeof(mdDir));
-    }
-
-    /* 3.3) write user defined streams (if any) */
-    if (UserStreamParam)
-    {
-        for (i = 0; i < UserStreamParam->UserStreamCount; i++)
+        DWORD  exit_code;
+        HANDLE h = CreateThread(NULL, 0, write_minidump, &dc, 0, NULL);
+        if (h)
         {
-            mdDir.StreamType = UserStreamParam->UserStreamArray[i].Type;
-            mdDir.Location.DataSize = UserStreamParam->UserStreamArray[i].BufferSize;
-            mdDir.Location.Rva = dc.rva;
-            writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
-                    &mdDir, sizeof(mdDir));
-            append(&dc, UserStreamParam->UserStreamArray[i].Buffer, 
-                   UserStreamParam->UserStreamArray[i].BufferSize);
+            if (WaitForSingleObject(h, INFINITE) == WAIT_OBJECT_0 && GetExitCodeThread(h, &exit_code))
+                ret = exit_code;
+            else
+                TerminateThread(h, 0);
+            CloseHandle(h);
         }
     }
 
-    /* 3.4) write full memory (if requested) */
-    if (DumpType & MiniDumpWithFullMemory)
-    {
-        fetch_memory64_info(&dc);
-
-        mdDir.StreamType = Memory64ListStream;
-        mdDir.Location.Rva = dc.rva;
-        mdDir.Location.DataSize = dump_memory64_info(&dc);
-        writeat(&dc, mdHead.StreamDirectoryRva + idx_stream++ * sizeof(mdDir),
-                &mdDir, sizeof(mdDir));
-    }
-
-    /* fill the remaining directory entries with 0's (unused stream types) */
-    /* NOTE: this should always come last in the dump! */
-    for (i = idx_stream; i < nStreams; i++)
-        writeat(&dc, mdHead.StreamDirectoryRva + i * sizeof(emptyDir), &emptyDir, sizeof(emptyDir));
-
     if (sym_initialized)
         SymCleanup(hProcess);
 
@@ -1076,7 +1144,7 @@ BOOL WINAPI MiniDumpWriteDump(HANDLE hProcess, DWORD pid, HANDLE hFile,
     HeapFree(GetProcessHeap(), 0, dc.modules);
     HeapFree(GetProcessHeap(), 0, dc.threads);
 
-    return TRUE;
+    return ret;
 }
 
 /******************************************************************
