@@ -24,6 +24,8 @@
 #include <string.h>
 #include <assert.h>
 
+#include "ntstatus.h"
+#define WIN32_NO_STATUS
 #include "dbghelp_private.h"
 #include "image_private.h"
 #include "psapi.h"
@@ -210,6 +212,7 @@ struct module* module_new(struct process* pcs, const WCHAR* name,
     module->module.BaseOfImage = mod_addr;
     module->module.ImageSize = size;
     module_set_module(module, name);
+    module->alt_modulename = NULL;
     module->module.ImageName[0] = '\0';
     lstrcpynW(module->module.LoadedImageName, name, ARRAY_SIZE(module->module.LoadedImageName));
     module->module.SymType = SymDeferred;
@@ -250,14 +253,13 @@ struct module* module_new(struct process* pcs, const WCHAR* name,
         module->cpu = dbghelp_current_cpu;
     module->debug_format_bitmask = 0;
 
-    vector_init(&module->vsymt, sizeof(struct symt*), 128);
-    vector_init(&module->vcustom_symt, sizeof(struct symt*), 16);
+    vector_init(&module->vsymt, sizeof(symref_t), 0);
+    vector_init(&module->vcustom_symt, sizeof(symref_t), 0);
     /* FIXME: this seems a bit too high (on a per module basis)
      * need some statistics about this
      */
     hash_table_init(&module->pool, &module->ht_symbols, 4096);
     hash_table_init(&module->pool, &module->ht_types,   4096);
-    vector_init(&module->vtypes, sizeof(struct symt*),  32);
 
     module->sources_used      = 0;
     module->sources_alloc     = 0;
@@ -288,6 +290,7 @@ struct module* module_find_by_nameW(const struct process* pcs, const WCHAR* name
     for (module = pcs->lmodules; module; module = module->next)
     {
         if (!wcsicmp(name, module->modulename)) return module;
+        if (module->alt_modulename && !wcsicmp(name, module->alt_modulename)) return module;
     }
     SetLastError(ERROR_INVALID_NAME);
     return NULL;
@@ -374,8 +377,12 @@ BOOL module_load_debug(struct module* module)
     if (module->module.SymType == SymDeferred)
     {
         BOOL ret;
-        
-        if (module->is_virtual) ret = FALSE;
+
+        if (module->is_virtual)
+        {
+            module->module.SymType = SymVirtual;
+            ret = TRUE;
+        }
         else if (module->type == DMT_PE)
         {
             idslW64.SizeOfStruct = sizeof(idslW64);
@@ -398,8 +405,9 @@ BOOL module_load_debug(struct module* module)
         if (!ret) module->module.SymType = SymNone;
         assert(module->module.SymType != SymDeferred);
         module->module.NumSyms = module->ht_symbols.num_elts;
+        return ret;
     }
-    return module->module.SymType != SymNone;
+    return TRUE;
 }
 
 /******************************************************************
@@ -410,7 +418,6 @@ BOOL module_load_debug(struct module* module)
  *   the module itself)
  * - if the module has no debug info and has an ELF container, then return the ELF
  *   container (and also force the ELF container's debug info loading if deferred)
- * - otherwise return the module itself if it has some debug info
  */
 BOOL module_get_debug(struct module_pair* pair)
 {
@@ -519,6 +526,7 @@ static BOOL image_check_debug_link_gnu_id(const WCHAR* file, struct image_file_m
 {
     struct image_section_map buildid_sect;
     DWORD read_bytes;
+    GUID guid;
     HANDLE handle;
     WCHAR *path;
     WORD magic;
@@ -537,6 +545,9 @@ static BOOL image_check_debug_link_gnu_id(const WCHAR* file, struct image_file_m
         ret = elf_map_handle(handle, fmap);
     CloseHandle(handle);
 
+    if (ret && pe_has_buildid_debug(fmap, &guid))
+        return TRUE;
+
     if (ret && image_find_section(fmap, ".note.gnu.build-id", &buildid_sect))
     {
         const UINT32* note;
@@ -549,12 +560,13 @@ static BOOL image_check_debug_link_gnu_id(const WCHAR* file, struct image_file_m
             {
                 if (note[1] == idlen && !memcmp(note + 3 + ((note[0] + 3) >> 2), id, idlen))
                     return TRUE;
-                WARN("mismatch in buildid information for %s\n", wine_dbgstr_w(file));
             }
         }
         image_unmap_section(&buildid_sect);
         image_unmap_file(fmap);
     }
+    if (ret)
+        WARN("mismatch in buildid information for %s\n", debugstr_w(file));
     return FALSE;
 }
 
@@ -634,13 +646,13 @@ static struct image_file_map* image_locate_debug_link(const struct module* modul
     if (image_check_debug_link_crc(slash, fmap_link, crc)) goto found;
 
 
-    WARN("Couldn't locate or map %s\n", filename);
+    WARN("Couldn't locate or map %s\n", debugstr_a(filename));
     HeapFree(GetProcessHeap(), 0, p);
     HeapFree(GetProcessHeap(), 0, fmap_link);
     return NULL;
 
 found:
-    TRACE("Located debug information file %s at %s\n", filename, debugstr_w(p));
+    TRACE("Located debug information file %s at %s\n", debugstr_a(filename), debugstr_w(p));
     HeapFree(GetProcessHeap(), 0, p);
     return fmap_link;
 }
@@ -656,69 +668,75 @@ static WCHAR* append_hex(WCHAR* dst, const BYTE* id, const BYTE* end)
     return dst;
 }
 
-/******************************************************************
- *		image_locate_build_id_target
- *
- * Try to find the .so file containing the debug info out of the build-id note information
- */
-static struct image_file_map* image_locate_build_id_target(const BYTE* id, unsigned idlen)
+static BOOL image_locate_build_id_target_in_dir(struct image_file_map *fmap_link, const BYTE* id, unsigned idlen, const WCHAR *from)
 {
-    struct image_file_map* fmap_link = NULL;
-    DWORD sz;
-    WCHAR* p;
-    WCHAR* z;
+    size_t from_len = wcslen(from);
+    BOOL found = FALSE;
+    WCHAR *p, *z;
 
-    fmap_link = HeapAlloc(GetProcessHeap(), 0, sizeof(*fmap_link));
-    if (!fmap_link) return NULL;
-
-    p = malloc(sizeof(L"/usr/lib/debug/.build-id/") +
-               (idlen * 2 + 1) * sizeof(WCHAR) + sizeof(L".debug"));
-    if (!p) goto fail;
-    wcscpy(p, L"/usr/lib/debug/.build-id/");
-    z = p + wcslen(p);
-    if (idlen)
+    if ((p = malloc((from_len + idlen * 2 + 1) * sizeof(WCHAR) + sizeof(L".debug"))))
     {
+        memcpy(p, from, from_len * sizeof(WCHAR));
+        z = p + from_len;
         z = append_hex(z, id, id + 1);
         if (idlen > 1)
         {
             *z++ = L'/';
             z = append_hex(z, id + 1, id + idlen);
         }
-    }
-    wcscpy(z, L".debug");
-    TRACE("checking %s\n", wine_dbgstr_w(p));
+        wcscpy(z, L".debug");
+        TRACE("checking %s\n", debugstr_w(p));
+        found = image_check_debug_link_gnu_id(p, fmap_link, id, idlen);
 
-    if (image_check_debug_link_gnu_id(p, fmap_link, id, idlen))
-    {
         free(p);
-        return fmap_link;
     }
+    return found;
+}
+
+/******************************************************************
+ *		image_locate_build_id_target
+ *
+ * Try to find an image file containing the debug info out of the build-id
+ * note information.
+ */
+static struct image_file_map* image_locate_build_id_target(const BYTE* id, unsigned idlen)
+{
+    struct image_file_map* fmap_link;
+    DWORD sz;
+
+    if (!idlen) return NULL;
+
+    if (!(fmap_link = HeapAlloc(GetProcessHeap(), 0, sizeof(*fmap_link))))
+        return NULL;
+    if (image_locate_build_id_target_in_dir(fmap_link, id, idlen, L"/usr/lib/debug/.build-id/"))
+        return fmap_link;
+    if (image_locate_build_id_target_in_dir(fmap_link, id, idlen, L"/usr/lib/.build-id/"))
+        return fmap_link;
 
     sz = GetEnvironmentVariableW(L"WINEHOMEDIR", NULL, 0);
     if (sz)
     {
-        z = realloc(p, sz * sizeof(WCHAR) +
-                    sizeof(L"\\.cache\\debuginfod_client\\") +
-                    idlen * 2 * sizeof(WCHAR) + sizeof(L"\\debuginfo") + 500);
-        if (!z) goto fail;
-        p = z;
-        GetEnvironmentVariableW(L"WINEHOMEDIR", p, sz);
-        z = p + sz - 1;
-        wcscpy(z, L"\\.cache\\debuginfod_client\\");
-        z += wcslen(z);
-        z = append_hex(z, id, id + idlen);
-        wcscpy(z, L"\\debuginfo");
-        TRACE("checking %ls\n", p);
-        if (image_check_debug_link_gnu_id(p, fmap_link, id, idlen))
+        WCHAR *p, *z;
+        p = malloc(sz * sizeof(WCHAR) +
+                   sizeof(L"\\.cache\\debuginfod_client\\") +
+                   idlen * 2 * sizeof(WCHAR) + sizeof(L"\\debuginfo") + 500);
+        if (p && GetEnvironmentVariableW(L"WINEHOMEDIR", p, sz) == sz - 1)
         {
+            BOOL found;
+
+            wcscpy(p + sz - 1, L"\\.cache\\debuginfod_client\\");
+            z = p + wcslen(p);
+            z = append_hex(z, id, id + idlen);
+            wcscpy(z, L"\\debuginfo");
+            TRACE("checking %ls\n", p);
+            found = image_check_debug_link_gnu_id(p, fmap_link, id, idlen);
             free(p);
-            return fmap_link;
+            if (found) return fmap_link;
         }
     }
 
     TRACE("not found\n");
-fail:
-    free(p);
+
     HeapFree(GetProcessHeap(), 0, fmap_link);
     return NULL;
 }
@@ -807,7 +825,8 @@ struct image_file_map* image_load_debugaltlink(struct image_file_map* fmap, stru
                     HeapFree(GetProcessHeap(), 0, fmap_link);
                     /* didn't work out with filename, try file lookup based on build-id */
                     if (!(fmap_link = image_locate_build_id_target(id, idlen)))
-                        WARN("Couldn't find a match for .gnu_debugaltlink section %s for %s\n", data, debugstr_w(module->modulename));
+                        WARN("Couldn't find a match for .gnu_debugaltlink section %s for %s\n",
+                             debugstr_a(data), debugstr_w(module->modulename));
                 }
             }
         }
@@ -829,7 +848,18 @@ BOOL image_check_alternate(struct image_file_map* fmap, const struct module* mod
     struct image_file_map* fmap_link = NULL;
 
     /* if present, add the .gnu_debuglink file as an alternate to current one */
-    if (image_find_section(fmap, ".note.gnu.build-id", &buildid_sect))
+    if (fmap->modtype == DMT_PE)
+    {
+        GUID guid;
+
+        if (pe_has_buildid_debug(fmap, &guid))
+        {
+            /* reorder bytes to match little endian order */
+            fmap_link = image_locate_build_id_target((const BYTE*)&guid, sizeof(guid));
+        }
+    }
+    /* if present, add the .note.gnu.build-id as an alternate to current one */
+    if (!fmap_link && image_find_section(fmap, ".note.gnu.build-id", &buildid_sect))
     {
         const UINT32* note;
 
@@ -872,6 +902,7 @@ BOOL image_check_alternate(struct image_file_map* fmap, const struct module* mod
     return FALSE;
 }
 
+#ifndef _WIN64
 /***********************************************************************
  *			SymLoadModule (DBGHELP.@)
  */
@@ -881,6 +912,7 @@ DWORD WINAPI SymLoadModule(HANDLE hProcess, HANDLE hFile, PCSTR ImageName,
     return SymLoadModuleEx(hProcess, hFile, ImageName, ModuleName, BaseOfDll,
                            SizeOfDll, NULL, 0);
 }
+#endif
 
 /***********************************************************************
  *			SymLoadModuleEx (DBGHELP.@)
@@ -940,7 +972,7 @@ DWORD64 WINAPI  SymLoadModuleExW(HANDLE hProcess, HANDLE hFile, PCWSTR wImageNam
 
     if (!(pcs = process_find_by_handle(hProcess))) return 0;
 
-    if (Flags & ~(SLMFLAG_VIRTUAL))
+    if (Flags & ~(SLMFLAG_VIRTUAL | SLMFLAG_NO_SYMBOLS))
         FIXME("Unsupported Flags %08lx for %s\n", Flags, debugstr_w(wImageName));
 
     /* Trying to load a new module at the same address of an existing one,
@@ -957,7 +989,7 @@ DWORD64 WINAPI  SymLoadModuleExW(HANDLE hProcess, HANDLE hFile, PCWSTR wImageNam
         }
 
     /* this is a Wine extension to the API just to redo the synchronisation */
-    if (!wImageName && !hFile)
+    if (!wImageName && !hFile && !Flags)
     {
         pcs->loader->synchronize_module_list(pcs);
         return 0;
@@ -965,14 +997,13 @@ DWORD64 WINAPI  SymLoadModuleExW(HANDLE hProcess, HANDLE hFile, PCWSTR wImageNam
 
     if (Flags & SLMFLAG_VIRTUAL)
     {
-        if (!wImageName) return 0;
+        if (!wImageName) wImageName = L"";
         module = module_new(pcs, wImageName, DMT_PE, FALSE, TRUE, BaseOfDll, SizeOfDll, 0, 0, IMAGE_FILE_MACHINE_UNKNOWN);
         if (!module) return 0;
-        module->module.SymType = SymVirtual;
     }
     else
     {
-        /* otherwise, try a regular PE image */
+        /* try PE image */
         module = pe_load_native_module(pcs, wImageName, hFile, BaseOfDll, SizeOfDll);
         if (!module && wImageName)
         {
@@ -981,7 +1012,6 @@ DWORD64 WINAPI  SymLoadModuleExW(HANDLE hProcess, HANDLE hFile, PCWSTR wImageNam
              * In both cases, ensure system module list is up-to-date.
              */
             pcs->loader->synchronize_module_list(pcs);
-            /* try a dll.so... */
             if (module_is_container_loaded(pcs, wImageName, BaseOfDll))
                 module = pe_load_builtin_module(pcs, wImageName, BaseOfDll, SizeOfDll);
             /* at last, try ELF or Mach-O module */
@@ -995,11 +1025,11 @@ DWORD64 WINAPI  SymLoadModuleExW(HANDLE hProcess, HANDLE hFile, PCWSTR wImageNam
             return 0;
         }
     }
-    /* by default module_new fills module.ModuleName from a derivation
-     * of LoadedImageName. Overwrite it, if we have better information
-     */
+    if (Flags & SLMFLAG_NO_SYMBOLS) module->dont_load_symbols = 1;
+
+    /* Store alternate name for module when provided. */
     if (wModuleName)
-        module_set_module(module, wModuleName);
+        module->alt_modulename = pool_wcsdup(&module->pool, wModuleName);
     if (wImageName)
         lstrcpynW(module->module.ImageName, wImageName, ARRAY_SIZE(module->module.ImageName));
 
@@ -1051,9 +1081,8 @@ DWORD64 WINAPI SymLoadModule64(HANDLE hProcess, HANDLE hFile, PCSTR ImageName,
  */
 BOOL module_remove(struct process* pcs, struct module* module)
 {
-    struct module_format*modfmt;
+    struct module_format_vtable_iterator iter = {};
     struct module**     p;
-    unsigned            i;
 
     TRACE("%s (%p)\n", debugstr_w(module->modulename), module);
 
@@ -1065,20 +1094,20 @@ BOOL module_remove(struct process* pcs, struct module* module)
             locsym = &symt_get_function_from_inlined((struct symt_function*)locsym)->symt;
         if (symt_check_tag(locsym, SymTagFunction))
         {
-            locsym = ((struct symt_function*)locsym)->container;
-            if (symt_check_tag(locsym, SymTagCompiland) &&
-                module == ((struct symt_compiland*)locsym)->container->module)
+            struct symt_compiland *compiland = (struct symt_compiland*)SYMT_SYMREF_TO_PTR(((struct symt_function*)locsym)->container);
+            if (symt_check_tag(&compiland->symt, SymTagCompiland))
             {
-                pcs->localscope_pc = 0;
-                pcs->localscope_symt = NULL;
+                if (module == ((struct symt_module*)SYMT_SYMREF_TO_PTR(compiland->container))->module)
+                {
+                    pcs->localscope_pc = 0;
+                    pcs->localscope_symt = NULL;
+                }
             }
         }
     }
-    for (i = 0; i < DFI_LAST; i++)
-    {
-        if ((modfmt = module->format_info[i]) && modfmt->remove)
-            modfmt->remove(pcs, module->format_info[i]);
-    }
+    while (module_format_vtable_iterator_next(module, &iter, MODULE_FORMAT_VTABLE_INDEX(remove)))
+        iter.modfmt->vtable->remove(iter.modfmt);
+
     hash_table_destroy(&module->ht_symbols);
     hash_table_destroy(&module->ht_types);
     HeapFree(GetProcessHeap(), 0, module->sources);
@@ -1100,6 +1129,7 @@ BOOL module_remove(struct process* pcs, struct module* module)
     return FALSE;
 }
 
+#ifndef _WIN64
 /******************************************************************
  *		SymUnloadModule (DBGHELP.@)
  *
@@ -1108,6 +1138,7 @@ BOOL WINAPI SymUnloadModule(HANDLE hProcess, DWORD BaseOfDll)
 {
     return SymUnloadModule64(hProcess, BaseOfDll);
 }
+#endif
 
 /******************************************************************
  *		SymUnloadModule64 (DBGHELP.@)
@@ -1126,6 +1157,7 @@ BOOL WINAPI SymUnloadModule64(HANDLE hProcess, DWORD64 BaseOfDll)
     return TRUE;
 }
 
+#ifndef _WIN64
 /******************************************************************
  *		SymEnumerateModules (DBGHELP.@)
  *
@@ -1156,6 +1188,7 @@ BOOL  WINAPI SymEnumerateModules(HANDLE hProcess,
 
     return SymEnumerateModulesW64(hProcess, enum_modW64_32, &x);
 }
+#endif
 
 /******************************************************************
  *		SymEnumerateModules64 (DBGHELP.@)
@@ -1245,6 +1278,7 @@ BOOL  WINAPI EnumerateLoadedModules64(HANDLE hProcess,
     return EnumerateLoadedModulesW64(hProcess, enum_load_modW64_64, &x);
 }
 
+#ifndef _WIN64
 /******************************************************************
  *		EnumerateLoadedModules (DBGHELP.@)
  *
@@ -1275,6 +1309,7 @@ BOOL  WINAPI EnumerateLoadedModules(HANDLE hProcess,
 
     return EnumerateLoadedModulesW64(hProcess, enum_load_modW64_32, &x);
 }
+#endif
 
 static unsigned int load_and_grow_modules(HANDLE process, HMODULE** hmods, unsigned start, unsigned* alloc, DWORD filter)
 {
@@ -1301,6 +1336,7 @@ BOOL  WINAPI EnumerateLoadedModulesW64(HANDLE process,
                                        PENUMLOADED_MODULES_CALLBACKW64 enum_cb,
                                        PVOID user)
 {
+    OBJECT_BASIC_INFORMATION obi;
     HMODULE*            hmods;
     unsigned            alloc = 256, count, count32, i;
     USHORT              pcs_machine, native_machine;
@@ -1311,8 +1347,21 @@ BOOL  WINAPI EnumerateLoadedModulesW64(HANDLE process,
     WCHAR*              wowdir = NULL;
     size_t              sysdir_len = 0, wowdir_len = 0;
 
+    if (process != GetCurrentProcess() &&
+        RtlIsCurrentProcess( process ) &&
+        !NtQueryObject(process, ObjectBasicInformation, &obi, sizeof(obi), NULL) &&
+        obi.GrantedAccess & PROCESS_VM_READ)
+    {
+        TRACE("same process.\n");
+        process = GetCurrentProcess();
+    }
+
     /* process might not be a handle to a live process */
-    if (!IsWow64Process2(process, &pcs_machine, &native_machine)) return FALSE;
+    if (!IsWow64Process2(process, &pcs_machine, &native_machine))
+    {
+        SetLastError(STATUS_INVALID_CID);
+        return FALSE;
+    }
     with_32bit_modules = sizeof(void*) > sizeof(int) &&
         pcs_machine != IMAGE_FILE_MACHINE_UNKNOWN &&
         (dbghelp_options & SYMOPT_INCLUDE_32BIT_MODULES);
@@ -1396,6 +1445,7 @@ static void dbghelp_str_WtoA(const WCHAR *src, char *dst, int dst_len)
     dst[dst_len - 1] = 0;
 }
 
+#ifndef _WIN64
 /******************************************************************
  *		SymGetModuleInfo (DBGHELP.@)
  *
@@ -1456,6 +1506,7 @@ BOOL  WINAPI SymGetModuleInfoW(HANDLE hProcess, DWORD dwAddr,
 
     return TRUE;
 }
+#endif
 
 /******************************************************************
  *		SymGetModuleInfo64 (DBGHELP.@)
@@ -1531,6 +1582,11 @@ BOOL  WINAPI SymGetModuleInfoW64(HANDLE hProcess, DWORD64 dwAddr,
 
     if (dbghelp_opt_real_path && module->real_path)
         lstrcpynW(miw64.LoadedImageName, module->real_path, ARRAY_SIZE(miw64.LoadedImageName));
+    else if (miw64.SymType == SymDeferred)
+    {
+        miw64.LoadedImageName[0] = '\0';
+        miw64.TimeDateStamp = 0;
+    }
 
     /* update debug information from container if any */
     if (module->module.SymType == SymNone)
@@ -1546,6 +1602,7 @@ BOOL  WINAPI SymGetModuleInfoW64(HANDLE hProcess, DWORD64 dwAddr,
     return TRUE;
 }
 
+#ifndef _WIN64
 /***********************************************************************
  *		SymGetModuleBase (DBGHELP.@)
  */
@@ -1553,6 +1610,7 @@ DWORD WINAPI SymGetModuleBase(HANDLE hProcess, DWORD dwAddr)
 {
     return (DWORD)SymGetModuleBase64(hProcess, dwAddr);
 }
+#endif
 
 /***********************************************************************
  *		SymGetModuleBase64 (DBGHELP.@)
@@ -1584,26 +1642,49 @@ void module_reset_debug_info(struct module* module)
     hash_table_destroy(&module->ht_types);
     module->ht_types.num_buckets = 0;
     module->ht_types.buckets = NULL;
-    module->vtypes.num_elts = 0;
     hash_table_destroy(&module->ht_symbols);
     module->sources_used = module->sources_alloc = 0;
     module->sources = NULL;
 }
 
+static BOOL WINAPI process_invade_cb(PCWSTR name, ULONG64 base, ULONG size, PVOID user)
+{
+    HANDLE      hProcess = user;
+
+    /* Note: this follows native behavior:
+     * If a PE module has been unloaded from debuggee, it's not immediately removed
+     * from module list in dbghelp.
+     * Removal may eventually happen when loading a another module with SymLoadModule:
+     * if the module to be loaded overlaps an existing one, SymLoadModule will
+     * automatically unload the eldest one.
+     */
+    SymLoadModuleExW(hProcess, 0, name, NULL, base, size, NULL, 0);
+    return TRUE;
+}
+
+BOOL module_refresh_list(struct process *pcs)
+{
+    BOOL ret;
+
+    ret = pcs->loader->synchronize_module_list(pcs);
+    ret = EnumerateLoadedModulesW64(pcs->handle, process_invade_cb, pcs->handle) && ret;
+    return ret;
+}
+
 /******************************************************************
  *              SymRefreshModuleList (DBGHELP.@)
  */
 BOOL WINAPI SymRefreshModuleList(HANDLE hProcess)
 {
-    struct process*     pcs;
+    struct process *pcs;
 
     TRACE("(%p)\n", hProcess);
 
     if (!(pcs = process_find_by_handle(hProcess))) return FALSE;
-
-    return pcs->loader->synchronize_module_list(pcs);
+    return module_refresh_list(pcs);
 }
 
+#ifndef _WIN64
 /***********************************************************************
  *		SymFunctionTableAccess (DBGHELP.@)
  */
@@ -1611,6 +1692,7 @@ PVOID WINAPI SymFunctionTableAccess(HANDLE hProcess, DWORD AddrBase)
 {
     return SymFunctionTableAccess64(hProcess, AddrBase);
 }
+#endif
 
 /***********************************************************************
  *		SymFunctionTableAccess64 (DBGHELP.@)
