@@ -842,7 +842,7 @@ static BOOL get_mono_path_datadir(LPWSTR path)
 {
     static const WCHAR winedatadirW[] = {'W','I','N','E','D','A','T','A','D','I','R',0};
     static const WCHAR winebuilddirW[] = {'W','I','N','E','B','U','I','L','D','D','I','R',0};
-    static const WCHAR unix_prefix[] = {'\\','?','?','\\','u','n','i','x','\\'};
+    static const WCHAR unix_prefix[] = {'\\','?','?','\\','u','n','i','x','\\',0};
     static const WCHAR monoW[] = {'\\','m','o','n','o',0};
     static const WCHAR dotdotmonoW[] = {'\\','.','.','\\','m','o','n','o',0};
     const WCHAR *data_dir, *suffix;
@@ -1639,7 +1639,7 @@ static DWORD get_assembly_search_flags(MonoAssemblyName *aname)
         return result;
     }
 
-    name_copy = malloc((strlen(name) + 3) * sizeof(WCHAR));
+    name_copy = malloc(strlen(name) + 3);
     if (!name_copy)
     {
         ERR("out of memory\n");
