@@ -70,61 +70,8 @@ static inline HTMLFrameBase *impl_from_IHTMLFrameBase(IHTMLFrameBase *iface)
     return CONTAINING_RECORD(iface, HTMLFrameBase, IHTMLFrameBase_iface);
 }
 
-static HRESULT WINAPI HTMLFrameBase_QueryInterface(IHTMLFrameBase *iface, REFIID riid, void **ppv)
-{
-    HTMLFrameBase *This = impl_from_IHTMLFrameBase(iface);
-
-    return IHTMLDOMNode_QueryInterface(&This->element.node.IHTMLDOMNode_iface, riid, ppv);
-}
-
-static ULONG WINAPI HTMLFrameBase_AddRef(IHTMLFrameBase *iface)
-{
-    HTMLFrameBase *This = impl_from_IHTMLFrameBase(iface);
-
-    return IHTMLDOMNode_AddRef(&This->element.node.IHTMLDOMNode_iface);
-}
-
-static ULONG WINAPI HTMLFrameBase_Release(IHTMLFrameBase *iface)
-{
-    HTMLFrameBase *This = impl_from_IHTMLFrameBase(iface);
-
-    return IHTMLDOMNode_Release(&This->element.node.IHTMLDOMNode_iface);
-}
-
-static HRESULT WINAPI HTMLFrameBase_GetTypeInfoCount(IHTMLFrameBase *iface, UINT *pctinfo)
-{
-    HTMLFrameBase *This = impl_from_IHTMLFrameBase(iface);
-
-    return IDispatchEx_GetTypeInfoCount(&This->element.node.event_target.dispex.IDispatchEx_iface, pctinfo);
-}
-
-static HRESULT WINAPI HTMLFrameBase_GetTypeInfo(IHTMLFrameBase *iface, UINT iTInfo,
-        LCID lcid, ITypeInfo **ppTInfo)
-{
-    HTMLFrameBase *This = impl_from_IHTMLFrameBase(iface);
-
-    return IDispatchEx_GetTypeInfo(&This->element.node.event_target.dispex.IDispatchEx_iface, iTInfo, lcid,
-            ppTInfo);
-}
-
-static HRESULT WINAPI HTMLFrameBase_GetIDsOfNames(IHTMLFrameBase *iface, REFIID riid,
-        LPOLESTR *rgszNames, UINT cNames, LCID lcid, DISPID *rgDispId)
-{
-    HTMLFrameBase *This = impl_from_IHTMLFrameBase(iface);
-
-    return IDispatchEx_GetIDsOfNames(&This->element.node.event_target.dispex.IDispatchEx_iface, riid, rgszNames,
-            cNames, lcid, rgDispId);
-}
-
-static HRESULT WINAPI HTMLFrameBase_Invoke(IHTMLFrameBase *iface, DISPID dispIdMember,
-        REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS *pDispParams,
-        VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT *puArgErr)
-{
-    HTMLFrameBase *This = impl_from_IHTMLFrameBase(iface);
-
-    return IDispatchEx_Invoke(&This->element.node.event_target.dispex.IDispatchEx_iface, dispIdMember, riid,
-            lcid, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
-}
+DISPEX_IDISPATCH_IMPL(HTMLFrameBase, IHTMLFrameBase,
+                      impl_from_IHTMLFrameBase(iface)->element.node.event_target.dispex)
 
 static HRESULT WINAPI HTMLFrameBase_put_src(IHTMLFrameBase *iface, BSTR v)
 {
@@ -132,7 +79,7 @@ static HRESULT WINAPI HTMLFrameBase_put_src(IHTMLFrameBase *iface, BSTR v)
 
     TRACE("(%p)->(%s)\n", This, debugstr_w(v));
 
-    if(!This->content_window || !This->element.node.doc || !This->element.node.doc->window || !This->element.node.doc->window->base.outer_window) {
+    if(!This->content_window || !This->element.node.doc || !This->element.node.doc->window || is_detached_window(This->element.node.doc->window)) {
         nsAString nsstr;
         nsresult nsres;
 
@@ -748,59 +695,8 @@ static inline HTMLFrameElement *impl_from_IHTMLFrameElement3(IHTMLFrameElement3
     return CONTAINING_RECORD(iface, HTMLFrameElement, IHTMLFrameElement3_iface);
 }
 
-static HRESULT WINAPI HTMLFrameElement3_QueryInterface(IHTMLFrameElement3 *iface,
-        REFIID riid, void **ppv)
-{
-    HTMLFrameElement *This = impl_from_IHTMLFrameElement3(iface);
-
-    return IHTMLDOMNode_QueryInterface(&This->framebase.element.node.IHTMLDOMNode_iface, riid, ppv);
-}
-
-static ULONG WINAPI HTMLFrameElement3_AddRef(IHTMLFrameElement3 *iface)
-{
-    HTMLFrameElement *This = impl_from_IHTMLFrameElement3(iface);
-
-    return IHTMLDOMNode_AddRef(&This->framebase.element.node.IHTMLDOMNode_iface);
-}
-
-static ULONG WINAPI HTMLFrameElement3_Release(IHTMLFrameElement3 *iface)
-{
-    HTMLFrameElement *This = impl_from_IHTMLFrameElement3(iface);
-
-    return IHTMLDOMNode_Release(&This->framebase.element.node.IHTMLDOMNode_iface);
-}
-
-static HRESULT WINAPI HTMLFrameElement3_GetTypeInfoCount(IHTMLFrameElement3 *iface, UINT *pctinfo)
-{
-    HTMLFrameElement *This = impl_from_IHTMLFrameElement3(iface);
-    return IDispatchEx_GetTypeInfoCount(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface,
-            pctinfo);
-}
-
-static HRESULT WINAPI HTMLFrameElement3_GetTypeInfo(IHTMLFrameElement3 *iface, UINT iTInfo,
-        LCID lcid, ITypeInfo **ppTInfo)
-{
-    HTMLFrameElement *This = impl_from_IHTMLFrameElement3(iface);
-    return IDispatchEx_GetTypeInfo(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, iTInfo,
-            lcid, ppTInfo);
-}
-
-static HRESULT WINAPI HTMLFrameElement3_GetIDsOfNames(IHTMLFrameElement3 *iface, REFIID riid,
-        LPOLESTR *rgszNames, UINT cNames, LCID lcid, DISPID *rgDispId)
-{
-    HTMLFrameElement *This = impl_from_IHTMLFrameElement3(iface);
-    return IDispatchEx_GetIDsOfNames(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, riid,
-            rgszNames, cNames, lcid, rgDispId);
-}
-
-static HRESULT WINAPI HTMLFrameElement3_Invoke(IHTMLFrameElement3 *iface, DISPID dispIdMember,
-        REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS *pDispParams,
-        VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT *puArgErr)
-{
-    HTMLFrameElement *This = impl_from_IHTMLFrameElement3(iface);
-    return IDispatchEx_Invoke(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, dispIdMember,
-            riid, lcid, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
-}
+DISPEX_IDISPATCH_IMPL(HTMLFrameElement3, IHTMLFrameElement3,
+                      impl_from_IHTMLFrameElement3(iface)->framebase.element.node.event_target.dispex)
 
 static HRESULT WINAPI HTMLFrameElement3_get_contentDocument(IHTMLFrameElement3 *iface, IDispatch **p)
 {
@@ -964,7 +860,7 @@ static void HTMLFrameElement_destructor(DispatchEx *dispex)
     HTMLFrameBase_destructor(&This->framebase);
 }
 
-static HRESULT HTMLFrameElement_get_dispid(DispatchEx *dispex, BSTR name, DWORD grfdex, DISPID *dispid)
+static HRESULT HTMLFrameElement_get_dispid(DispatchEx *dispex, const WCHAR *name, DWORD grfdex, DISPID *dispid)
 {
     HTMLFrameElement *This = frame_from_DispatchEx(dispex);
 
@@ -974,21 +870,19 @@ static HRESULT HTMLFrameElement_get_dispid(DispatchEx *dispex, BSTR name, DWORD
     return search_window_props(This->framebase.content_window->base.inner_window, name, grfdex, dispid);
 }
 
-static HRESULT HTMLFrameElement_get_name(DispatchEx *dispex, DISPID id, BSTR *name)
+static HRESULT HTMLFrameElement_get_prop_desc(DispatchEx *dispex, DISPID id, struct property_info *desc)
 {
     HTMLFrameElement *This = frame_from_DispatchEx(dispex);
-    DWORD idx = id - MSHTML_DISPID_CUSTOM_MIN;
 
-    if(!This->framebase.content_window ||
-       idx >= This->framebase.content_window->base.inner_window->global_prop_cnt)
+    if(!This->framebase.content_window)
         return DISP_E_MEMBERNOTFOUND;
 
-    *name = SysAllocString(This->framebase.content_window->base.inner_window->global_props[idx].name);
-    return *name ? S_OK : E_OUTOFMEMORY;
+    return HTMLWindow_get_prop_desc(&This->framebase.content_window->base.inner_window->event_target.dispex,
+                                    id, desc);
 }
 
-static HRESULT HTMLFrameElement_invoke(DispatchEx *dispex, IDispatch *this_obj, DISPID id, LCID lcid, WORD flags,
-        DISPPARAMS *params, VARIANT *res, EXCEPINFO *ei, IServiceProvider *caller)
+static HRESULT HTMLFrameElement_invoke(DispatchEx *dispex, DISPID id, LCID lcid, WORD flags, DISPPARAMS *params,
+        VARIANT *res, EXCEPINFO *ei, IServiceProvider *caller)
 {
     HTMLFrameElement *This = frame_from_DispatchEx(dispex);
 
@@ -997,8 +891,8 @@ static HRESULT HTMLFrameElement_invoke(DispatchEx *dispex, IDispatch *this_obj,
         return E_FAIL;
     }
 
-    return dispex_invoke(&This->framebase.content_window->base.inner_window->event_target.dispex,
-                         this_obj, id, lcid, flags, params, res, ei, caller);
+    return HTMLWindow_invoke(&This->framebase.content_window->base.inner_window->event_target.dispex,
+                             id, lcid, flags, params, res, ei, caller);
 }
 
 static const NodeImplVtbl HTMLFrameElementImplVtbl = {
@@ -1019,7 +913,7 @@ static const event_target_vtbl_t HTMLFrameElement_event_target_vtbl = {
         .traverse       = HTMLFrameElement_traverse,
         .unlink         = HTMLFrameElement_unlink,
         .get_dispid     = HTMLFrameElement_get_dispid,
-        .get_name       = HTMLFrameElement_get_name,
+        .get_prop_desc  = HTMLFrameElement_get_prop_desc,
         .invoke         = HTMLFrameElement_invoke
     },
     HTMLELEMENT_EVENT_TARGET_VTBL_ENTRIES,
@@ -1027,7 +921,6 @@ static const event_target_vtbl_t HTMLFrameElement_event_target_vtbl = {
 };
 
 static const tid_t HTMLFrameElement_iface_tids[] = {
-    HTMLELEMENT_TIDS,
     IHTMLFrameBase_tid,
     IHTMLFrameBase2_tid,
     IHTMLFrameElement3_tid,
@@ -1035,12 +928,13 @@ static const tid_t HTMLFrameElement_iface_tids[] = {
 };
 
 dispex_static_data_t HTMLFrameElement_dispex = {
-    "HTMLFrameElement",
-    &HTMLFrameElement_event_target_vtbl.dispex_vtbl,
-    PROTO_ID_HTMLFrameElement,
-    DispHTMLFrameElement_tid,
-    HTMLFrameElement_iface_tids,
-    HTMLElement_init_dispex_info
+    .id           = PROT_HTMLFrameElement,
+    .prototype_id = PROT_HTMLElement,
+    .vtbl         = &HTMLFrameElement_event_target_vtbl.dispex_vtbl,
+    .disp_tid     = DispHTMLFrameElement_tid,
+    .iface_tids   = HTMLFrameElement_iface_tids,
+    .init_info    = HTMLElement_init_dispex_info,
+    .js_flags     = HOSTOBJ_VOLATILE_PROPS
 };
 
 HRESULT HTMLFrameElement_Create(HTMLDocumentNode *doc, nsIDOMElement *nselem, HTMLElement **elem)
@@ -1072,59 +966,8 @@ static inline HTMLIFrame *impl_from_IHTMLIFrameElement(IHTMLIFrameElement *iface
     return CONTAINING_RECORD(iface, HTMLIFrame, IHTMLIFrameElement_iface);
 }
 
-static HRESULT WINAPI HTMLIFrameElement_QueryInterface(IHTMLIFrameElement *iface,
-        REFIID riid, void **ppv)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement(iface);
-
-    return IHTMLDOMNode_QueryInterface(&This->framebase.element.node.IHTMLDOMNode_iface, riid, ppv);
-}
-
-static ULONG WINAPI HTMLIFrameElement_AddRef(IHTMLIFrameElement *iface)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement(iface);
-
-    return IHTMLDOMNode_AddRef(&This->framebase.element.node.IHTMLDOMNode_iface);
-}
-
-static ULONG WINAPI HTMLIFrameElement_Release(IHTMLIFrameElement *iface)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement(iface);
-
-    return IHTMLDOMNode_Release(&This->framebase.element.node.IHTMLDOMNode_iface);
-}
-
-static HRESULT WINAPI HTMLIFrameElement_GetTypeInfoCount(IHTMLIFrameElement *iface, UINT *pctinfo)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement(iface);
-    return IDispatchEx_GetTypeInfoCount(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface,
-            pctinfo);
-}
-
-static HRESULT WINAPI HTMLIFrameElement_GetTypeInfo(IHTMLIFrameElement *iface, UINT iTInfo,
-        LCID lcid, ITypeInfo **ppTInfo)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement(iface);
-    return IDispatchEx_GetTypeInfo(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, iTInfo,
-            lcid, ppTInfo);
-}
-
-static HRESULT WINAPI HTMLIFrameElement_GetIDsOfNames(IHTMLIFrameElement *iface, REFIID riid,
-        LPOLESTR *rgszNames, UINT cNames, LCID lcid, DISPID *rgDispId)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement(iface);
-    return IDispatchEx_GetIDsOfNames(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, riid,
-            rgszNames, cNames, lcid, rgDispId);
-}
-
-static HRESULT WINAPI HTMLIFrameElement_Invoke(IHTMLIFrameElement *iface, DISPID dispIdMember,
-        REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS *pDispParams,
-        VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT *puArgErr)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement(iface);
-    return IDispatchEx_Invoke(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, dispIdMember,
-            riid, lcid, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
-}
+DISPEX_IDISPATCH_IMPL(HTMLIFrameElement, IHTMLIFrameElement,
+                      impl_from_IHTMLIFrameElement(iface)->framebase.element.node.event_target.dispex)
 
 static HRESULT WINAPI HTMLIFrameElement_put_vspace(IHTMLIFrameElement *iface, LONG v)
 {
@@ -1189,59 +1032,8 @@ static inline HTMLIFrame *impl_from_IHTMLIFrameElement2(IHTMLIFrameElement2 *ifa
     return CONTAINING_RECORD(iface, HTMLIFrame, IHTMLIFrameElement2_iface);
 }
 
-static HRESULT WINAPI HTMLIFrameElement2_QueryInterface(IHTMLIFrameElement2 *iface,
-        REFIID riid, void **ppv)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement2(iface);
-
-    return IHTMLDOMNode_QueryInterface(&This->framebase.element.node.IHTMLDOMNode_iface, riid, ppv);
-}
-
-static ULONG WINAPI HTMLIFrameElement2_AddRef(IHTMLIFrameElement2 *iface)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement2(iface);
-
-    return IHTMLDOMNode_AddRef(&This->framebase.element.node.IHTMLDOMNode_iface);
-}
-
-static ULONG WINAPI HTMLIFrameElement2_Release(IHTMLIFrameElement2 *iface)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement2(iface);
-
-    return IHTMLDOMNode_Release(&This->framebase.element.node.IHTMLDOMNode_iface);
-}
-
-static HRESULT WINAPI HTMLIFrameElement2_GetTypeInfoCount(IHTMLIFrameElement2 *iface, UINT *pctinfo)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement2(iface);
-    return IDispatchEx_GetTypeInfoCount(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface,
-            pctinfo);
-}
-
-static HRESULT WINAPI HTMLIFrameElement2_GetTypeInfo(IHTMLIFrameElement2 *iface, UINT iTInfo,
-        LCID lcid, ITypeInfo **ppTInfo)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement2(iface);
-    return IDispatchEx_GetTypeInfo(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, iTInfo,
-            lcid, ppTInfo);
-}
-
-static HRESULT WINAPI HTMLIFrameElement2_GetIDsOfNames(IHTMLIFrameElement2 *iface, REFIID riid,
-        LPOLESTR *rgszNames, UINT cNames, LCID lcid, DISPID *rgDispId)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement2(iface);
-    return IDispatchEx_GetIDsOfNames(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, riid,
-            rgszNames, cNames, lcid, rgDispId);
-}
-
-static HRESULT WINAPI HTMLIFrameElement2_Invoke(IHTMLIFrameElement2 *iface, DISPID dispIdMember,
-        REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS *pDispParams,
-        VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT *puArgErr)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement2(iface);
-    return IDispatchEx_Invoke(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, dispIdMember,
-            riid, lcid, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
-}
+DISPEX_IDISPATCH_IMPL(HTMLIFrameElement2, IHTMLIFrameElement2,
+                      impl_from_IHTMLIFrameElement2(iface)->framebase.element.node.event_target.dispex)
 
 static HRESULT WINAPI HTMLIFrameElement2_put_height(IHTMLIFrameElement2 *iface, VARIANT v)
 {
@@ -1338,59 +1130,8 @@ static inline HTMLIFrame *impl_from_IHTMLIFrameElement3(IHTMLIFrameElement3 *ifa
     return CONTAINING_RECORD(iface, HTMLIFrame, IHTMLIFrameElement3_iface);
 }
 
-static HRESULT WINAPI HTMLIFrameElement3_QueryInterface(IHTMLIFrameElement3 *iface,
-        REFIID riid, void **ppv)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement3(iface);
-
-    return IHTMLDOMNode_QueryInterface(&This->framebase.element.node.IHTMLDOMNode_iface, riid, ppv);
-}
-
-static ULONG WINAPI HTMLIFrameElement3_AddRef(IHTMLIFrameElement3 *iface)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement3(iface);
-
-    return IHTMLDOMNode_AddRef(&This->framebase.element.node.IHTMLDOMNode_iface);
-}
-
-static ULONG WINAPI HTMLIFrameElement3_Release(IHTMLIFrameElement3 *iface)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement3(iface);
-
-    return IHTMLDOMNode_Release(&This->framebase.element.node.IHTMLDOMNode_iface);
-}
-
-static HRESULT WINAPI HTMLIFrameElement3_GetTypeInfoCount(IHTMLIFrameElement3 *iface, UINT *pctinfo)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement3(iface);
-    return IDispatchEx_GetTypeInfoCount(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface,
-            pctinfo);
-}
-
-static HRESULT WINAPI HTMLIFrameElement3_GetTypeInfo(IHTMLIFrameElement3 *iface, UINT iTInfo,
-        LCID lcid, ITypeInfo **ppTInfo)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement3(iface);
-    return IDispatchEx_GetTypeInfo(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, iTInfo,
-            lcid, ppTInfo);
-}
-
-static HRESULT WINAPI HTMLIFrameElement3_GetIDsOfNames(IHTMLIFrameElement3 *iface, REFIID riid,
-        LPOLESTR *rgszNames, UINT cNames, LCID lcid, DISPID *rgDispId)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement3(iface);
-    return IDispatchEx_GetIDsOfNames(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, riid,
-            rgszNames, cNames, lcid, rgDispId);
-}
-
-static HRESULT WINAPI HTMLIFrameElement3_Invoke(IHTMLIFrameElement3 *iface, DISPID dispIdMember,
-        REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS *pDispParams,
-        VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT *puArgErr)
-{
-    HTMLIFrame *This = impl_from_IHTMLIFrameElement3(iface);
-    return IDispatchEx_Invoke(&This->framebase.element.node.event_target.dispex.IDispatchEx_iface, dispIdMember,
-            riid, lcid, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
-}
+DISPEX_IDISPATCH_IMPL(HTMLIFrameElement3, IHTMLIFrameElement3,
+                      impl_from_IHTMLIFrameElement3(iface)->framebase.element.node.event_target.dispex)
 
 static HRESULT WINAPI HTMLIFrameElement3_get_contentDocument(IHTMLIFrameElement3 *iface, IDispatch **p)
 {
@@ -1555,7 +1296,7 @@ static void HTMLIFrame_destructor(DispatchEx *dispex)
     HTMLFrameBase_destructor(&This->framebase);
 }
 
-static HRESULT HTMLIFrame_get_dispid(DispatchEx *dispex, BSTR name, DWORD grfdex, DISPID *dispid)
+static HRESULT HTMLIFrame_get_dispid(DispatchEx *dispex, const WCHAR *name, DWORD grfdex, DISPID *dispid)
 {
     HTMLIFrame *This = iframe_from_DispatchEx(dispex);
 
@@ -1565,21 +1306,19 @@ static HRESULT HTMLIFrame_get_dispid(DispatchEx *dispex, BSTR name, DWORD grfdex
     return search_window_props(This->framebase.content_window->base.inner_window, name, grfdex, dispid);
 }
 
-static HRESULT HTMLIFrame_get_name(DispatchEx *dispex, DISPID id, BSTR *name)
+static HRESULT HTMLIFrame_get_prop_desc(DispatchEx *dispex, DISPID id, struct property_info *desc)
 {
     HTMLIFrame *This = iframe_from_DispatchEx(dispex);
-    DWORD idx = id - MSHTML_DISPID_CUSTOM_MIN;
 
-    if(!This->framebase.content_window ||
-       idx >= This->framebase.content_window->base.inner_window->global_prop_cnt)
+    if(!This->framebase.content_window)
         return DISP_E_MEMBERNOTFOUND;
 
-    *name = SysAllocString(This->framebase.content_window->base.inner_window->global_props[idx].name);
-    return *name ? S_OK : E_OUTOFMEMORY;
+    return HTMLWindow_get_prop_desc(&This->framebase.content_window->base.inner_window->event_target.dispex,
+                                    id, desc);
 }
 
-static HRESULT HTMLIFrame_invoke(DispatchEx *dispex, IDispatch *this_obj, DISPID id, LCID lcid, WORD flags,
-        DISPPARAMS *params, VARIANT *res, EXCEPINFO *ei, IServiceProvider *caller)
+static HRESULT HTMLIFrame_invoke(DispatchEx *dispex, DISPID id, LCID lcid, WORD flags, DISPPARAMS *params,
+        VARIANT *res, EXCEPINFO *ei, IServiceProvider *caller)
 {
     HTMLIFrame *This = iframe_from_DispatchEx(dispex);
 
@@ -1588,8 +1327,8 @@ static HRESULT HTMLIFrame_invoke(DispatchEx *dispex, IDispatch *this_obj, DISPID
         return E_FAIL;
     }
 
-    return dispex_invoke(&This->framebase.content_window->base.inner_window->event_target.dispex,
-                         this_obj, id, lcid, flags, params, res, ei, caller);
+    return HTMLWindow_invoke(&This->framebase.content_window->base.inner_window->event_target.dispex,
+                             id, lcid, flags, params, res, ei, caller);
 }
 
 static const NodeImplVtbl HTMLIFrameImplVtbl = {
@@ -1602,7 +1341,7 @@ static const NodeImplVtbl HTMLIFrameImplVtbl = {
     .bind_to_tree          = HTMLIFrame_bind_to_tree,
 };
 
-static const event_target_vtbl_t HTMLIFrame_event_target_vtbl = {
+static const event_target_vtbl_t HTMLIFrameElement_event_target_vtbl = {
     {
         HTMLELEMENT_DISPEX_VTBL_ENTRIES,
         .query_interface= HTMLIFrame_query_interface,
@@ -1610,15 +1349,14 @@ static const event_target_vtbl_t HTMLIFrame_event_target_vtbl = {
         .traverse       = HTMLIFrame_traverse,
         .unlink         = HTMLIFrame_unlink,
         .get_dispid     = HTMLIFrame_get_dispid,
-        .get_name       = HTMLIFrame_get_name,
+        .get_prop_desc  = HTMLIFrame_get_prop_desc,
         .invoke         = HTMLIFrame_invoke
     },
     HTMLELEMENT_EVENT_TARGET_VTBL_ENTRIES,
     .handle_event       = HTMLElement_handle_event
 };
 
-static const tid_t HTMLIFrame_iface_tids[] = {
-    HTMLELEMENT_TIDS,
+static const tid_t HTMLIFrameElement_iface_tids[] = {
     IHTMLFrameBase_tid,
     IHTMLFrameBase2_tid,
     IHTMLIFrameElement_tid,
@@ -1627,13 +1365,14 @@ static const tid_t HTMLIFrame_iface_tids[] = {
     0
 };
 
-dispex_static_data_t HTMLIFrame_dispex = {
-    "HTMLIFrameElement",
-    &HTMLIFrame_event_target_vtbl.dispex_vtbl,
-    PROTO_ID_HTMLIFrameElement,
-    DispHTMLIFrame_tid,
-    HTMLIFrame_iface_tids,
-    HTMLElement_init_dispex_info
+dispex_static_data_t HTMLIFrameElement_dispex = {
+    .id           = PROT_HTMLIFrameElement,
+    .prototype_id = PROT_HTMLElement,
+    .vtbl         = &HTMLIFrameElement_event_target_vtbl.dispex_vtbl,
+    .disp_tid     = DispHTMLIFrame_tid,
+    .iface_tids   = HTMLIFrameElement_iface_tids,
+    .init_info    = HTMLElement_init_dispex_info,
+    .js_flags     = HOSTOBJ_VOLATILE_PROPS
 };
 
 HRESULT HTMLIFrame_Create(HTMLDocumentNode *doc, nsIDOMElement *nselem, HTMLElement **elem)
@@ -1649,7 +1388,7 @@ HRESULT HTMLIFrame_Create(HTMLDocumentNode *doc, nsIDOMElement *nselem, HTMLElem
     ret->IHTMLIFrameElement3_iface.lpVtbl = &HTMLIFrameElement3Vtbl;
     ret->framebase.element.node.vtbl = &HTMLIFrameImplVtbl;
 
-    HTMLFrameBase_Init(&ret->framebase, doc, nselem, &HTMLIFrame_dispex);
+    HTMLFrameBase_Init(&ret->framebase, doc, nselem, &HTMLIFrameElement_dispex);
 
     *elem = &ret->framebase.element;
     return S_OK;
