@@ -46,58 +46,8 @@ static inline HTMLImg *impl_from_IHTMLImgElement(IHTMLImgElement *iface)
     return CONTAINING_RECORD(iface, HTMLImg, IHTMLImgElement_iface);
 }
 
-static HRESULT WINAPI HTMLImgElement_QueryInterface(IHTMLImgElement *iface, REFIID riid, void **ppv)
-{
-    HTMLImg *This = impl_from_IHTMLImgElement(iface);
-
-    return IHTMLDOMNode_QueryInterface(&This->element.node.IHTMLDOMNode_iface, riid, ppv);
-}
-
-static ULONG WINAPI HTMLImgElement_AddRef(IHTMLImgElement *iface)
-{
-    HTMLImg *This = impl_from_IHTMLImgElement(iface);
-
-    return IHTMLDOMNode_AddRef(&This->element.node.IHTMLDOMNode_iface);
-}
-
-static ULONG WINAPI HTMLImgElement_Release(IHTMLImgElement *iface)
-{
-    HTMLImg *This = impl_from_IHTMLImgElement(iface);
-
-    return IHTMLDOMNode_Release(&This->element.node.IHTMLDOMNode_iface);
-}
-
-static HRESULT WINAPI HTMLImgElement_GetTypeInfoCount(IHTMLImgElement *iface, UINT *pctinfo)
-{
-    HTMLImg *This = impl_from_IHTMLImgElement(iface);
-    return IDispatchEx_GetTypeInfoCount(&This->element.node.event_target.dispex.IDispatchEx_iface, pctinfo);
-}
-
-static HRESULT WINAPI HTMLImgElement_GetTypeInfo(IHTMLImgElement *iface, UINT iTInfo,
-                                              LCID lcid, ITypeInfo **ppTInfo)
-{
-    HTMLImg *This = impl_from_IHTMLImgElement(iface);
-    return IDispatchEx_GetTypeInfo(&This->element.node.event_target.dispex.IDispatchEx_iface, iTInfo, lcid,
-            ppTInfo);
-}
-
-static HRESULT WINAPI HTMLImgElement_GetIDsOfNames(IHTMLImgElement *iface, REFIID riid,
-                                                LPOLESTR *rgszNames, UINT cNames,
-                                                LCID lcid, DISPID *rgDispId)
-{
-    HTMLImg *This = impl_from_IHTMLImgElement(iface);
-    return IDispatchEx_GetIDsOfNames(&This->element.node.event_target.dispex.IDispatchEx_iface, riid, rgszNames,
-            cNames, lcid, rgDispId);
-}
-
-static HRESULT WINAPI HTMLImgElement_Invoke(IHTMLImgElement *iface, DISPID dispIdMember,
-                            REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS *pDispParams,
-                            VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT *puArgErr)
-{
-    HTMLImg *This = impl_from_IHTMLImgElement(iface);
-    return IDispatchEx_Invoke(&This->element.node.event_target.dispex.IDispatchEx_iface, dispIdMember, riid,
-            lcid, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
-}
+DISPEX_IDISPATCH_IMPL(HTMLImgElement, IHTMLImgElement,
+                      impl_from_IHTMLImgElement(iface)->element.node.event_target.dispex)
 
 static HRESULT WINAPI HTMLImgElement_put_isMap(IHTMLImgElement *iface, VARIANT_BOOL v)
 {
@@ -716,11 +666,6 @@ static const event_target_vtbl_t HTMLImgElement_event_target_vtbl = {
     .handle_event       = HTMLElement_handle_event
 };
 
-static const tid_t HTMLImgElement_iface_tids[] = {
-    HTMLELEMENT_TIDS,
-    0
-};
-
 static void HTMLImgElement_init_dispex_info(dispex_data_t *info, compat_mode_t mode)
 {
     static const dispex_hook_t img_ie11_hooks[] = {
@@ -733,13 +678,12 @@ static void HTMLImgElement_init_dispex_info(dispex_data_t *info, compat_mode_t m
     dispex_info_add_interface(info, IHTMLImgElement_tid, mode >= COMPAT_MODE_IE11 ? img_ie11_hooks : NULL);
 }
 
-dispex_static_data_t HTMLImgElement_dispex = {
-    "HTMLImageElement",
-    &HTMLImgElement_event_target_vtbl.dispex_vtbl,
-    PROTO_ID_HTMLImgElement,
-    DispHTMLImg_tid,
-    HTMLImgElement_iface_tids,
-    HTMLImgElement_init_dispex_info
+dispex_static_data_t HTMLImageElement_dispex = {
+    .id           = PROT_HTMLImageElement,
+    .prototype_id = PROT_HTMLElement,
+    .vtbl         = &HTMLImgElement_event_target_vtbl.dispex_vtbl,
+    .disp_tid     = DispHTMLImg_tid,
+    .init_info    = HTMLImgElement_init_dispex_info,
 };
 
 HRESULT HTMLImgElement_Create(HTMLDocumentNode *doc, nsIDOMElement *nselem, HTMLElement **elem)
@@ -754,7 +698,7 @@ HRESULT HTMLImgElement_Create(HTMLDocumentNode *doc, nsIDOMElement *nselem, HTML
     ret->IHTMLImgElement_iface.lpVtbl = &HTMLImgElementVtbl;
     ret->element.node.vtbl = &HTMLImgElementImplVtbl;
 
-    HTMLElement_Init(&ret->element, doc, nselem, &HTMLImgElement_dispex);
+    HTMLElement_Init(&ret->element, doc, nselem, &HTMLImageElement_dispex);
 
     nsres = nsIDOMElement_QueryInterface(nselem, &IID_nsIDOMHTMLImageElement, (void**)&ret->nsimg);
     assert(nsres == NS_OK);
@@ -763,61 +707,13 @@ HRESULT HTMLImgElement_Create(HTMLDocumentNode *doc, nsIDOMElement *nselem, HTML
     return S_OK;
 }
 
-static inline struct global_ctor *impl_from_IHTMLImageElementFactory(IHTMLImageElementFactory *iface)
+static inline HTMLImageElementFactory *impl_from_IHTMLImageElementFactory(IHTMLImageElementFactory *iface)
 {
-    return CONTAINING_RECORD(iface, struct global_ctor, IHTMLImageElementFactory_iface);
+    return CONTAINING_RECORD(iface, HTMLImageElementFactory, IHTMLImageElementFactory_iface);
 }
 
-static HRESULT WINAPI HTMLImageElementFactory_QueryInterface(IHTMLImageElementFactory *iface,
-        REFIID riid, void **ppv)
-{
-    struct global_ctor *This = impl_from_IHTMLImageElementFactory(iface);
-    return IDispatchEx_QueryInterface(&This->dispex.IDispatchEx_iface, riid, ppv);
-}
-
-static ULONG WINAPI HTMLImageElementFactory_AddRef(IHTMLImageElementFactory *iface)
-{
-    struct global_ctor *This = impl_from_IHTMLImageElementFactory(iface);
-    return IDispatchEx_AddRef(&This->dispex.IDispatchEx_iface);
-}
-
-static ULONG WINAPI HTMLImageElementFactory_Release(IHTMLImageElementFactory *iface)
-{
-    struct global_ctor *This = impl_from_IHTMLImageElementFactory(iface);
-    return IDispatchEx_Release(&This->dispex.IDispatchEx_iface);
-}
-
-static HRESULT WINAPI HTMLImageElementFactory_GetTypeInfoCount(IHTMLImageElementFactory *iface,
-        UINT *pctinfo)
-{
-    struct global_ctor *This = impl_from_IHTMLImageElementFactory(iface);
-    return IDispatchEx_GetTypeInfoCount(&This->dispex.IDispatchEx_iface, pctinfo);
-}
-
-static HRESULT WINAPI HTMLImageElementFactory_GetTypeInfo(IHTMLImageElementFactory *iface,
-        UINT iTInfo, LCID lcid, ITypeInfo **ppTInfo)
-{
-    struct global_ctor *This = impl_from_IHTMLImageElementFactory(iface);
-    return IDispatchEx_GetTypeInfo(&This->dispex.IDispatchEx_iface, iTInfo, lcid, ppTInfo);
-}
-
-static HRESULT WINAPI HTMLImageElementFactory_GetIDsOfNames(IHTMLImageElementFactory *iface,
-        REFIID riid, LPOLESTR *rgszNames, UINT cNames, LCID lcid,
-        DISPID *rgDispId)
-{
-    struct global_ctor *This = impl_from_IHTMLImageElementFactory(iface);
-    return IDispatchEx_GetIDsOfNames(&This->dispex.IDispatchEx_iface, riid, rgszNames, cNames, lcid, rgDispId);
-}
-
-static HRESULT WINAPI HTMLImageElementFactory_Invoke(IHTMLImageElementFactory *iface,
-        DISPID dispIdMember, REFIID riid, LCID lcid, WORD wFlags,
-        DISPPARAMS *pDispParams, VARIANT *pVarResult, EXCEPINFO *pExcepInfo,
-        UINT *puArgErr)
-{
-    struct global_ctor *This = impl_from_IHTMLImageElementFactory(iface);
-    return IDispatchEx_Invoke(&This->dispex.IDispatchEx_iface, dispIdMember, riid, lcid, wFlags,
-            pDispParams, pVarResult, pExcepInfo, puArgErr);
-}
+DISPEX_IDISPATCH_IMPL(HTMLImageElementFactory, IHTMLImageElementFactory,
+                      impl_from_IHTMLImageElementFactory(iface)->dispex)
 
 static LONG var_to_size(const VARIANT *v)
 {
@@ -846,7 +742,7 @@ static LONG var_to_size(const VARIANT *v)
 static HRESULT WINAPI HTMLImageElementFactory_create(IHTMLImageElementFactory *iface,
         VARIANT width, VARIANT height, IHTMLImgElement **img_elem)
 {
-    struct global_ctor *This = impl_from_IHTMLImageElementFactory(iface);
+    HTMLImageElementFactory *This = impl_from_IHTMLImageElementFactory(iface);
     HTMLDocumentNode *doc = This->window->doc;
     IHTMLImgElement *img;
     HTMLElement *elem;
@@ -889,7 +785,7 @@ static HRESULT WINAPI HTMLImageElementFactory_create(IHTMLImageElementFactory *i
     return S_OK;
 }
 
-const IHTMLImageElementFactoryVtbl HTMLImageElementFactoryVtbl = {
+static const IHTMLImageElementFactoryVtbl HTMLImageElementFactoryVtbl = {
     HTMLImageElementFactory_QueryInterface,
     HTMLImageElementFactory_AddRef,
     HTMLImageElementFactory_Release,
@@ -900,14 +796,14 @@ const IHTMLImageElementFactoryVtbl HTMLImageElementFactoryVtbl = {
     HTMLImageElementFactory_create
 };
 
-static inline struct global_ctor *impl_from_DispatchEx(DispatchEx *iface)
+static inline HTMLImageElementFactory *impl_from_DispatchEx(DispatchEx *iface)
 {
-    return CONTAINING_RECORD(iface, struct global_ctor, dispex);
+    return CONTAINING_RECORD(iface, HTMLImageElementFactory, dispex);
 }
 
 static void *HTMLImageElementFactory_query_interface(DispatchEx *dispex, REFIID riid)
 {
-    struct global_ctor *This = impl_from_DispatchEx(dispex);
+    HTMLImageElementFactory *This = impl_from_DispatchEx(dispex);
 
     if(IsEqualGUID(&IID_IHTMLImageElementFactory, riid))
         return &This->IHTMLImageElementFactory_iface;
@@ -915,19 +811,41 @@ static void *HTMLImageElementFactory_query_interface(DispatchEx *dispex, REFIID
     return NULL;
 }
 
+static void HTMLImageElementFactory_traverse(DispatchEx *dispex, nsCycleCollectionTraversalCallback *cb)
+{
+    HTMLImageElementFactory *This = impl_from_DispatchEx(dispex);
+
+    if(This->window)
+        note_cc_edge((nsISupports*)&This->window->base.IHTMLWindow2_iface, "window", cb);
+}
+
+static void HTMLImageElementFactory_unlink(DispatchEx *dispex)
+{
+    HTMLImageElementFactory *This = impl_from_DispatchEx(dispex);
+
+    if(This->window) {
+        HTMLInnerWindow *window = This->window;
+        This->window = NULL;
+        IHTMLWindow2_Release(&window->base.IHTMLWindow2_iface);
+    }
+}
+
+static void HTMLImageElementFactory_destructor(DispatchEx *dispex)
+{
+    HTMLImageElementFactory *This = impl_from_DispatchEx(dispex);
+    free(This);
+}
+
 static HRESULT HTMLImageElementFactory_value(DispatchEx *dispex, LCID lcid,
         WORD flags, DISPPARAMS *params, VARIANT *res, EXCEPINFO *ei,
         IServiceProvider *caller)
 {
-    struct global_ctor *This = impl_from_DispatchEx(dispex);
+    HTMLImageElementFactory *This = impl_from_DispatchEx(dispex);
     IHTMLImgElement *img;
     VARIANT empty, *width, *height;
     HRESULT hres;
     int argc = params->cArgs - params->cNamedArgs;
 
-    if(flags != DISPATCH_CONSTRUCT)
-        return S_FALSE;
-
     V_VT(res) = VT_NULL;
 
     V_VT(&empty) = VT_EMPTY;
@@ -946,56 +864,43 @@ static HRESULT HTMLImageElementFactory_value(DispatchEx *dispex, LCID lcid,
     return S_OK;
 }
 
-static const dispex_static_data_vtbl_t HTMLImageElementFactory_dispex_vtbl = {
-    .query_interface  = HTMLImageElementFactory_query_interface,
-    .destructor       = global_ctor_destructor,
-    .traverse         = global_ctor_traverse,
-    .unlink           = global_ctor_unlink,
-    .value            = HTMLImageElementFactory_value,
-    .get_dispid       = legacy_ctor_get_dispid,
-    .get_name         = legacy_ctor_get_name,
-    .invoke           = legacy_ctor_invoke,
-    .delete           = legacy_ctor_delete
-};
-
-static const tid_t HTMLImageElementFactory_iface_tids[] = {
-    IHTMLImageElementFactory_tid,
-    0
-};
-
-dispex_static_data_t HTMLImageElementFactory_dispex = {
-    "HTMLImageElement",
-    &HTMLImageElementFactory_dispex_vtbl,
-    PROTO_ID_NULL,
-    IHTMLImageElementFactory_tid,
-    HTMLImageElementFactory_iface_tids
-};
-
-static HRESULT HTMLImageCtor_value(DispatchEx *iface, LCID lcid, WORD flags, DISPPARAMS *params,
-        VARIANT *res, EXCEPINFO *ei, IServiceProvider *caller)
+static void HTMLImageElementFactory_init_dispex_info(dispex_data_t *info, compat_mode_t mode)
 {
-    if(flags == DISPATCH_CONSTRUCT)
-        return HTMLImageElementFactory_value(iface, lcid, flags, params, res, ei, caller);
-
-    return global_ctor_value(iface, lcid, flags, params, res, ei, caller);
+    if(mode < COMPAT_MODE_IE9)
+        dispex_info_add_interface(info, IHTMLImageElementFactory_tid, NULL);
 }
 
-static const dispex_static_data_vtbl_t HTMLImageCtor_dispex_vtbl = {
+static const dispex_static_data_vtbl_t HTMLImageElementFactory_dispex_vtbl = {
     .query_interface  = HTMLImageElementFactory_query_interface,
-    .destructor       = global_ctor_destructor,
-    .traverse         = global_ctor_traverse,
-    .unlink           = global_ctor_unlink,
-    .value            = HTMLImageCtor_value,
-    .get_dispid       = legacy_ctor_get_dispid,
-    .get_name         = legacy_ctor_get_name,
-    .invoke           = legacy_ctor_invoke,
-    .delete           = legacy_ctor_delete
+    .destructor       = HTMLImageElementFactory_destructor,
+    .traverse         = HTMLImageElementFactory_traverse,
+    .unlink           = HTMLImageElementFactory_unlink,
+    .value            = HTMLImageElementFactory_value,
 };
 
-dispex_static_data_t HTMLImageCtor_dispex = {
-    "HTMLImageElement",
-    &HTMLImageCtor_dispex_vtbl,
-    PROTO_ID_NULL,
-    IHTMLImageElementFactory_tid,
-    HTMLImageElementFactory_iface_tids
+static dispex_static_data_t HTMLImageElementFactory_dispex = {
+    .name           = "Function",
+    .constructor_id = PROT_HTMLImageElement,
+    .vtbl           = &HTMLImageElementFactory_dispex_vtbl,
+    .disp_tid       = IHTMLImageElementFactory_tid,
+    .init_info      = HTMLImageElementFactory_init_dispex_info,
 };
+
+HRESULT HTMLImageElementFactory_Create(HTMLInnerWindow *window, DispatchEx **ret_val)
+{
+    HTMLImageElementFactory *ret;
+
+    ret = malloc(sizeof(HTMLImageElementFactory));
+    if(!ret)
+        return E_OUTOFMEMORY;
+
+    ret->IHTMLImageElementFactory_iface.lpVtbl = &HTMLImageElementFactoryVtbl;
+    ret->window = window;
+    IHTMLWindow2_AddRef(&window->base.IHTMLWindow2_iface);
+
+    init_dispatch(&ret->dispex, &HTMLImageElementFactory_dispex, window,
+                  dispex_compat_mode(&window->event_target.dispex));
+
+    *ret_val = &ret->dispex;
+    return S_OK;
+}
