@@ -507,14 +507,16 @@ static inline void mark_block_tail( struct block *block, DWORD flags )
 static inline void initialize_block( struct block *block, SIZE_T old_size, SIZE_T size, DWORD flags )
 {
     char *data = (char *)(block + 1);
-    SIZE_T i;
+    SIZE_T i, aligned_size;
 
     if (size <= old_size) return;
 
     if (flags & HEAP_ZERO_MEMORY)
     {
-        valgrind_make_writable( data + old_size, size - old_size );
-        memset( data + old_size, 0, size - old_size );
+        aligned_size = ROUND_SIZE( size, sizeof(void *) - 1 );
+        valgrind_make_writable( data + old_size, aligned_size - old_size );
+        memset( data + old_size, 0, aligned_size - old_size );
+        valgrind_make_noaccess( data + size, aligned_size - size );
     }
     else if (flags & HEAP_FREE_CHECKING_ENABLED)
     {
@@ -972,6 +974,7 @@ static struct block *split_block( struct heap *heap, ULONG flags, struct block *
 
 static void *allocate_region( struct heap *heap, ULONG flags, SIZE_T *region_size, SIZE_T *commit_size )
 {
+    const SIZE_T align = 0x400 * sizeof(void*);  /* minimum alignment for virtual allocations */
     ULONG reserve_flags = MEM_RESERVE;
     void *addr = NULL;
     NTSTATUS status;
@@ -982,6 +985,9 @@ static void *allocate_region( struct heap *heap, ULONG flags, SIZE_T *region_siz
         return NULL;
     }
 
+    *region_size = ROUND_SIZE( *region_size, align - 1 );
+    *commit_size = ROUND_SIZE( *commit_size, align - 1 );
+
     if (heap_top_down_hack) reserve_flags |= MEM_TOP_DOWN;
 
     /* allocate the memory block */
@@ -1153,10 +1159,9 @@ static struct block *find_free_block( struct heap *heap, ULONG flags, SIZE_T blo
 
 static BOOL is_valid_free_block( const struct heap *heap, const struct block *block )
 {
-    const SUBHEAP *subheap;
     unsigned int i;
 
-    if ((subheap = find_subheap( heap, block, FALSE ))) return TRUE;
+    if (find_subheap( heap, block, FALSE )) return TRUE;
     for (i = 0; i < FREE_LIST_COUNT; i++) if (block == &heap->free_lists[i].block) return TRUE;
     return FALSE;
 }
@@ -1497,23 +1502,9 @@ static void heap_set_debug_flags( HANDLE handle )
 
 /***********************************************************************
  *           RtlCreateHeap   (NTDLL.@)
- *
- * Create a new Heap.
- *
- * PARAMS
- *  flags      [I] HEAP_ flags from "winnt.h"
- *  addr       [I] Desired base address
- *  totalSize  [I] Total size of the heap, or 0 for a growable heap
- *  commitSize [I] Amount of heap space to commit
- *  unknown    [I] Not yet understood
- *  definition [I] Heap definition
- *
- * RETURNS
- *  Success: A HANDLE to the newly created heap.
- *  Failure: a NULL HANDLE.
  */
 HANDLE WINAPI RtlCreateHeap( ULONG flags, void *addr, SIZE_T total_size, SIZE_T commit_size,
-                             void *unknown, RTL_HEAP_DEFINITION *definition )
+                             void *lock, RTL_HEAP_PARAMETERS *params )
 {
     struct entry *entry;
     struct heap *heap;
@@ -1521,8 +1512,8 @@ HANDLE WINAPI RtlCreateHeap( ULONG flags, void *addr, SIZE_T total_size, SIZE_T
     SUBHEAP *subheap;
     unsigned int i;
 
-    TRACE( "flags %#lx, addr %p, total_size %#Ix, commit_size %#Ix, unknown %p, definition %p\n",
-           flags, addr, total_size, commit_size, unknown, definition );
+    TRACE( "flags %#lx, addr %p, total_size %#Ix, commit_size %#Ix, lock %p, params %p\n",
+           flags, addr, total_size, commit_size, lock, params );
 
     if (heap_zero_hack)
         flags |= HEAP_ZERO_MEMORY;
@@ -1572,7 +1563,7 @@ HANDLE WINAPI RtlCreateHeap( ULONG flags, void *addr, SIZE_T total_size, SIZE_T
     }
     else
     {
-        RtlInitializeCriticalSection( &heap->cs );
+        RtlInitializeCriticalSectionEx( &heap->cs, 0, RTL_CRITICAL_SECTION_FLAG_FORCE_DEBUG_INFO );
         heap->cs.DebugInfo->Spare[0] = (DWORD_PTR)(__FILE__ ": heap.cs");
     }
 
