@@ -50,6 +50,8 @@ static BOOL is_wow64(void)
     return !!NtCurrentTeb()->WowTebOffset;
 }
 
+static UINT64 call_gl_debug_message_callback;
+
 static pthread_mutex_t wgl_lock = PTHREAD_MUTEX_INITIALIZER;
 
 /* handle management */
@@ -66,8 +68,8 @@ enum wgl_handle_type
 struct opengl_context
 {
     DWORD tid;                   /* thread that the context is current in */
-    void (CALLBACK *debug_callback)( GLenum, GLenum, GLuint, GLenum, GLsizei, const GLchar *, const void * ); /* debug callback */
-    const void *debug_user;      /* debug user parameter */
+    UINT64 debug_callback;       /* client pointer */
+    UINT64 debug_user;           /* client pointer */
     GLubyte *extensions;         /* extension string */
     GLuint *disabled_exts;       /* indices of disabled extensions */
     struct wgl_context *drv_ctx; /* driver context */
@@ -318,13 +320,13 @@ static HKEY open_hkcu_key( const char *name )
             return 0;
 
         sid = ((TOKEN_USER *)sid_data)->User.Sid;
-        len = sprintf( buffer, "\\Registry\\User\\S-%u-%u", sid->Revision,
-                       (int)MAKELONG( MAKEWORD( sid->IdentifierAuthority.Value[5],
-                                                sid->IdentifierAuthority.Value[4] ),
-                                      MAKEWORD( sid->IdentifierAuthority.Value[3],
-                                                sid->IdentifierAuthority.Value[2] )));
+        len = snprintf( buffer, sizeof(buffer), "\\Registry\\User\\S-%u-%u", sid->Revision,
+                        (int)MAKELONG( MAKEWORD( sid->IdentifierAuthority.Value[5],
+                                                 sid->IdentifierAuthority.Value[4] ),
+                                       MAKEWORD( sid->IdentifierAuthority.Value[3],
+                                                 sid->IdentifierAuthority.Value[2] )));
         for (i = 0; i < sid->SubAuthorityCount; i++)
-            len += sprintf( buffer + len, "-%u", (int)sid->SubAuthority[i] );
+            len += snprintf( buffer + len, sizeof(buffer) - len, "-%u", (int)sid->SubAuthority[i] );
 
         ascii_to_unicode( bufferW, buffer, len );
         hkcu = reg_open_key( NULL, bufferW, len * sizeof(WCHAR) );
@@ -491,9 +493,9 @@ static const GLubyte *wrap_glGetString( TEB *teb, GLenum name )
             /* 4.4 depends on ARB_buffer_storage, which we don't support on wow64. */
             if (major > 4 || (major == 4 && minor >= 4))
             {
-                char *str = malloc( 3 + strlen( rest ) + 1 );
+                char *str = NULL;
 
-                sprintf( str, "4.3%s", rest );
+                asprintf( &str, "4.3%s", rest );
                 if (InterlockedCompareExchangePointer( (void **)&ptr->u.context->version_string, str, NULL ))
                     free( str );
                 return ptr->u.context->version_string;
@@ -536,7 +538,7 @@ static char *build_extension_list( TEB *teb )
         capacity = max( capacity, len + strlen( extension ) + 2 );
         if (!(tmp = realloc( available_extensions, capacity ))) break;
         available_extensions = tmp;
-        len += sprintf( available_extensions + len, "%s ", extension );
+        len += snprintf( available_extensions + len, capacity - len, "%s ", extension );
     }
     if (len) available_extensions[len - 1] = 0;
 
@@ -622,6 +624,7 @@ static PROC wrap_wglGetProcAddress( TEB *teb, LPCSTR name )
             {
                 { "glCopyTexSubImage3DEXT", "glCopyTexSubImage3D" },     /* needed by RuneScape */
                 { "glVertexAttribDivisor", "glVertexAttribDivisorARB"},  /* needed by Caffeine */
+                { "glCompressedTexImage2DARB", "glCompressedTexImage2D" }, /* needed by Grim Fandango Remastered */
             };
 
             for (i = 0; i < ARRAY_SIZE(alternatives); i++)
@@ -913,21 +916,15 @@ static BOOL wrap_wglSetPbufferAttribARB( HPBUFFERARB handle, const int *attribs
 }
 
 static void gl_debug_message_callback( GLenum source, GLenum type, GLuint id, GLenum severity,
-                                       GLsizei length, const GLchar *message, const void *userParam )
+                                       GLsizei length, const GLchar *message, const void *user )
 {
-    struct wine_gl_debug_message_params params =
-    {
-        .source = source,
-        .type = type,
-        .id = id,
-        .severity = severity,
-        .length = length,
-        .message = message,
-    };
+    struct gl_debug_message_callback_params *params;
     void *ret_ptr;
     ULONG ret_len;
-    struct wgl_handle *ptr = (struct wgl_handle *)userParam;
+    struct wgl_handle *ptr = (struct wgl_handle *)user;
+    UINT len = strlen( message ) + 1, size;
 
+    if (!ptr->u.context->debug_callback) return;
     if (!NtCurrentTeb())
     {
         fprintf( stderr, "msg:gl_debug_message_callback called from native thread, severity %#x, message \"%.*s\".\n",
@@ -935,46 +932,55 @@ static void gl_debug_message_callback( GLenum source, GLenum type, GLuint id, GL
         return;
     }
 
-    if (!(params.user_callback = ptr->u.context->debug_callback)) return;
-    params.user_data = ptr->u.context->debug_user;
+    size = offsetof(struct gl_debug_message_callback_params, message[len] );
+    if (!(params = malloc( size ))) return;
+    params->dispatch.callback = call_gl_debug_message_callback;
+    params->debug_callback = ptr->u.context->debug_callback;
+    params->debug_user = ptr->u.context->debug_user;
+    params->source = source;
+    params->type = type;
+    params->id = id;
+    params->severity = severity;
+    params->length = length;
+    memcpy( params->message, message, len );
 
-    KeUserModeCallback( NtUserCallOpenGLDebugMessageCallback, &params, sizeof(params),
-                        &ret_ptr, &ret_len );
+    KeUserDispatchCallback( &params->dispatch, size, &ret_ptr, &ret_len );
+    free( params );
 }
 
-static void wrap_glDebugMessageCallback( TEB *teb, GLDEBUGPROC callback, const void *userParam )
+static void wrap_glDebugMessageCallback( TEB *teb, GLDEBUGPROC callback, const void *user )
 {
     struct wgl_handle *ptr = get_current_context_ptr( teb );
     const struct opengl_funcs *funcs = teb->glTable;
 
     if (!funcs->ext.p_glDebugMessageCallback) return;
 
-    ptr->u.context->debug_callback = callback;
-    ptr->u.context->debug_user     = userParam;
+    ptr->u.context->debug_callback = (UINT_PTR)callback;
+    ptr->u.context->debug_user     = (UINT_PTR)user;
     funcs->ext.p_glDebugMessageCallback( gl_debug_message_callback, ptr );
 }
 
-static void wrap_glDebugMessageCallbackAMD( TEB *teb, GLDEBUGPROCAMD callback, void *userParam )
+static void wrap_glDebugMessageCallbackAMD( TEB *teb, GLDEBUGPROCAMD callback, void *user )
 {
     struct wgl_handle *ptr = get_current_context_ptr( teb );
     const struct opengl_funcs *funcs = teb->glTable;
 
     if (!funcs->ext.p_glDebugMessageCallbackAMD) return;
 
-    ptr->u.context->debug_callback = callback;
-    ptr->u.context->debug_user     = userParam;
+    ptr->u.context->debug_callback = (UINT_PTR)callback;
+    ptr->u.context->debug_user     = (UINT_PTR)user;
     funcs->ext.p_glDebugMessageCallbackAMD( gl_debug_message_callback, ptr );
 }
 
-static void wrap_glDebugMessageCallbackARB( TEB *teb, GLDEBUGPROCARB callback, const void *userParam )
+static void wrap_glDebugMessageCallbackARB( TEB *teb, GLDEBUGPROCARB callback, const void *user )
 {
     struct wgl_handle *ptr = get_current_context_ptr( teb );
     const struct opengl_funcs *funcs = teb->glTable;
 
     if (!funcs->ext.p_glDebugMessageCallbackARB) return;
 
-    ptr->u.context->debug_callback = callback;
-    ptr->u.context->debug_user     = userParam;
+    ptr->u.context->debug_callback = (UINT_PTR)callback;
+    ptr->u.context->debug_user     = (UINT_PTR)user;
     funcs->ext.p_glDebugMessageCallbackARB( gl_debug_message_callback, ptr );
 }
 
@@ -1162,6 +1168,13 @@ NTSTATUS ext_wglSetPbufferAttribARB( void *args )
     return STATUS_SUCCESS;
 }
 
+NTSTATUS process_attach( void *args )
+{
+    struct process_attach_params *params = args;
+    call_gl_debug_message_callback = params->call_gl_debug_message_callback;
+    return STATUS_SUCCESS;
+}
+
 NTSTATUS thread_attach( void *args )
 {
     TEB *teb = args;
@@ -1174,6 +1187,16 @@ NTSTATUS process_detach( void *args )
     return STATUS_SUCCESS;
 }
 
+NTSTATUS get_pixel_formats( void *args )
+{
+    struct get_pixel_formats_params *params = args;
+    const struct opengl_funcs *funcs = get_dc_funcs( params->hdc );
+    if (!funcs || !funcs->wgl.p_get_pixel_formats) return STATUS_NOT_IMPLEMENTED;
+    funcs->wgl.p_get_pixel_formats( params->formats, params->max_formats,
+                                    &params->num_formats, &params->num_onscreen_formats );
+    return STATUS_SUCCESS;
+}
+
 #ifdef _WIN64
 
 typedef ULONG PTR32;
@@ -1872,10 +1895,15 @@ static void unmap_named_buffer( TEB *teb, GLint buffer )
 static NTSTATUS wow64_map_buffer( TEB *teb, GLint buffer, GLenum target, void *ptr, SIZE_T size,
                                   GLbitfield access, PTR32 *ret )
 {
+    static unsigned int once;
+
     if (*ret)  /* wow64 pointer provided, map buffer to it */
     {
         if (!(access & (GL_MAP_INVALIDATE_RANGE_BIT | GL_MAP_INVALIDATE_BUFFER_BIT)))
         {
+            if (!once++)
+                FIXME( "Doing a copy of a mapped buffer (expect performance issues)\n" );
+
             TRACE( "Copying %#zx from buffer at %p to wow64 buffer %p\n", size, ptr, UlongToPtr(*ret) );
             memcpy( UlongToPtr(*ret), ptr, size );
         }
@@ -2251,4 +2279,29 @@ NTSTATUS wow64_process_detach( void *args )
     return STATUS_SUCCESS;
 }
 
+NTSTATUS wow64_get_pixel_formats( void *args )
+{
+    struct
+    {
+        PTR32 teb;
+        PTR32 hdc;
+        PTR32 formats;
+        UINT max_formats;
+        UINT num_formats;
+        UINT num_onscreen_formats;
+    } *params32 = args;
+    struct get_pixel_formats_params params =
+    {
+        .teb = get_teb64(params32->teb),
+        .hdc = ULongToPtr(params32->hdc),
+        .formats = ULongToPtr(params32->formats),
+        .max_formats = params32->max_formats,
+    };
+    NTSTATUS status;
+    status = get_pixel_formats( &params );
+    params32->num_formats = params.num_formats;
+    params32->num_onscreen_formats = params.num_onscreen_formats;
+    return status;
+}
+
 #endif
