@@ -925,7 +925,7 @@ HRESULT mem_allocator_create(IUnknown *lpUnkOuter, IUnknown **out)
     if (!(pMemAlloc = CoTaskMemAlloc(sizeof(*pMemAlloc))))
         return E_OUTOFMEMORY;
 
-    InitializeCriticalSection(&pMemAlloc->csState);
+    InitializeCriticalSectionEx(&pMemAlloc->csState, 0, RTL_CRITICAL_SECTION_FLAG_FORCE_DEBUG_INFO);
     pMemAlloc->csState.DebugInfo->Spare[0] = (DWORD_PTR)(__FILE__ ": StdMemAllocator.csState");
 
     pMemAlloc->pMemory = NULL;
