@@ -31,6 +31,8 @@
 #include "dmoreg.h"
 #include "gst_guids.h"
 #include "wmcodecdsp.h"
+#include "mferror.h"
+#include "mfapi.h"
 
 WINE_DEFAULT_DEBUG_CHANNEL(quartz);
 WINE_DECLARE_DEBUG_CHANNEL(mfplat);
@@ -39,8 +41,6 @@ WINE_DECLARE_DEBUG_CHANNEL(wmvcore);
 DEFINE_GUID(GUID_NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
 DEFINE_GUID(MEDIASUBTYPE_VC1S,MAKEFOURCC('V','C','1','S'),0x0000,0x0010,0x80,0x00,0x00,0xaa,0x00,0x38,0x9b,0x71);
 
-static const GUID MEDIASUBTYPE_MP3 = {0x00000055, 0x0000, 0x0010, {0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71}};
-
 bool array_reserve(void **elements, size_t *capacity, size_t count, size_t size)
 {
     unsigned int new_capacity, max_capacity;
@@ -68,18 +68,112 @@ bool array_reserve(void **elements, size_t *capacity, size_t count, size_t size)
     return TRUE;
 }
 
-wg_parser_t wg_parser_create(enum wg_parser_type type, bool output_compressed, bool use_opengl)
+static HRESULT video_format_from_media_type(IMFMediaType *media_type, MFVIDEOFORMAT **format, UINT32 *format_size)
+{
+    GUID subtype;
+    HRESULT hr;
+
+    if (FAILED(hr = IMFMediaType_GetGUID(media_type, &MF_MT_SUBTYPE, &subtype)))
+        return hr;
+    if (FAILED(hr = MFCreateMFVideoFormatFromMFMediaType(media_type, format, format_size)))
+        return hr;
+
+    /* fixup MPEG video formats here, so we can consistently use MFVIDEOFORMAT internally */
+    if (IsEqualGUID(&subtype, &MEDIASUBTYPE_MPEG1Payload)
+            || IsEqualGUID(&subtype, &MEDIASUBTYPE_MPEG1Packet)
+            || IsEqualGUID(&subtype, &MEDIASUBTYPE_MPEG2_VIDEO))
+    {
+        struct mpeg_video_format *mpeg;
+        UINT32 mpeg_size, len;
+
+        if (FAILED(IMFMediaType_GetBlobSize(media_type, &MF_MT_MPEG_SEQUENCE_HEADER, &len)))
+            len = 0;
+        mpeg_size = offsetof(struct mpeg_video_format, sequence_header[len]);
+
+        if ((mpeg = CoTaskMemAlloc(mpeg_size)))
+        {
+            memset(mpeg, 0, mpeg_size);
+            mpeg->hdr = **format;
+
+            IMFMediaType_GetBlob(media_type, &MF_MT_MPEG_SEQUENCE_HEADER, mpeg->sequence_header, len, NULL);
+            IMFMediaType_GetUINT32(media_type, &MF_MT_MPEG_START_TIME_CODE, (UINT32 *)&mpeg->start_time_code);
+            IMFMediaType_GetUINT32(media_type, &MF_MT_MPEG2_PROFILE, &mpeg->profile);
+            IMFMediaType_GetUINT32(media_type, &MF_MT_MPEG2_LEVEL, &mpeg->level);
+            IMFMediaType_GetUINT32(media_type, &MF_MT_MPEG2_FLAGS, &mpeg->flags);
+
+            CoTaskMemFree(*format);
+            *format = &mpeg->hdr;
+            *format_size = mpeg_size;
+        }
+    }
+
+    return hr;
+}
+
+static HRESULT wg_media_type_from_mf(IMFMediaType *media_type, struct wg_media_type *wg_media_type)
+{
+    HRESULT hr;
+
+    if (FAILED(hr = IMFMediaType_GetMajorType(media_type, &wg_media_type->major)))
+        return hr;
+
+    if (IsEqualGUID(&wg_media_type->major, &MFMediaType_Video))
+        return video_format_from_media_type(media_type, &wg_media_type->u.video,
+                &wg_media_type->format_size);
+    if (IsEqualGUID(&wg_media_type->major, &MFMediaType_Audio))
+        return MFCreateWaveFormatExFromMFMediaType(media_type, &wg_media_type->u.audio,
+                &wg_media_type->format_size, 0);
+
+    FIXME("Unsupported major type %s\n", debugstr_guid(&wg_media_type->major));
+    return E_NOTIMPL;
+}
+
+static HRESULT media_type_from_video_format(const MFVIDEOFORMAT *format, IMFMediaType **media_type)
+{
+    HRESULT hr;
+
+    if (FAILED(hr = MFCreateVideoMediaType(format, (IMFVideoMediaType **)media_type)) || format->dwSize <= sizeof(*format))
+        return hr;
+
+    /* fixup MPEG video formats here, so we can consistently use MFVIDEOFORMAT internally */
+    if (IsEqualGUID(&format->guidFormat, &MEDIASUBTYPE_MPEG1Payload)
+            || IsEqualGUID(&format->guidFormat, &MEDIASUBTYPE_MPEG1Packet)
+            || IsEqualGUID(&format->guidFormat, &MEDIASUBTYPE_MPEG2_VIDEO))
+    {
+        struct mpeg_video_format *mpeg = (struct mpeg_video_format *)format;
+        IMFMediaType_SetBlob(*media_type, &MF_MT_MPEG_SEQUENCE_HEADER, mpeg->sequence_header, mpeg->sequence_header_count);
+        IMFMediaType_SetUINT32(*media_type, &MF_MT_MPEG_START_TIME_CODE, mpeg->start_time_code);
+        IMFMediaType_SetUINT32(*media_type, &MF_MT_MPEG2_PROFILE, mpeg->profile);
+        IMFMediaType_SetUINT32(*media_type, &MF_MT_MPEG2_LEVEL, mpeg->level);
+        IMFMediaType_SetUINT32(*media_type, &MF_MT_MPEG2_FLAGS, mpeg->flags);
+        IMFMediaType_DeleteItem(*media_type, &MF_MT_USER_DATA);
+    }
+
+    return hr;
+}
+
+static HRESULT wg_media_type_to_mf(const struct wg_media_type *wg_media_type, IMFMediaType **media_type)
+{
+    if (IsEqualGUID(&wg_media_type->major, &MFMediaType_Video))
+        return media_type_from_video_format(wg_media_type->u.video, media_type);
+    if (IsEqualGUID(&wg_media_type->major, &MFMediaType_Audio))
+        return MFCreateAudioMediaType(wg_media_type->u.audio, (IMFAudioMediaType **)media_type);
+
+    FIXME("Unsupported major type %s\n", debugstr_guid(&wg_media_type->major));
+    return E_NOTIMPL;
+}
+
+wg_parser_t wg_parser_create(bool output_compressed, bool use_opengl)
 {
     struct wg_parser_create_params params =
     {
-        .type = type,
         .output_compressed = output_compressed,
         .use_opengl = use_opengl,
         .err_on = ERR_ON(quartz),
         .warn_on = WARN_ON(quartz),
     };
 
-    TRACE("type %#x, use_opengl %u.\n", type, use_opengl);
+    TRACE("output_compressed %d.\n", output_compressed);
 
     if (WINE_UNIX_CALL(unix_wg_parser_create, &params))
         return 0;
@@ -176,9 +270,9 @@ wg_parser_stream_t wg_parser_get_stream(wg_parser_t parser, uint32_t index)
     return params.stream;
 }
 
-void wg_parser_stream_get_preferred_format(wg_parser_stream_t stream, struct wg_format *format)
+void wg_parser_stream_get_current_format(wg_parser_stream_t stream, struct wg_format *format)
 {
-    struct wg_parser_stream_get_preferred_format_params params =
+    struct wg_parser_stream_get_current_format_params params =
     {
         .stream = stream,
         .format = format,
@@ -186,7 +280,7 @@ void wg_parser_stream_get_preferred_format(wg_parser_stream_t stream, struct wg_
 
     TRACE("stream %#I64x, format %p.\n", stream, format);
 
-    WINE_UNIX_CALL(unix_wg_parser_stream_get_preferred_format, &params);
+    WINE_UNIX_CALL(unix_wg_parser_stream_get_current_format, &params);
 }
 
 void wg_parser_stream_get_codec_format(wg_parser_stream_t stream, struct wg_format *format)
@@ -202,14 +296,12 @@ void wg_parser_stream_get_codec_format(wg_parser_stream_t stream, struct wg_form
     WINE_UNIX_CALL(unix_wg_parser_stream_get_codec_format, &params);
 }
 
-void wg_parser_stream_enable(wg_parser_stream_t stream, const struct wg_format *format,
-        uint32_t flags)
+void wg_parser_stream_enable(wg_parser_stream_t stream, const struct wg_format *format)
 {
     struct wg_parser_stream_enable_params params =
     {
         .stream = stream,
         .format = format,
-        .flags = flags,
     };
 
     TRACE("stream %#I64x, format %p.\n", stream, format);
@@ -342,23 +434,59 @@ void wg_parser_stream_seek(wg_parser_stream_t stream, double rate,
     WINE_UNIX_CALL(unix_wg_parser_stream_seek, &params);
 }
 
-wg_transform_t wg_transform_create(const struct wg_format *input_format,
-        const struct wg_format *output_format, const struct wg_transform_attrs *attrs)
+HRESULT wg_transform_create_mf(IMFMediaType *input_type, IMFMediaType *output_type,
+        const struct wg_transform_attrs *attrs, wg_transform_t *transform)
 {
     struct wg_transform_create_params params =
     {
-        .input_format = input_format,
-        .output_format = output_format,
-        .attrs = attrs,
+        .attrs = *attrs,
     };
+    NTSTATUS status;
+    HRESULT hr;
+
+    TRACE("input_type %p, output_type %p.\n", input_type, output_type);
+
+    if (FAILED(hr = wg_media_type_from_mf(input_type, &params.input_type)))
+        return hr;
+    if (FAILED(hr = wg_media_type_from_mf(output_type, &params.output_type)))
+    {
+        CoTaskMemFree(params.input_type.u.format);
+        return hr;
+    }
+
+    if ((status = WINE_UNIX_CALL(unix_wg_transform_create, &params)))
+    {
+        WARN("Failed to create transform, status %#lx\n", status);
+        hr = HRESULT_FROM_NT(status);
+    }
+
+    CoTaskMemFree(params.output_type.u.format);
+    CoTaskMemFree(params.input_type.u.format);
+    *transform = params.transform;
+    return hr;
+}
+
+HRESULT wg_transform_create_quartz(const AM_MEDIA_TYPE *input_format, const AM_MEDIA_TYPE *output_format,
+        const struct wg_transform_attrs *attrs, wg_transform_t *transform)
+{
+    IMFMediaType *input_type, *output_type;
+    HRESULT hr;
 
     TRACE("input_format %p, output_format %p.\n", input_format, output_format);
 
-    if (WINE_UNIX_CALL(unix_wg_transform_create, &params))
+    /* through IMFMediaType to normalize representation to MFVIDEOFORMAT / WAVEFORMATEX */
+    if (FAILED(hr = MFCreateMediaTypeFromRepresentation(AM_MEDIA_TYPE_REPRESENTATION, (void *)input_format, &input_type)))
         return 0;
+    if (FAILED(hr = MFCreateMediaTypeFromRepresentation(AM_MEDIA_TYPE_REPRESENTATION, (void *)output_format, &output_type)))
+    {
+        IMFMediaType_Release(input_type);
+        return 0;
+    }
 
-    TRACE("Returning transform %#I64x.\n", params.transform);
-    return params.transform;
+    hr = wg_transform_create_mf(input_type, output_type, attrs, transform);
+    IMFMediaType_Release(output_type);
+    IMFMediaType_Release(input_type);
+    return hr;
 }
 
 void wg_transform_destroy(wg_transform_t transform)
@@ -385,18 +513,16 @@ HRESULT wg_transform_push_data(wg_transform_t transform, struct wg_sample *sampl
     return params.result;
 }
 
-HRESULT wg_transform_read_data(wg_transform_t transform, struct wg_sample *sample,
-        struct wg_format *format)
+HRESULT wg_transform_read_data(wg_transform_t transform, struct wg_sample *sample)
 {
     struct wg_transform_read_data_params params =
     {
         .transform = transform,
         .sample = sample,
-        .format = format,
     };
     NTSTATUS status;
 
-    TRACE("transform %#I64x, sample %p, format %p.\n", transform, sample, format);
+    TRACE("transform %#I64x, sample %p.\n", transform, sample);
 
     if ((status = WINE_UNIX_CALL(unix_wg_transform_read_data, &params)))
         return HRESULT_FROM_NT(status);
@@ -420,17 +546,61 @@ bool wg_transform_get_status(wg_transform_t transform, bool *accepts_input)
     return true;
 }
 
-bool wg_transform_set_output_format(wg_transform_t transform, struct wg_format *format)
+HRESULT wg_transform_get_output_type(wg_transform_t transform, IMFMediaType **media_type)
 {
-    struct wg_transform_set_output_format_params params =
+    struct wg_transform_get_output_type_params params =
     {
         .transform = transform,
-        .format = format,
     };
+    NTSTATUS status;
+    HRESULT hr;
 
-    TRACE("transform %#I64x, format %p.\n", transform, format);
+    TRACE("transform %#I64x, media_type %p.\n", transform, media_type);
 
-    return !WINE_UNIX_CALL(unix_wg_transform_set_output_format, &params);
+    if ((status = WINE_UNIX_CALL(unix_wg_transform_get_output_type, &params))
+            && status == STATUS_BUFFER_TOO_SMALL)
+    {
+        if (!(params.media_type.u.format = CoTaskMemAlloc(params.media_type.format_size)))
+            return ERROR_OUTOFMEMORY;
+        status = WINE_UNIX_CALL(unix_wg_transform_get_output_type, &params);
+    }
+
+    if (status)
+    {
+        CoTaskMemFree(params.media_type.u.format);
+        WARN("Failed to get output media type, status %#lx\n", status);
+        return HRESULT_FROM_NT(status);
+    }
+
+    hr = wg_media_type_to_mf(&params.media_type, media_type);
+    CoTaskMemFree(params.media_type.u.format);
+    return hr;
+}
+
+HRESULT wg_transform_set_output_type(wg_transform_t transform, IMFMediaType *media_type)
+{
+    struct wg_transform_set_output_type_params params =
+    {
+        .transform = transform,
+    };
+    NTSTATUS status;
+    HRESULT hr;
+
+    TRACE("transform %#I64x, media_type %p.\n", transform, media_type);
+
+    if (FAILED(hr = wg_media_type_from_mf(media_type, &params.media_type)))
+    {
+        WARN("Failed to initialize media type, hr %#lx\n", hr);
+        return hr;
+    }
+    if ((status = WINE_UNIX_CALL(unix_wg_transform_set_output_type, &params)))
+    {
+       WARN("Failed to set transform output type, status %#lx\n", status);
+       hr = HRESULT_FROM_NT(status);
+    }
+
+    CoTaskMemFree(params.media_type.u.format);
+    return hr;
 }
 
 HRESULT wg_transform_drain(wg_transform_t transform)
@@ -606,6 +776,63 @@ HRESULT wg_muxer_finalize(wg_muxer_t muxer)
     return S_OK;
 }
 
+static HRESULT check_transform_support(const struct wg_media_type *input, const struct wg_media_type *output)
+{
+    IMFMediaType *input_type, *output_type;
+    struct wg_transform_attrs attrs = {0};
+    wg_transform_t transform;
+    HRESULT hr;
+
+    if (FAILED(hr = wg_media_type_to_mf(input, &input_type)))
+        return hr;
+    if (FAILED(hr = wg_media_type_to_mf(output, &output_type)))
+    {
+        IMFMediaType_Release(input_type);
+        return hr;
+    }
+
+    if (SUCCEEDED(hr = wg_transform_create_mf(input_type, output_type, &attrs, &transform)))
+        wg_transform_destroy(transform);
+
+    IMFMediaType_Release(output_type);
+    IMFMediaType_Release(input_type);
+    return hr;
+}
+
+HRESULT check_audio_transform_support(const WAVEFORMATEX *input, const WAVEFORMATEX *output)
+{
+    const struct wg_media_type input_type =
+    {
+        .major = MFMediaType_Audio,
+        .format_size = sizeof(*input) + input->cbSize,
+        .u.audio = (WAVEFORMATEX *)input,
+    };
+    const struct wg_media_type output_type =
+    {
+        .major = MFMediaType_Audio,
+        .format_size = sizeof(*output) + output->cbSize,
+        .u.audio = (WAVEFORMATEX *)output,
+    };
+    return check_transform_support(&input_type, &output_type);
+}
+
+HRESULT check_video_transform_support(const MFVIDEOFORMAT *input, const MFVIDEOFORMAT *output)
+{
+    const struct wg_media_type input_type =
+    {
+        .major = MFMediaType_Video,
+        .format_size = input->dwSize,
+        .u.video = (MFVIDEOFORMAT *)input,
+    };
+    const struct wg_media_type output_type =
+    {
+        .major = MFMediaType_Video,
+        .format_size = output->dwSize,
+        .u.video = (MFVIDEOFORMAT *)output,
+    };
+    return check_transform_support(&input_type, &output_type);
+}
+
 #define ALIGN(n, alignment) (((n) + (alignment) - 1) & ~((alignment) - 1))
 
 unsigned int wg_format_get_stride(const struct wg_format *format)
@@ -619,6 +846,7 @@ unsigned int wg_format_get_stride(const struct wg_format *format)
 
         case WG_VIDEO_FORMAT_BGRA:
         case WG_VIDEO_FORMAT_BGRx:
+        case WG_VIDEO_FORMAT_RGBA:
             return width * 4;
 
         case WG_VIDEO_FORMAT_BGR:
@@ -634,10 +862,13 @@ unsigned int wg_format_get_stride(const struct wg_format *format)
             return ALIGN(width * 2, 4);
 
         case WG_VIDEO_FORMAT_I420:
-        case WG_VIDEO_FORMAT_NV12:
         case WG_VIDEO_FORMAT_YV12:
             return ALIGN(width, 4); /* Y plane */
 
+        /* NV12 stride in Windows has alignment 2. GStreamer output is reformatted to 2 where necessary. */
+        case WG_VIDEO_FORMAT_NV12:
+            return ALIGN(width, 2); /* Y plane */
+
         case WG_VIDEO_FORMAT_UNKNOWN:
             FIXME("Cannot calculate stride for unknown video format.\n");
     }
@@ -652,19 +883,12 @@ bool wg_video_format_is_rgb(enum wg_video_format format)
         case WG_VIDEO_FORMAT_BGRA:
         case WG_VIDEO_FORMAT_BGRx:
         case WG_VIDEO_FORMAT_BGR:
-        case WG_VIDEO_FORMAT_RGBA:
         case WG_VIDEO_FORMAT_RGB15:
         case WG_VIDEO_FORMAT_RGB16:
+        case WG_VIDEO_FORMAT_RGBA:
             return true;
 
-        case WG_VIDEO_FORMAT_AYUV:
-        case WG_VIDEO_FORMAT_I420:
-        case WG_VIDEO_FORMAT_NV12:
-        case WG_VIDEO_FORMAT_UYVY:
-        case WG_VIDEO_FORMAT_YUY2:
-        case WG_VIDEO_FORMAT_YV12:
-        case WG_VIDEO_FORMAT_YVYU:
-        case WG_VIDEO_FORMAT_UNKNOWN:
+        default:
             break;
     }
 
@@ -769,6 +993,18 @@ static struct class_factory mpeg4_sink_class_factory_cf = {{&class_factory_vtbl}
 
 HRESULT WINAPI DllGetClassObject(REFCLSID clsid, REFIID iid, void **out)
 {
+    static const GUID CLSID_wg_avi_splitter = {0x272bfbfb,0x50d0,0x4078,{0xb6,0x00,0x1e,0x95,0x9c,0x30,0x13,0x37}};
+    static const GUID CLSID_wg_color_converter = {0xf47e2da5,0xe370,0x47b7,{0x90,0x3a,0x07,0x8d,0xdd,0x45,0xa5,0xcc}};
+    static const GUID CLSID_wg_mp3_sink_factory = {0x1f302877,0xaaab,0x40a3,{0xb9,0xe0,0x9f,0x48,0xda,0xf3,0x5b,0xc8}};
+    static const GUID CLSID_wg_mpeg4_sink_factory = {0x5d5407d9,0xc6ca,0x4770,{0xa7,0xcc,0x27,0xc0,0xcb,0x8a,0x76,0x27}};
+    static const GUID CLSID_wg_mpeg_audio_decoder = {0xc9f285f8,0x4380,0x4121,{0x97,0x1f,0x49,0xa9,0x53,0x16,0xc2,0x7b}};
+    static const GUID CLSID_wg_mpeg_video_decoder = {0x5ed2e5f6,0xbf3e,0x4180,{0x83,0xa4,0x48,0x47,0xcc,0x5b,0x4e,0xa3}};
+    static const GUID CLSID_wg_resampler = {0x92f35e78,0x15a5,0x486b,{0x88,0x8e,0x57,0x5f,0x99,0x65,0x1c,0xe2}};
+    static const GUID CLSID_wg_wma_decoder = {0x5b4d4e54,0x0620,0x4cf9,{0x94,0xae,0x78,0x23,0x96,0x5c,0x28,0xb6}};
+    static const GUID CLSID_wg_wmv_decoder = {0x62ee5ddb,0x4f52,0x48e2,{0x89,0x28,0x78,0x7b,0x02,0x53,0xa0,0xbc}};
+    static const GUID CLSID_wg_mp3_decoder = {0x84cd8e3e,0xb221,0x434a,{0x88,0x82,0x9d,0x6c,0x8d,0xf4,0x90,0xe1}};
+    static const GUID CLSID_wg_mpeg1_splitter = {0xa8edbf98,0x2442,0x42c5,{0x85,0xa1,0xab,0x05,0xa5,0x80,0xdf,0x53}};
+    static const GUID CLSID_wg_wave_parser = {0x3f839ec7,0x5ea6,0x49e1,{0x80,0xc2,0x1e,0xa3,0x00,0xf8,0xb0,0xe0}};
     struct class_factory *factory;
     HRESULT hr;
 
@@ -780,31 +1016,31 @@ HRESULT WINAPI DllGetClassObject(REFCLSID clsid, REFIID iid, void **out)
     if (SUCCEEDED(hr = mfplat_get_class_object(clsid, iid, out)))
         return hr;
 
-    if (IsEqualGUID(clsid, &CLSID_AviSplitter))
+    if (IsEqualGUID(clsid, &CLSID_wg_avi_splitter))
         factory = &avi_splitter_cf;
     else if (IsEqualGUID(clsid, &CLSID_decodebin_parser))
         factory = &decodebin_parser_cf;
-    else if (IsEqualGUID(clsid, &CLSID_CMpegAudioCodec))
+    else if (IsEqualGUID(clsid, &CLSID_wg_mpeg_audio_decoder))
         factory = &mpeg_audio_codec_cf;
-    else if (IsEqualGUID(clsid, &CLSID_CMpegVideoCodec))
+    else if (IsEqualGUID(clsid, &CLSID_wg_mpeg_video_decoder))
         factory = &mpeg_video_codec_cf;
-    else if (IsEqualGUID(clsid, &CLSID_mpeg_layer3_decoder))
+    else if (IsEqualGUID(clsid, &CLSID_wg_mp3_decoder))
         factory = &mpeg_layer3_decoder_cf;
-    else if (IsEqualGUID(clsid, &CLSID_MPEG1Splitter))
+    else if (IsEqualGUID(clsid, &CLSID_wg_mpeg1_splitter))
         factory = &mpeg_splitter_cf;
-    else if (IsEqualGUID(clsid, &CLSID_WAVEParser))
+    else if (IsEqualGUID(clsid, &CLSID_wg_wave_parser))
         factory = &wave_parser_cf;
-    else if (IsEqualGUID(clsid, &CLSID_WMADecMediaObject))
+    else if (IsEqualGUID(clsid, &CLSID_wg_wma_decoder))
         factory = &wma_decoder_cf;
-    else if (IsEqualGUID(clsid, &CLSID_WMVDecoderMFT))
+    else if (IsEqualGUID(clsid, &CLSID_wg_wmv_decoder))
         factory = &wmv_decoder_cf;
-    else if (IsEqualGUID(clsid, &CLSID_CResamplerMediaObject))
+    else if (IsEqualGUID(clsid, &CLSID_wg_resampler))
         factory = &resampler_cf;
-    else if (IsEqualGUID(clsid, &CLSID_CColorConvertDMO))
+    else if (IsEqualGUID(clsid, &CLSID_wg_color_converter))
         factory = &color_convert_cf;
-    else if (IsEqualGUID(clsid, &CLSID_MFMP3SinkClassFactory))
+    else if (IsEqualGUID(clsid, &CLSID_wg_mp3_sink_factory))
         factory = &mp3_sink_class_factory_cf;
-    else if (IsEqualGUID(clsid, &CLSID_MFMPEG4SinkClassFactory))
+    else if (IsEqualGUID(clsid, &CLSID_wg_mpeg4_sink_factory))
         factory = &mpeg4_sink_class_factory_cf;
     else
     {
@@ -851,199 +1087,6 @@ static const REGPINTYPES reg_audio_mt = {&MEDIATYPE_Audio, &GUID_NULL};
 static const REGPINTYPES reg_stream_mt = {&MEDIATYPE_Stream, &GUID_NULL};
 static const REGPINTYPES reg_video_mt = {&MEDIATYPE_Video, &GUID_NULL};
 
-static const REGPINTYPES reg_avi_splitter_sink_mt = {&MEDIATYPE_Stream, &MEDIASUBTYPE_Avi};
-
-static const REGFILTERPINS2 reg_avi_splitter_pins[2] =
-{
-    {
-        .nMediaTypes = 1,
-        .lpMediaType = &reg_avi_splitter_sink_mt,
-    },
-    {
-        .dwFlags = REG_PINFLAG_B_OUTPUT,
-        .nMediaTypes = 1,
-        .lpMediaType = &reg_video_mt,
-    },
-};
-
-static const REGFILTER2 reg_avi_splitter =
-{
-    .dwVersion = 2,
-    .dwMerit = MERIT_NORMAL,
-    .u.s2.cPins2 = 2,
-    .u.s2.rgPins2 = reg_avi_splitter_pins,
-};
-
-static const REGPINTYPES reg_mpeg_audio_codec_sink_mts[3] =
-{
-    {&MEDIATYPE_Audio, &MEDIASUBTYPE_MPEG1Packet},
-    {&MEDIATYPE_Audio, &MEDIASUBTYPE_MPEG1Payload},
-    {&MEDIATYPE_Audio, &MEDIASUBTYPE_MPEG1AudioPayload},
-};
-
-static const REGPINTYPES reg_mpeg_audio_codec_source_mts[1] =
-{
-    {&MEDIATYPE_Audio, &MEDIASUBTYPE_PCM},
-};
-
-static const REGFILTERPINS2 reg_mpeg_audio_codec_pins[2] =
-{
-    {
-        .nMediaTypes = 3,
-        .lpMediaType = reg_mpeg_audio_codec_sink_mts,
-    },
-    {
-        .dwFlags = REG_PINFLAG_B_OUTPUT,
-        .nMediaTypes = 1,
-        .lpMediaType = reg_mpeg_audio_codec_source_mts,
-    },
-};
-
-static const REGFILTER2 reg_mpeg_audio_codec =
-{
-    .dwVersion = 2,
-    .dwMerit = 0x03680001,
-    .u.s2.cPins2 = 2,
-    .u.s2.rgPins2 = reg_mpeg_audio_codec_pins,
-};
-
-static const REGPINTYPES reg_mpeg_video_codec_sink_mts[2] =
-{
-    {&MEDIATYPE_Video, &MEDIASUBTYPE_MPEG1Packet},
-    {&MEDIATYPE_Video, &MEDIASUBTYPE_MPEG1Payload},
-};
-
-static const REGPINTYPES reg_mpeg_video_codec_source_mts[1] =
-{
-    {&MEDIATYPE_Video, &GUID_NULL},
-};
-
-static const REGFILTERPINS2 reg_mpeg_video_codec_pins[2] =
-{
-    {
-        .nMediaTypes = 2,
-        .lpMediaType = reg_mpeg_video_codec_sink_mts,
-    },
-    {
-        .dwFlags = REG_PINFLAG_B_OUTPUT,
-        .nMediaTypes = 1,
-        .lpMediaType = reg_mpeg_video_codec_source_mts,
-    },
-};
-
-static const REGFILTER2 reg_mpeg_video_codec =
-{
-    .dwVersion = 2,
-    .dwMerit = 0x40000001,
-    .u.s2.cPins2 = 2,
-    .u.s2.rgPins2 = reg_mpeg_video_codec_pins,
-};
-
-static const REGPINTYPES reg_mpeg_layer3_decoder_sink_mts[1] =
-{
-    {&MEDIATYPE_Audio, &MEDIASUBTYPE_MP3},
-};
-
-static const REGPINTYPES reg_mpeg_layer3_decoder_source_mts[1] =
-{
-    {&MEDIATYPE_Audio, &MEDIASUBTYPE_PCM},
-};
-
-static const REGFILTERPINS2 reg_mpeg_layer3_decoder_pins[2] =
-{
-    {
-        .nMediaTypes = 1,
-        .lpMediaType = reg_mpeg_layer3_decoder_sink_mts,
-    },
-    {
-        .dwFlags = REG_PINFLAG_B_OUTPUT,
-        .nMediaTypes = 1,
-        .lpMediaType = reg_mpeg_layer3_decoder_source_mts,
-    },
-};
-
-static const REGFILTER2 reg_mpeg_layer3_decoder =
-{
-    .dwVersion = 2,
-    .dwMerit = 0x00810000,
-    .u.s2.cPins2 = 2,
-    .u.s2.rgPins2 = reg_mpeg_layer3_decoder_pins,
-};
-
-static const REGPINTYPES reg_mpeg_splitter_sink_mts[4] =
-{
-    {&MEDIATYPE_Stream, &MEDIASUBTYPE_MPEG1Audio},
-    {&MEDIATYPE_Stream, &MEDIASUBTYPE_MPEG1Video},
-    {&MEDIATYPE_Stream, &MEDIASUBTYPE_MPEG1System},
-    {&MEDIATYPE_Stream, &MEDIASUBTYPE_MPEG1VideoCD},
-};
-
-static const REGPINTYPES reg_mpeg_splitter_audio_mts[2] =
-{
-    {&MEDIATYPE_Audio, &MEDIASUBTYPE_MPEG1Packet},
-    {&MEDIATYPE_Audio, &MEDIASUBTYPE_MPEG1AudioPayload},
-};
-
-static const REGPINTYPES reg_mpeg_splitter_video_mts[2] =
-{
-    {&MEDIATYPE_Video, &MEDIASUBTYPE_MPEG1Packet},
-    {&MEDIATYPE_Video, &MEDIASUBTYPE_MPEG1Payload},
-};
-
-static const REGFILTERPINS2 reg_mpeg_splitter_pins[3] =
-{
-    {
-        .nMediaTypes = 4,
-        .lpMediaType = reg_mpeg_splitter_sink_mts,
-    },
-    {
-        .dwFlags = REG_PINFLAG_B_ZERO | REG_PINFLAG_B_OUTPUT,
-        .nMediaTypes = 2,
-        .lpMediaType = reg_mpeg_splitter_audio_mts,
-    },
-    {
-        .dwFlags = REG_PINFLAG_B_ZERO | REG_PINFLAG_B_OUTPUT,
-        .nMediaTypes = 2,
-        .lpMediaType = reg_mpeg_splitter_video_mts,
-    },
-};
-
-static const REGFILTER2 reg_mpeg_splitter =
-{
-    .dwVersion = 2,
-    .dwMerit = MERIT_NORMAL,
-    .u.s2.cPins2 = 3,
-    .u.s2.rgPins2 = reg_mpeg_splitter_pins,
-};
-
-static const REGPINTYPES reg_wave_parser_sink_mts[3] =
-{
-    {&MEDIATYPE_Stream, &MEDIASUBTYPE_WAVE},
-    {&MEDIATYPE_Stream, &MEDIASUBTYPE_AU},
-    {&MEDIATYPE_Stream, &MEDIASUBTYPE_AIFF},
-};
-
-static const REGFILTERPINS2 reg_wave_parser_pins[2] =
-{
-    {
-        .nMediaTypes = 3,
-        .lpMediaType = reg_wave_parser_sink_mts,
-    },
-    {
-        .dwFlags = REG_PINFLAG_B_OUTPUT,
-        .nMediaTypes = 1,
-        .lpMediaType = &reg_audio_mt,
-    },
-};
-
-static const REGFILTER2 reg_wave_parser =
-{
-    .dwVersion = 2,
-    .dwMerit = MERIT_UNLIKELY,
-    .u.s2.cPins2 = 2,
-    .u.s2.rgPins2 = reg_wave_parser_pins,
-};
-
 static const REGFILTERPINS2 reg_decodebin_parser_pins[3] =
 {
     {
@@ -1065,98 +1108,13 @@ static const REGFILTERPINS2 reg_decodebin_parser_pins[3] =
 static const REGFILTER2 reg_decodebin_parser =
 {
     .dwVersion = 2,
-    .dwMerit = MERIT_PREFERRED,
+    .dwMerit = MERIT_NORMAL - 1,
     .u.s2.cPins2 = 3,
     .u.s2.rgPins2 = reg_decodebin_parser_pins,
 };
 
 HRESULT WINAPI DllRegisterServer(void)
 {
-    DMO_PARTIAL_MEDIATYPE audio_convert_types[2] =
-    {
-        {.type = MEDIATYPE_Audio, .subtype = MEDIASUBTYPE_PCM},
-        {.type = MEDIATYPE_Audio, .subtype = MEDIASUBTYPE_IEEE_FLOAT},
-    };
-    DMO_PARTIAL_MEDIATYPE wma_decoder_output[2] =
-    {
-        {.type = MEDIATYPE_Audio, .subtype = MEDIASUBTYPE_PCM},
-        {.type = MEDIATYPE_Audio, .subtype = MEDIASUBTYPE_IEEE_FLOAT},
-    };
-    DMO_PARTIAL_MEDIATYPE wma_decoder_input[4] =
-    {
-        {.type = MEDIATYPE_Audio, .subtype = MEDIASUBTYPE_MSAUDIO1},
-        {.type = MEDIATYPE_Audio, .subtype = MEDIASUBTYPE_WMAUDIO2},
-        {.type = MEDIATYPE_Audio, .subtype = MEDIASUBTYPE_WMAUDIO3},
-        {.type = MEDIATYPE_Audio, .subtype = MEDIASUBTYPE_WMAUDIO_LOSSLESS},
-    };
-    DMO_PARTIAL_MEDIATYPE wmv_decoder_output[11] =
-    {
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YV12},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YUY2},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_UYVY},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YVYU},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_NV11},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_NV12},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB32},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB24},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB565},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB555},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB8},
-    };
-    DMO_PARTIAL_MEDIATYPE wmv_decoder_input[8] =
-    {
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_WMV1},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_WMV2},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_WMV3},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_WMVA},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_WVC1},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_WMVP},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_WVP2},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_VC1S},
-    };
-    DMO_PARTIAL_MEDIATYPE color_convert_input[20] =
-    {
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YV12},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YUY2},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_UYVY},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_AYUV},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_NV12},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB32},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB565},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_I420},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_IYUV},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YVYU},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB24},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB555},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB8},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_V216},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_V410},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_NV11},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_Y41P},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_Y41T},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_Y42T},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YVU9},
-    };
-    DMO_PARTIAL_MEDIATYPE color_convert_output[16] =
-    {
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YV12},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YUY2},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_UYVY},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_AYUV},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_NV12},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB32},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB565},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_I420},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_IYUV},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_YVYU},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB24},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB555},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_RGB8},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_V216},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_V410},
-        {.type = MEDIATYPE_Video, .subtype = MEDIASUBTYPE_NV11},
-    };
-
     IFilterMapper2 *mapper;
     HRESULT hr;
 
@@ -1169,35 +1127,12 @@ HRESULT WINAPI DllRegisterServer(void)
             &IID_IFilterMapper2, (void **)&mapper)))
         return hr;
 
-    IFilterMapper2_RegisterFilter(mapper, &CLSID_AviSplitter, L"AVI Splitter", NULL, NULL, NULL, &reg_avi_splitter);
     IFilterMapper2_RegisterFilter(mapper, &CLSID_decodebin_parser,
             L"GStreamer splitter filter", NULL, NULL, NULL, &reg_decodebin_parser);
-    IFilterMapper2_RegisterFilter(mapper, &CLSID_CMpegAudioCodec,
-            L"MPEG Audio Decoder", NULL, NULL, NULL, &reg_mpeg_audio_codec);
-    IFilterMapper2_RegisterFilter(mapper, &CLSID_CMpegVideoCodec,
-            L"MPEG Video Decoder", NULL, NULL, NULL, &reg_mpeg_video_codec);
-    IFilterMapper2_RegisterFilter(mapper, &CLSID_mpeg_layer3_decoder,
-            L"MPEG Layer-3 Decoder", NULL, NULL, NULL, &reg_mpeg_layer3_decoder);
-    IFilterMapper2_RegisterFilter(mapper, &CLSID_MPEG1Splitter,
-            L"MPEG-I Stream Splitter", NULL, NULL, NULL, &reg_mpeg_splitter);
-    IFilterMapper2_RegisterFilter(mapper, &CLSID_WAVEParser, L"Wave Parser", NULL, NULL, NULL, &reg_wave_parser);
 
     IFilterMapper2_Release(mapper);
 
-    if (FAILED(hr = DMORegister(L"WMAudio Decoder DMO", &CLSID_WMADecMediaObject, &DMOCATEGORY_AUDIO_DECODER,
-            0, ARRAY_SIZE(wma_decoder_input), wma_decoder_input, ARRAY_SIZE(wma_decoder_output), wma_decoder_output)))
-        return hr;
-    if (FAILED(hr = DMORegister(L"WMVideo Decoder DMO", &CLSID_WMVDecoderMFT, &DMOCATEGORY_VIDEO_DECODER,
-            0, ARRAY_SIZE(wmv_decoder_input), wmv_decoder_input, ARRAY_SIZE(wmv_decoder_output), wmv_decoder_output)))
-        return hr;
-    if (FAILED(hr = DMORegister(L"Resampler DMO", &CLSID_CResamplerMediaObject, &DMOCATEGORY_AUDIO_EFFECT,
-            0, ARRAY_SIZE(audio_convert_types), audio_convert_types, ARRAY_SIZE(audio_convert_types), audio_convert_types)))
-        return hr;
-    if (FAILED(hr = DMORegister(L"Color Converter DMO", &CLSID_CColorConvertDMO, &DMOCATEGORY_VIDEO_EFFECT,
-            0, ARRAY_SIZE(color_convert_input), color_convert_input, ARRAY_SIZE(color_convert_output), color_convert_output)))
-        return hr;
-
-    return mfplat_DllRegisterServer();
+    return S_OK;
 }
 
 HRESULT WINAPI DllUnregisterServer(void)
@@ -1214,24 +1149,9 @@ HRESULT WINAPI DllUnregisterServer(void)
             &IID_IFilterMapper2, (void **)&mapper)))
         return hr;
 
-    IFilterMapper2_UnregisterFilter(mapper, NULL, NULL, &CLSID_AviSplitter);
     IFilterMapper2_UnregisterFilter(mapper, NULL, NULL, &CLSID_decodebin_parser);
-    IFilterMapper2_UnregisterFilter(mapper, NULL, NULL, &CLSID_CMpegAudioCodec);
-    IFilterMapper2_UnregisterFilter(mapper, NULL, NULL, &CLSID_CMpegVideoCodec);
-    IFilterMapper2_UnregisterFilter(mapper, NULL, NULL, &CLSID_mpeg_layer3_decoder);
-    IFilterMapper2_UnregisterFilter(mapper, NULL, NULL, &CLSID_MPEG1Splitter);
-    IFilterMapper2_UnregisterFilter(mapper, NULL, NULL, &CLSID_WAVEParser);
 
     IFilterMapper2_Release(mapper);
 
-    if (FAILED(hr = DMOUnregister(&CLSID_CColorConvertDMO, &DMOCATEGORY_VIDEO_EFFECT)))
-        return hr;
-    if (FAILED(hr = DMOUnregister(&CLSID_CResamplerMediaObject, &DMOCATEGORY_AUDIO_EFFECT)))
-        return hr;
-    if (FAILED(hr = DMOUnregister(&CLSID_WMADecMediaObject, &DMOCATEGORY_AUDIO_DECODER)))
-        return hr;
-    if (FAILED(hr = DMOUnregister(&CLSID_WMVDecoderMFT, &DMOCATEGORY_VIDEO_DECODER)))
-        return hr;
-
     return S_OK;
 }
