@@ -59,6 +59,7 @@
 #endif
 
 #include "media-converter.h"
+#include <gst/base/gstadapter.h>
 
 #include <assert.h>
 
@@ -69,7 +70,8 @@
 #define VIDEO_CONV_FOZ_TAG_OGVDATA   1
 #define VIDEO_CONV_FOZ_TAG_STREAM    2
 #define VIDEO_CONV_FOZ_TAG_MKVDATA   3
-#define VIDEO_CONV_FOZ_NUM_TAGS      4
+#define VIDEO_CONV_FOZ_TAG_CODEC     4
+#define VIDEO_CONV_FOZ_NUM_TAGS      5
 
 #define DURATION_NONE (UINT64_MAX)
 
@@ -88,14 +90,24 @@ struct hashes_reader
     GList *current_hash;
 };
 
+enum video_conv_state_flags
+{
+    VIDEO_CONV_STREAM_STARTED = 1,
+    VIDEO_CONV_HAS_TRANSCODED = 2,
+    VIDEO_CONV_IS_DUMPING = 4,
+};
+
 struct video_conv_state
 {
-    struct payload_hash transcode_hash;
+    struct fozdb_hash transcode_hash;
     struct fozdb *read_fozdb;
+    int blank_file;
     uint64_t upstream_duration;
     uint64_t our_duration;
     uint32_t transcoded_tag;
-    bool has_transcoded, need_stream_start;
+    uint32_t state_flags;
+    uint64_t read_offset;
+    GList *chunk_hashes;
 };
 
 typedef struct
@@ -104,6 +116,8 @@ typedef struct
     GstPad *sink_pad, *src_pad;
     pthread_mutex_t state_mutex;
     struct video_conv_state *state;
+    GstPadMode active_mode;
+    GstAdapter *adapter;
 } VideoConv;
 
 typedef struct
@@ -116,7 +130,7 @@ G_DEFINE_TYPE(VideoConv, video_conv, GST_TYPE_ELEMENT);
 #define VIDEO_CONV(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj), VIDEO_CONV_TYPE, VideoConv))
 #define parent_class    (video_conv_parent_class)
 GST_ELEMENT_REGISTER_DEFINE(protonvideoconverter, "protonvideoconverter",
-        GST_RANK_MARGINAL, VIDEO_CONV_TYPE);
+        GST_RANK_MARGINAL + 1, VIDEO_CONV_TYPE);
 
 static GstStaticPadTemplate video_conv_sink_template = GST_STATIC_PAD_TEMPLATE("sink",
         GST_PAD_SINK, GST_PAD_ALWAYS,
@@ -135,7 +149,7 @@ void hashes_reader_init(struct hashes_reader *reader, GList *hashes)
 
 static int hashes_reader_read(void *reader, uint8_t *buffer, size_t size, size_t *read_size)
 {
-    struct payload_hash *hash = (struct payload_hash *)buffer;
+    struct fozdb_hash *hash = (struct fozdb_hash *)buffer;
     struct hashes_reader *hashes_reader = reader;
 
     if (!size)
@@ -147,7 +161,7 @@ static int hashes_reader_read(void *reader, uint8_t *buffer, size_t size, size_t
     if (!hashes_reader->current_hash)
         return CONV_ERROR_DATA_END;
 
-    *hash = *(struct payload_hash *)(hashes_reader->current_hash->data);
+    *hash = *(struct fozdb_hash *)(hashes_reader->current_hash->data);
     hashes_reader->current_hash = hashes_reader->current_hash->next;
 
     *read_size = sizeof(*hash);
@@ -161,11 +175,10 @@ static int dump_fozdb_open_video(bool create)
 
 static void dump_fozdb_discard_transcoded(void)
 {
-    GList *to_discard_chunks = NULL;
-    struct payload_hash *stream_id;
+    struct rb_tree to_discard_chunks = {fozdb_entry_compare};
+    struct fozdb_entry *entry;
     struct fozdb *read_fozdb;
     char *read_fozdb_path;
-    GHashTableIter iter;
     int ret;
 
     if (dump_fozdb.already_cleaned)
@@ -192,43 +205,40 @@ static void dump_fozdb_discard_transcoded(void)
         return;
     }
 
-    fozdb_iter_tag(dump_fozdb.fozdb, VIDEO_CONV_FOZ_TAG_STREAM, &iter);
-    while (g_hash_table_iter_next(&iter, (void **)&stream_id, NULL))
+    FOZDB_FOR_EACH_TAG_ENTRY(entry, VIDEO_CONV_FOZ_TAG_STREAM, dump_fozdb.fozdb)
     {
-        struct payload_hash chunk_id;
-        uint32_t chunks_size, i;
+        struct fozdb_hash chunk_id;
+        uint32_t i;
         size_t read_size;
 
-        if (fozdb_has_entry(read_fozdb, VIDEO_CONV_FOZ_TAG_OGVDATA, stream_id))
+        if (fozdb_has_entry(read_fozdb, VIDEO_CONV_FOZ_TAG_OGVDATA, &entry->key.hash))
         {
-            if (fozdb_entry_size(dump_fozdb.fozdb, VIDEO_CONV_FOZ_TAG_STREAM, stream_id, &chunks_size) == CONV_OK)
+            if (entry->full_size)
             {
-                uint8_t *buffer = calloc(1, chunks_size);
-                if (fozdb_read_entry_data(dump_fozdb.fozdb, VIDEO_CONV_FOZ_TAG_STREAM, stream_id,
-                        0, buffer, chunks_size, &read_size, true) == CONV_OK)
+                uint8_t *buffer = calloc(1, entry->full_size);
+                if (fozdb_read_entry_data(dump_fozdb.fozdb, VIDEO_CONV_FOZ_TAG_STREAM, &entry->key.hash,
+                        0, buffer, entry->full_size, &read_size, true) == CONV_OK)
                 {
                     for (i = 0; i < read_size / sizeof(chunk_id); ++i)
                     {
-                        payload_hash_from_bytes(&chunk_id, buffer + i * sizeof(chunk_id));
-                        to_discard_chunks = g_list_append(to_discard_chunks,
-                                entry_name_create(VIDEO_CONV_FOZ_TAG_VIDEODATA, &chunk_id));
+                        fozdb_hash_from_bytes(&chunk_id, buffer + i * sizeof(chunk_id));
+                        fozdb_entry_put(&to_discard_chunks, VIDEO_CONV_FOZ_TAG_VIDEODATA, &chunk_id);
                     }
                 }
                 free(buffer);
             }
 
-            to_discard_chunks = g_list_append(to_discard_chunks,
-                    entry_name_create(VIDEO_CONV_FOZ_TAG_STREAM, stream_id));
+            fozdb_entry_put(&to_discard_chunks, VIDEO_CONV_FOZ_TAG_STREAM, &entry->key.hash);
         }
     }
 
-    if ((ret = fozdb_discard_entries(dump_fozdb.fozdb, to_discard_chunks)) < 0)
+    if ((ret = fozdb_discard_entries(dump_fozdb.fozdb, &to_discard_chunks)) < 0)
     {
         GST_ERROR("Failed to discard entries, ret %d.", ret);
         dump_fozdb_close(&dump_fozdb);
     }
 
-    g_list_free_full(to_discard_chunks, free);
+    rb_destroy(&to_discard_chunks, fozdb_entry_destroy, NULL);
 }
 
 struct pad_reader *pad_reader_create_with_stride(GstPad *pad, size_t stride)
@@ -331,8 +341,9 @@ static int video_conv_state_create(struct video_conv_state **out)
 {
     struct video_conv_state *state;
     struct fozdb *fozdb = NULL;
-    char *read_fozdb_path;
-    int ret;
+    char *read_fozdb_path, *blank_video;
+    uint64_t blank_file_size;
+    int ret, fd;
 
     if ((read_fozdb_path = getenv("MEDIACONV_VIDEO_TRANSCODED_FILE")))
     {
@@ -345,13 +356,26 @@ static int video_conv_state_create(struct video_conv_state **out)
         GST_ERROR("Env MEDIACONV_VIDEO_TRANSCODED_FILE is not set.");
         ret = CONV_ERROR_ENV_NOT_SET;
     }
+    if (!(blank_video = getenv("MEDIACONV_BLANK_VIDEO_FILE")))
+    {
+        GST_ERROR("Env MEDIACONV_BLANK_VIDEO_FILE not set.");
+        return CONV_ERROR_ENV_NOT_SET;
+    }
+
+    if (!open_file(blank_video, O_RDONLY, &fd))
+        return CONV_ERROR_OPEN_FAILED;
+    if (!get_file_size(fd, &blank_file_size))
+    {
+        close(fd);
+        return CONV_ERROR_OPEN_FAILED;
+    }
 
     state = calloc(1, sizeof(*state));
     state->read_fozdb = fozdb;
+    state->blank_file = fd;
     state->upstream_duration = DURATION_NONE;
-    state->our_duration = DURATION_NONE;
+    state->our_duration = blank_file_size;
     state->transcoded_tag = VIDEO_CONV_FOZ_TAG_MKVDATA;
-    state->need_stream_start = true;
 
     *out = state;
     return ret;
@@ -359,20 +383,21 @@ static int video_conv_state_create(struct video_conv_state **out)
 
 static void video_conv_state_release(struct video_conv_state *state)
 {
+    if ((state->state_flags & VIDEO_CONV_IS_DUMPING))
+        pthread_mutex_unlock(&dump_fozdb.mutex);
     if (state->read_fozdb)
         fozdb_release(state->read_fozdb);
+    close(state->blank_file);
     free(state);
 }
 
 /* Return true if the file is transcoded, false if not. */
-bool video_conv_state_begin_transcode(struct video_conv_state *state, struct payload_hash *hash)
+bool video_conv_state_begin_transcode(struct video_conv_state *state, struct fozdb_hash *hash)
 {
-    const char *blank_video;
-    uint64_t file_size = 0;
-    int fd;
-
     GST_DEBUG("state %p, hash %s.", state, format_hash(hash));
 
+    state->transcode_hash = *hash;
+
     if (state->read_fozdb)
     {
         uint32_t entry_size;
@@ -380,38 +405,24 @@ bool video_conv_state_begin_transcode(struct video_conv_state *state, struct pay
         if (fozdb_entry_size(state->read_fozdb, VIDEO_CONV_FOZ_TAG_MKVDATA, hash, &entry_size) == CONV_OK)
         {
             GST_DEBUG("Found an MKV video for hash %s.", format_hash(hash));
-            state->transcode_hash = *hash;
             state->our_duration = entry_size;
             state->transcoded_tag = VIDEO_CONV_FOZ_TAG_MKVDATA;
-            state->has_transcoded = true;
+            state->state_flags |= VIDEO_CONV_HAS_TRANSCODED;
             return true;
         }
 
         if (fozdb_entry_size(state->read_fozdb, VIDEO_CONV_FOZ_TAG_OGVDATA, hash, &entry_size) == CONV_OK)
         {
             GST_DEBUG("Found an OGV video for hash %s.", format_hash(hash));
-            state->transcode_hash = *hash;
             state->our_duration = entry_size;
             state->transcoded_tag = VIDEO_CONV_FOZ_TAG_OGVDATA;
-            state->has_transcoded = true;
+            state->state_flags |= VIDEO_CONV_HAS_TRANSCODED;
             return true;
         }
     }
 
     GST_INFO("No transcoded video for %s. Substituting a blank video.", format_hash(hash));
-
-    if (!(blank_video = getenv("MEDIACONV_BLANK_VIDEO_FILE")))
-    {
-        GST_ERROR("Env MEDIACONV_BLANK_VIDEO_FILE not set.");
-        return false;
-    }
-    if (open_file(blank_video, O_RDONLY, &fd))
-    {
-        get_file_size(fd, &file_size);
-        close(fd);
-    }
-    state->our_duration = file_size;
-    state->has_transcoded = false;
+    state->state_flags &= ~VIDEO_CONV_HAS_TRANSCODED;
 
     create_placeholder_file("placeholder-video-used");
 
@@ -421,13 +432,11 @@ bool video_conv_state_begin_transcode(struct video_conv_state *state, struct pay
 int video_conv_state_fill_buffer(struct video_conv_state *state, uint64_t offset,
         uint8_t *buffer, size_t size, size_t *fill_size)
 {
-    const char *blank_video;
-    uint64_t file_size;
     size_t to_copy;
     bool read_ok;
-    int fd, ret;
+    int ret;
 
-    if (state->has_transcoded)
+    if (state->state_flags & VIDEO_CONV_HAS_TRANSCODED)
     {
         if ((ret = fozdb_read_entry_data(state->read_fozdb, state->transcoded_tag, &state->transcode_hash,
                 offset, buffer, size, fill_size, false)) < 0)
@@ -436,38 +445,16 @@ int video_conv_state_fill_buffer(struct video_conv_state *state, uint64_t offset
     }
     else /* Fill blank video data to buffer. */
     {
-        if (!(blank_video = getenv("MEDIACONV_BLANK_VIDEO_FILE")))
-        {
-            GST_ERROR("Env MEDIACONV_BLANK_VIDEO_FILE not set.");
-            return CONV_ERROR_ENV_NOT_SET;
-        }
-        if (!open_file(blank_video, O_RDONLY, &fd))
-            return CONV_ERROR_OPEN_FAILED;
-        if (!get_file_size(fd, &file_size))
-        {
-            close(fd);
-            return CONV_ERROR;
-        }
-
-        /* Get copy size. */
-        if (offset >= file_size)
-        {
-            close(fd);
-            return CONV_OK;
-        }
-        to_copy = min(file_size - offset, size);
+        offset = offset % state->our_duration;
+        to_copy = min(state->our_duration - offset, size);
 
         /* Copy data. */
-        if (lseek(fd, offset, SEEK_SET) < 0)
+        if (lseek(state->blank_file, offset, SEEK_SET) < 0)
         {
-            GST_ERROR("Failed to seek %s to %#"PRIx64". %s.", blank_video, offset, strerror(errno));
-            close(fd);
+            GST_ERROR("Failed to seek to %#"PRIx64". %s.", offset, strerror(errno));
             return CONV_ERROR;
         }
-        read_ok = complete_read(fd, buffer, to_copy);
-        close(fd);
-
-        if (!read_ok)
+        if (!(read_ok = complete_read(state->blank_file, buffer, to_copy)))
         {
             GST_ERROR("Failed to read blank video data.");
             return CONV_ERROR_READ_FAILED;
@@ -528,7 +515,7 @@ static uint64_t video_conv_duration_ours_to_upstream(VideoConv *conv, uint64_t p
     struct video_conv_state *state = conv->state;
 
     if (state->upstream_duration != DURATION_NONE && state->our_duration != DURATION_NONE)
-        return pos * state->upstream_duration / state->our_duration;
+        return gst_util_uint64_scale_round(pos, state->upstream_duration, state->our_duration);
     else
         return DURATION_NONE;
 }
@@ -562,21 +549,79 @@ static bool video_conv_get_upstream_range(VideoConv *conv, uint64_t offset, uint
     return true;
 }
 
-static bool video_conv_hash_upstream_data(VideoConv *conv, struct payload_hash *hash)
+static uint64_t video_conv_duration_ours_to_downstream(VideoConv *conv, uint64_t pos)
 {
-    struct pad_reader *reader;
-    bool ret;
+    struct video_conv_state *state = conv->state;
+
+    if (state->upstream_duration != DURATION_NONE && state->our_duration != DURATION_NONE)
+        return gst_util_uint64_scale_round(pos, state->our_duration, state->upstream_duration);
+    else
+        return DURATION_NONE;
+}
+
+static bool video_conv_get_downstream_range(VideoConv *conv, uint64_t offset, uint32_t end,
+        uint64_t *downstream_offset, uint64_t *downstream_end)
+{
+    struct video_conv_state *state;
+
+    if (!(state = video_conv_lock_state(conv)))
+        return false;
+
+    if (state->upstream_duration == DURATION_NONE)
+        video_conv_query_upstream_duration(conv);
+
+    *downstream_offset = video_conv_duration_ours_to_downstream(conv, offset);
+    *downstream_end = video_conv_duration_ours_to_downstream(conv, end);
+
+    pthread_mutex_unlock(&conv->state_mutex);
+
+    return true;
+}
+
+static bool video_conv_hash_upstream_data(VideoConv *conv, struct fozdb_hash *hash)
+{
+    bool ret = false;
 
     memset(hash, 0, sizeof(*hash));
 
-    reader = pad_reader_create_with_stride(conv->sink_pad, HASH_STRIDE);
-    ret = murmur3_128(reader, pad_reader_read, HASH_SEED, hash);
-    pad_reader_release(reader);
+    if (conv->active_mode == GST_PAD_MODE_PUSH && gst_adapter_available(conv->adapter) > 0)
+    {
+        struct bytes_reader bytes_reader;
+        gsize read_size = gst_adapter_available(conv->adapter);
+        const void *buffer = gst_adapter_map(conv->adapter, read_size);
+        bytes_reader_init(&bytes_reader, buffer, read_size);
+        ret = murmur3_128(&bytes_reader, bytes_reader_read, HASH_SEED, hash);
+        gst_adapter_unmap(conv->adapter);
+        gst_adapter_clear(conv->adapter);
+    }
+    else if (gst_pad_activate_mode(conv->sink_pad, GST_PAD_MODE_PULL, true))
+    {
+        struct pad_reader *reader = pad_reader_create_with_stride(conv->sink_pad, HASH_STRIDE);
+        ret = murmur3_128(reader, pad_reader_read, HASH_SEED, hash);
+        pad_reader_release(reader);
+        conv->active_mode = GST_PAD_MODE_PULL;
+    }
 
     return ret;
 }
 
-static int video_conv_dump_upstream_data(VideoConv *conv, struct payload_hash *hash)
+static int video_conv_dump_upstream_chunk(VideoConv *conv, const void *buffer, size_t read_size,
+        GList **chunk_hashes)
+{
+    struct bytes_reader bytes_reader;
+    struct fozdb_hash *chunk_hash;
+
+    bytes_reader_init(&bytes_reader, buffer, read_size);
+    chunk_hash = calloc(1, sizeof(*chunk_hash));
+    murmur3_128(&bytes_reader, bytes_reader_read, HASH_SEED, chunk_hash);
+    *chunk_hashes = g_list_append(*chunk_hashes, chunk_hash);
+
+    bytes_reader_init(&bytes_reader, buffer, read_size);
+    return fozdb_write_entry(dump_fozdb.fozdb, VIDEO_CONV_FOZ_TAG_VIDEODATA, chunk_hash,
+            &bytes_reader, bytes_reader_read, true);
+}
+
+static int video_conv_dump_upstream_data(VideoConv *conv, struct fozdb_hash *hash)
 {
     struct hashes_reader chunk_hashes_reader;
     struct pad_reader *pad_reader = NULL;
@@ -593,21 +638,17 @@ static int video_conv_dump_upstream_data(VideoConv *conv, struct payload_hash *h
         goto done;
     }
 
+    if (conv->active_mode == GST_PAD_MODE_PUSH)
+    {
+        conv->state->state_flags |= VIDEO_CONV_IS_DUMPING;
+        return 0;
+    }
+
     buffer = calloc(1, HASH_CHUNK_SIZE);
     pad_reader = pad_reader_create(conv->sink_pad);
     while ((ret = pad_reader_read(pad_reader, buffer, HASH_CHUNK_SIZE, &read_size)) == CONV_OK)
     {
-        struct bytes_reader bytes_reader;
-        struct payload_hash *chunk_hash;
-
-        bytes_reader_init(&bytes_reader, buffer, read_size);
-        chunk_hash = calloc(1, sizeof(*chunk_hash));
-        murmur3_128(&bytes_reader, bytes_reader_read, HASH_SEED, chunk_hash);
-        chunk_hashes = g_list_append(chunk_hashes, chunk_hash);
-
-        bytes_reader_init(&bytes_reader, buffer, read_size);
-        if ((ret = fozdb_write_entry(dump_fozdb.fozdb, VIDEO_CONV_FOZ_TAG_VIDEODATA, chunk_hash,
-                &bytes_reader, bytes_reader_read, true)) < 0)
+        if ((ret = video_conv_dump_upstream_chunk(conv, buffer, read_size, &chunk_hashes)) < 0)
         {
             GST_ERROR("Error writing video data to fozdb, ret %d.", ret);
             goto done;
@@ -638,10 +679,10 @@ done:
 static void video_conv_init_transcode(VideoConv *conv)
 {
     struct video_conv_state *state = conv->state;
-    struct payload_hash hash;
+    struct fozdb_hash hash;
     int ret;
 
-    if (state->has_transcoded)
+    if (state->state_flags & VIDEO_CONV_HAS_TRANSCODED)
         return;
 
     pthread_mutex_lock(&dump_fozdb.mutex);
@@ -660,55 +701,340 @@ static void video_conv_init_transcode(VideoConv *conv)
         GST_ERROR("Failed to hash upstream data.");
     }
 
-    pthread_mutex_unlock(&dump_fozdb.mutex);
+    if (!(state->state_flags & VIDEO_CONV_IS_DUMPING))
+        pthread_mutex_unlock(&dump_fozdb.mutex);
+}
+
+static uint32_t video_conv_get_state_flags(VideoConv *conv)
+{
+    struct video_conv_state *state;
+    uint32_t state_flags;
+
+    if (!(state = video_conv_lock_state(conv)))
+    {
+        GST_ERROR("VideoConv not yet in READY state?");
+        return 0;
+    }
+    state_flags = state->state_flags;
+    pthread_mutex_unlock(&conv->state_mutex);
+
+    return state_flags;
+}
+
+static gboolean video_conv_push_stream_start(VideoConv *conv, struct fozdb_hash *hash)
+{
+    struct video_conv_state *state;
+
+    push_event(conv->src_pad, gst_event_new_stream_start(format_hash(hash)));
+
+    if (!(state = video_conv_lock_state(conv)))
+    {
+        GST_ERROR("VideoConv not yet in READY state?");
+        return false;
+    }
+    state->state_flags |= VIDEO_CONV_STREAM_STARTED;
+    pthread_mutex_unlock(&conv->state_mutex);
+
+    return true;
+}
+
+static gboolean video_conv_push_caps(VideoConv *conv, uint32_t transcode_tag)
+{
+    GstCaps *caps;
+    gboolean ret;
+
+    if (transcode_tag == VIDEO_CONV_FOZ_TAG_MKVDATA)
+        caps = gst_caps_from_string("video/x-matroska");
+    else if (transcode_tag == VIDEO_CONV_FOZ_TAG_OGVDATA)
+        caps = gst_caps_from_string("application/ogg");
+    else
+        return false;
+
+    ret = push_event(conv->src_pad, gst_event_new_caps(caps));
+    gst_caps_unref(caps);
+    return ret;
+}
+
+static gboolean video_conv_push_segment(VideoConv *conv)
+{
+    struct video_conv_state *state;
+    GstSegment segment;
+
+    gst_segment_init(&segment, GST_FORMAT_BYTES);
+    if (!(state = video_conv_lock_state(conv)))
+    {
+        GST_ERROR("VideoConv not yet in READY state?");
+        return false;
+    }
+    segment.stop = state->our_duration;
+    pthread_mutex_unlock(&conv->state_mutex);
+
+    return push_event(conv->src_pad, gst_event_new_segment(&segment));
+}
+
+static gboolean video_conv_sink_event_caps(VideoConv *conv, GstEvent *event)
+{
+    struct video_conv_state *state;
+    uint32_t transcode_tag;
+
+    gst_event_unref(event);
+
+    /* push_event, below, can also grab state and cause a deadlock, so make sure it's
+     * unlocked before calling */
+    if (!(state = video_conv_lock_state(conv)))
+    {
+        GST_ERROR("VideoConv not yet in READY state?");
+        return false;
+    }
+
+    video_conv_init_transcode(conv);
+    transcode_tag = state->transcoded_tag;
+
+    pthread_mutex_unlock(&conv->state_mutex);
+
+    return video_conv_push_caps(conv, transcode_tag);
+}
+
+static gboolean video_conv_sink_event_eos(VideoConv *conv, GstEvent *event)
+{
+    struct video_conv_state *state;
+    struct fozdb_hash hash;
+    uint32_t transcode_tag;
+    uint32_t state_flags;
+    int ret;
+
+    gst_event_unref(event);
+
+    if (!(state = video_conv_lock_state(conv)))
+        return false;
+
+    if (state->state_flags & VIDEO_CONV_IS_DUMPING)
+    {
+        struct hashes_reader chunk_hashes_reader;
+        gsize read_bytes;
+
+        if ((read_bytes = gst_adapter_available(conv->adapter)))
+        {
+            const void *buffer = gst_adapter_map(conv->adapter, read_bytes);
+
+            if ((ret = video_conv_dump_upstream_chunk(conv, buffer, read_bytes, &state->chunk_hashes)) < 0)
+                GST_ERROR("Error writing stream data to fozdb, ret %d.", ret);
+
+            gst_adapter_unmap(conv->adapter);
+            gst_adapter_clear(conv->adapter);
+        }
+
+        hashes_reader_init(&chunk_hashes_reader, state->chunk_hashes);
+        if ((ret = fozdb_write_entry(dump_fozdb.fozdb, VIDEO_CONV_FOZ_TAG_STREAM, &state->transcode_hash,
+                &chunk_hashes_reader, hashes_reader_read, true)) < 0)
+            GST_ERROR("Error writing stream data to fozdb, ret %d.", ret);
+
+        if (state->chunk_hashes)
+            g_list_free_full(state->chunk_hashes, free);
+
+        pthread_mutex_unlock(&dump_fozdb.mutex);
+        state->state_flags &= ~VIDEO_CONV_IS_DUMPING;
+        pthread_mutex_unlock(&conv->state_mutex);
+
+        return gst_pad_push_event(conv->src_pad, gst_event_new_eos());
+    }
+
+    video_conv_init_transcode(conv);
+    hash = state->transcode_hash;
+    state_flags = state->state_flags;
+    transcode_tag = state->transcoded_tag;
+
+    pthread_mutex_unlock(&conv->state_mutex);
+
+    if (!(state_flags & VIDEO_CONV_STREAM_STARTED))
+    {
+        /* rewind and start a new stream for dumping or playback */
+        if (!push_event(conv->sink_pad, gst_event_new_seek(1.0, GST_FORMAT_BYTES, GST_SEEK_FLAG_FLUSH,
+                GST_SEEK_TYPE_SET, 0, GST_SEEK_TYPE_NONE, -1)))
+            return false;
+        if (!video_conv_push_stream_start(conv, &hash))
+            return false;
+        if (!video_conv_push_caps(conv, transcode_tag))
+            return false;
+        if (!video_conv_push_segment(conv))
+            return false;
+
+        /* return false to cancel upstream pads EOS event handling and avoid setting EOS flag */
+        return false;
+    }
+
+    return gst_pad_push_event(conv->src_pad, gst_event_new_eos());
 }
 
 static gboolean video_conv_sink_event(GstPad *pad, GstObject *parent, GstEvent *event)
 {
     VideoConv *conv = VIDEO_CONV(parent);
-    bool ret;
 
     GST_DEBUG_OBJECT(pad, "Got event %"GST_PTR_FORMAT".", event);
 
     if (event->type == GST_EVENT_CAPS)
+        return video_conv_sink_event_caps(conv, event);
+    if (event->type == GST_EVENT_EOS)
+        return video_conv_sink_event_eos(conv, event);
+
+    if (video_conv_get_state_flags(conv) & VIDEO_CONV_STREAM_STARTED)
+        return gst_pad_event_default(pad, parent, event);
+
+    gst_event_unref(event);
+    return true;
+}
+
+static gboolean video_conv_src_event_seek(VideoConv *conv, GstEvent *event)
+{
+    guint seqnum = gst_event_get_seqnum(event);
+    guint64 upstream_offset, upstream_size;
+    GstSeekType offset_type, stop_type;
+    struct video_conv_state *state;
+    gint64 offset, stop;
+    GstSeekFlags flags;
+    GstFormat format;
+    gdouble rate;
+
+    gst_event_parse_seek(event, &rate, &format, &flags, &offset_type, &offset, &stop_type, &stop);
+    gst_event_unref(event);
+    if (format != GST_FORMAT_BYTES)
+        return false;
+
+    GST_TRACE("conv %p, rate %f, format %s, flags %#x, offset_type %u, cur %#" G_GINT64_MODIFIER "x, "
+            "stop_type %u, stop %#" G_GINT64_MODIFIER "x.", conv, rate, gst_format_get_name(format),
+            flags, offset_type, offset, stop_type, stop);
+
+    if (!(state = video_conv_lock_state(conv)))
+        return GST_FLOW_ERROR;
+    if (state->state_flags & VIDEO_CONV_IS_DUMPING)
     {
-        struct video_conv_state *state;
-        uint32_t transcode_tag;
-        GstCaps *caps;
-
-        /* push_event, below, can also grab state and cause a deadlock, so make sure it's
-         * unlocked before calling */
-        if (!(state = video_conv_lock_state(conv)))
-        {
-            GST_ERROR("VideoConv not yet in READY state?");
-            return false;
-        }
-
-        if (!gst_pad_activate_mode(conv->sink_pad, GST_PAD_MODE_PULL, true))
-        {
-            GST_ERROR("Failed to activate sink pad in pull mode.");
-            pthread_mutex_unlock(&conv->state_mutex);
-            return false;
-        }
-
-        video_conv_init_transcode(conv);
-        transcode_tag = state->transcoded_tag;
-
         pthread_mutex_unlock(&conv->state_mutex);
+        return true;
+    }
+    pthread_mutex_unlock(&conv->state_mutex);
 
-        if (transcode_tag == VIDEO_CONV_FOZ_TAG_MKVDATA)
-            caps = gst_caps_from_string("video/x-matroska");
-        else if (transcode_tag == VIDEO_CONV_FOZ_TAG_OGVDATA)
-            caps = gst_caps_from_string("application/ogg");
-        else
-            return false;
+    if (!video_conv_get_upstream_range(conv, offset, HASH_CHUNK_SIZE, &upstream_offset, &upstream_size))
+        return false;
 
-        ret = push_event(conv->src_pad, gst_event_new_caps(caps));
-        gst_caps_unref(caps);
-        return ret;
+    if ((event = gst_event_new_seek(1.0, GST_FORMAT_BYTES, GST_SEEK_FLAG_FLUSH,
+            GST_SEEK_TYPE_SET, upstream_offset, GST_SEEK_TYPE_NONE, -1)))
+        gst_event_set_seqnum(event, seqnum);
+    return push_event(conv->sink_pad, event);
+}
+
+static gboolean video_conv_src_event(GstPad *pad, GstObject *parent, GstEvent *event)
+{
+    VideoConv *conv = VIDEO_CONV(parent);
+
+    GST_DEBUG_OBJECT(pad, "Got event %"GST_PTR_FORMAT".", event);
+
+    if (event->type == GST_EVENT_SEEK)
+        return video_conv_src_event_seek(conv, event);
+
+    if (video_conv_get_state_flags(conv) & VIDEO_CONV_STREAM_STARTED)
+        return gst_pad_event_default(pad, parent, event);
+
+    gst_event_unref(event);
+    return true;
+}
+
+static void video_conv_dump_buffered_chunks(VideoConv *conv, gsize bytes_available)
+{
+    struct video_conv_state *state = conv->state;
+    int ret;
+
+    while (bytes_available >= HASH_CHUNK_SIZE)
+    {
+        const void *buffer = gst_adapter_map(conv->adapter, HASH_CHUNK_SIZE);
+        if ((ret = video_conv_dump_upstream_chunk(conv, buffer, HASH_CHUNK_SIZE, &state->chunk_hashes)) < 0)
+        {
+            pthread_mutex_unlock(&dump_fozdb.mutex);
+            state->state_flags &= ~VIDEO_CONV_IS_DUMPING;
+            break;
+        }
+
+        gst_adapter_unmap(conv->adapter);
+        gst_adapter_flush(conv->adapter, HASH_CHUNK_SIZE);
+        bytes_available -= HASH_CHUNK_SIZE;
+    }
+}
+
+static GstFlowReturn video_conv_sink_chain(GstPad *pad, GstObject *parent, GstBuffer *buffer)
+{
+    gsize buffer_size = gst_buffer_get_size(buffer), buffer_offset = GST_BUFFER_OFFSET(buffer);
+    uint64_t downstream_offset, downstream_end, seek_offset = -1;
+    VideoConv *conv = VIDEO_CONV(parent);
+    struct video_conv_state *state;
+    GstBuffer *transcoded = NULL;
+    uint32_t state_flags;
+    int ret = 0;
+
+    GST_DEBUG_OBJECT(pad, "Got buffer %"GST_PTR_FORMAT".", buffer);
+
+    if (!(state = video_conv_lock_state(conv)))
+        return GST_FLOW_ERROR;
+    state_flags = state->state_flags;
+
+    if ((state_flags & VIDEO_CONV_IS_DUMPING) || !(state_flags & VIDEO_CONV_STREAM_STARTED))
+    {
+        gsize bytes_available = gst_adapter_available(conv->adapter) + buffer_size;
+        gst_adapter_push(conv->adapter, buffer);
+
+        if (state_flags & VIDEO_CONV_IS_DUMPING)
+            video_conv_dump_buffered_chunks(conv, bytes_available);
+        else if (!(bytes_available % HASH_CHUNK_SIZE) && bytes_available >= HASH_STRIDE)
+        {
+            state->read_offset += HASH_STRIDE;
+            seek_offset = state->read_offset;
+        }
     }
 
-    return gst_pad_event_default(pad, parent, event);
+    pthread_mutex_unlock(&conv->state_mutex);
+
+    if (ret < 0)
+    {
+        GST_ERROR("Failed to dump or read transcoded buffer, error %d.", ret);
+        if (transcoded)
+            gst_buffer_unref(transcoded);
+        return GST_FLOW_ERROR;
+    }
+
+    if (seek_offset != -1 && !push_event(pad, gst_event_new_seek(1.0, GST_FORMAT_BYTES, GST_SEEK_FLAG_FLUSH,
+            GST_SEEK_TYPE_SET, seek_offset, GST_SEEK_TYPE_NONE, -1)))
+        return GST_FLOW_ERROR;
+    if (!(state_flags & VIDEO_CONV_STREAM_STARTED))
+        return GST_FLOW_OK;
+
+    if (!video_conv_get_downstream_range(conv, buffer_offset, buffer_offset + buffer_size,
+            &downstream_offset, &downstream_end))
+        return GST_FLOW_ERROR;
+
+    if (downstream_end == downstream_offset)
+        return GST_FLOW_OK;
+    if (!(transcoded = gst_buffer_new_and_alloc(downstream_end - downstream_offset)))
+        return GST_FLOW_ERROR;
+    else
+    {
+        GstBufferMapInfo map;
+        size_t fill_size = 0;
+
+        if (gst_buffer_map(transcoded, &map, GST_MAP_READWRITE))
+        {
+            ret = video_conv_state_fill_buffer(state, downstream_offset, map.data, map.size, &fill_size);
+            gst_buffer_unmap(transcoded, &map);
+            gst_buffer_set_size(transcoded, fill_size);
+        }
+
+        if (gst_buffer_get_size(transcoded))
+        {
+            GST_BUFFER_OFFSET(transcoded) = downstream_offset;
+            return gst_pad_push(conv->src_pad, transcoded);
+        }
+
+        gst_buffer_unref(transcoded);
+    }
+
+    return GST_FLOW_OK;
 }
 
 static GstFlowReturn video_conv_src_get_range(GstPad *pad, GstObject *parent,
@@ -809,7 +1135,13 @@ static gboolean video_conv_src_query(GstPad *pad, GstObject *parent, GstQuery *q
         gst_query_unref(peer_query);
 
         gst_query_set_scheduling(query, flags, min, max, align);
-        gst_query_add_scheduling_mode(query, GST_PAD_MODE_PULL);
+        if (conv->active_mode != GST_PAD_MODE_NONE)
+            gst_query_add_scheduling_mode(query, conv->active_mode);
+        else
+        {
+            gst_query_add_scheduling_mode(query, GST_PAD_MODE_PULL);
+            gst_query_add_scheduling_mode(query, GST_PAD_MODE_PUSH);
+        }
 
         return true;
 
@@ -842,9 +1174,8 @@ static gboolean video_conv_src_active_mode(GstPad *pad, GstObject *parent, GstPa
 {
     VideoConv *conv = VIDEO_CONV(parent);
     struct video_conv_state *state;
-    struct payload_hash hash;
-    bool need_stream_start;
-    bool has_transcoded;
+    struct fozdb_hash hash;
+    uint32_t state_flags;
 
     GST_DEBUG_OBJECT(pad, "mode %s, active %d.", gst_pad_mode_get_name(mode), active);
 
@@ -855,6 +1186,7 @@ static gboolean video_conv_src_active_mode(GstPad *pad, GstObject *parent, GstPa
         return false;
     }
 
+    conv->active_mode = mode;
     if (mode != GST_PAD_MODE_PULL)
         return true;
 
@@ -866,34 +1198,22 @@ static gboolean video_conv_src_active_mode(GstPad *pad, GstObject *parent, GstPa
 
     video_conv_init_transcode(conv);
     hash = state->transcode_hash;
-    need_stream_start = state->need_stream_start;
-    has_transcoded = state->has_transcoded;
+    state_flags = state->state_flags;
 
     /* push_event, below, can also grab state and cause a deadlock, so make sure it's
      * unlocked before calling */
     pthread_mutex_unlock(&conv->state_mutex);
 
-    if (need_stream_start && active && has_transcoded)
-    {
-        push_event(conv->src_pad, gst_event_new_stream_start(format_hash(&hash)));
-
-        if (!(state = video_conv_lock_state(conv)))
-        {
-            GST_ERROR("VideoConv not yet in READY state?");
-            return false;
-        }
-        state->need_stream_start = false;
-        pthread_mutex_unlock(&conv->state_mutex);
-    }
-
+    if (active && !(state_flags & VIDEO_CONV_STREAM_STARTED) && (state_flags & VIDEO_CONV_HAS_TRANSCODED))
+        return video_conv_push_stream_start(conv, &hash);
     return true;
 }
 
-
 static void video_conv_finalize(GObject *object)
 {
     VideoConv *conv = VIDEO_CONV(object);
 
+    gst_object_unref(conv->adapter);
     pthread_mutex_destroy(&conv->state_mutex);
     if (conv->state)
         video_conv_state_release(conv->state);
@@ -925,9 +1245,11 @@ static void video_conv_init(VideoConv *conv)
 
     conv->sink_pad = gst_pad_new_from_static_template(&video_conv_sink_template, "sink");
     gst_pad_set_event_function(conv->sink_pad, GST_DEBUG_FUNCPTR(video_conv_sink_event));
+    gst_pad_set_chain_function(conv->sink_pad, GST_DEBUG_FUNCPTR(video_conv_sink_chain));
     gst_element_add_pad(element, conv->sink_pad);
 
     conv->src_pad = gst_pad_new_from_static_template(&video_conv_src_template, "src");
+    gst_pad_set_event_function(conv->src_pad, GST_DEBUG_FUNCPTR(video_conv_src_event));
     gst_pad_set_getrange_function(conv->src_pad, GST_DEBUG_FUNCPTR(video_conv_src_get_range));
     gst_pad_set_query_function(conv->src_pad, GST_DEBUG_FUNCPTR(video_conv_src_query));
     gst_pad_set_activatemode_function(conv->src_pad, GST_DEBUG_FUNCPTR(video_conv_src_active_mode));
@@ -935,4 +1257,166 @@ static void video_conv_init(VideoConv *conv)
 
     pthread_mutex_init(&conv->state_mutex, NULL);
     conv->state = NULL;
+    conv->adapter = gst_adapter_new();
+    conv->active_mode = GST_PAD_MODE_NONE;
+}
+
+static bool codec_info_to_wg_format(char *codec_info, struct wg_format *codec_format)
+{
+    char *codec_name = codec_info;
+
+    /* Get codec name. */
+    while (*codec_info && *codec_info != ' ')
+        ++codec_info;
+    *(codec_info++) = 0;
+
+    /* FIXME: Get width, height, fps etc. from codec info string. */
+    if (strcmp(codec_name, "cinepak") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_VIDEO_CINEPAK;
+    }
+    else if (strcmp(codec_name, "h264") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_VIDEO_H264;
+    }
+    else if (strcmp(codec_name, "wmv1") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_VIDEO_WMV;
+        codec_format->u.video.format = WG_VIDEO_FORMAT_WMV1;
+    }
+    else if (strcmp(codec_name, "wmv2") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_VIDEO_WMV;
+        codec_format->u.video.format = WG_VIDEO_FORMAT_WMV2;
+    }
+    else if (strcmp(codec_name, "wmv3") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_VIDEO_WMV;
+        codec_format->u.video.format = WG_VIDEO_FORMAT_WMV3;
+    }
+    else if  (strcmp(codec_name, "vc1") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_VIDEO_WMV;
+        codec_format->u.video.format = WG_VIDEO_FORMAT_WVC1;
+    }
+    else if  (strcmp(codec_name, "wmav1") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_AUDIO_WMA;
+        codec_format->u.audio.version = 1;
+    }
+    else if  (strcmp(codec_name, "wmav2") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_AUDIO_WMA;
+        codec_format->u.audio.version = 2;
+    }
+    else if  (strcmp(codec_name, "wmapro") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_AUDIO_WMA;
+        codec_format->u.audio.version = 3;
+    }
+    else if  (strcmp(codec_name, "wmalossless") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_AUDIO_WMA;
+        codec_format->u.audio.version = 4;
+    }
+    else if  (strcmp(codec_name, "xma1") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_AUDIO_WMA;
+        codec_format->u.audio.version = 1;
+        codec_format->u.audio.is_xma = true;
+    }
+    else if  (strcmp(codec_name, "xma2") == 0)
+    {
+        codec_format->major_type = WG_MAJOR_TYPE_AUDIO_WMA;
+        codec_format->u.audio.version = 2;
+        codec_format->u.audio.is_xma = true;
+    }
+    else
+    {
+        GST_FIXME("Unsupported codec name: %s.\n", codec_name);
+        return false;
+    }
+
+    GST_INFO("Got codec format major type %u.", codec_format->major_type);
+
+    return true;
+}
+
+gint compare_type(const GValue *value_element, GType type)
+{
+    GstElement *element = g_value_get_object(value_element);
+    return !G_TYPE_CHECK_INSTANCE_TYPE(element, type);
+}
+
+static GstElement *gst_bin_get_by_type(GstBin * bin, GType type)
+{
+    GstElement *element = NULL;
+    GstIterator *children;
+    GValue result = {0};
+    gboolean found;
+
+    children = gst_bin_iterate_recurse(bin);
+    found = gst_iterator_find_custom(children, (GCompareFunc)compare_type,
+            &result, (gpointer)type);
+    gst_iterator_free(children);
+
+    if (found)
+    {
+        element = g_value_dup_object(&result);
+        g_value_unset (&result);
+    }
+
+    return element;
+}
+
+bool get_untranscoded_stream_format(GstElement *container, uint32_t stream_index, struct wg_format *codec_format)
+{
+    struct video_conv_state *state;
+    uint8_t *buffer = NULL;
+    uint32_t entry_size, i;
+    char *codec_info;
+    size_t read_size;
+    bool ret = false;
+    VideoConv *conv;
+    int conv_ret;
+
+    if (!(conv = VIDEO_CONV(gst_bin_get_by_type(GST_BIN(container), VIDEO_CONV_TYPE))))
+    {
+        GST_WARNING("Failed to find video converter from %"GST_PTR_FORMAT".", container);
+        return false;
+    }
+
+    if (!(state = video_conv_lock_state(conv)))
+        return false;
+    if (!(state->state_flags & VIDEO_CONV_HAS_TRANSCODED))
+        goto done;
+
+    if (fozdb_entry_size(state->read_fozdb, VIDEO_CONV_FOZ_TAG_CODEC, &state->transcode_hash, &entry_size) < 0)
+    {
+        GST_WARNING("Failed to find codec info entry for stream %s.", format_hash(&state->transcode_hash));
+        goto done;
+    }
+
+    buffer = calloc(1, entry_size + 1);
+    if ((conv_ret = fozdb_read_entry_data(state->read_fozdb, VIDEO_CONV_FOZ_TAG_CODEC, &state->transcode_hash, 0,
+            buffer, entry_size, &read_size, false)) < 0)
+    {
+        GST_ERROR("Failed to read codec info, ret %d.", ret);
+        goto done;
+    }
+
+    /* Get stream codec info line by line. */
+    codec_info = strtok((char *)buffer, "\n");
+    for (i = 0; codec_info && i < stream_index; ++i)
+        codec_info = strtok(NULL, "\n");
+
+    GST_INFO("Got codec info \"%s\" for stream %d.\n", codec_info, stream_index);
+
+   ret = codec_info_to_wg_format(codec_info, codec_format);
+
+done:
+    if (buffer)
+        free(buffer);
+    pthread_mutex_unlock(&conv->state_mutex);
+    return ret;
 }
