@@ -20,6 +20,7 @@
 #define __WINE_D3DKMTHK_H
 
 #include <d3dukmdt.h>
+#include <winternl.h>
 
 typedef enum _D3DKMT_VIDPNSOURCEOWNER_TYPE
 {
@@ -189,8 +190,8 @@ typedef enum _KMTQUERYADAPTERINFOTYPE
     KMTQAITYPE_MODELIST,
     KMTQAITYPE_CHECKDRIVERUPDATESTATUS,
     KMTQAITYPE_VIRTUALADDRESSINFO,
-    KMTQAITYPE_DRIVERVERSION,
-    KMTQAITYPE_ADAPTERTYPE,
+    KMTQAITYPE_DRIVERVERSION = 13,
+    KMTQAITYPE_ADAPTERTYPE = 15,
     KMTQAITYPE_OUTPUTDUPLCONTEXTSCOUNT,
     KMTQAITYPE_WDDM_1_2_CAPS,
     KMTQAITYPE_UMD_DRIVER_VERSION,
@@ -243,10 +244,10 @@ typedef enum _KMTQUERYADAPTERINFOTYPE
     KMTQAITYPE_DRIVER_DESCRIPTION,
     KMTQAITYPE_DRIVER_DESCRIPTION_RENDER,
     KMTQAITYPE_SCANOUT_CAPS,
-    KMTQAITYPE_DISPLAY_UMDRIVERNAME,
     KMTQAITYPE_PARAVIRTUALIZATION_RENDER,
     KMTQAITYPE_SERVICENAME,
     KMTQAITYPE_WDDM_2_7_CAPS,
+    KMTQAITYPE_DISPLAY_UMDRIVERNAME = 71,
     KMTQAITYPE_TRACKEDWORKLOAD_SUPPORT,
     KMTQAITYPE_HYBRID_DLIST_DLL_SUPPORT,
     KMTQAITYPE_DISPLAY_CAPS,
@@ -781,12 +782,498 @@ typedef struct _D3DKMT_ADAPTERINFO
   BOOL          bPrecisePresentRegionsPreferred;
 } D3DKMT_ADAPTERINFO;
 
+#define MAX_ENUM_ADAPTERS 16
+
+typedef struct _D3DKMT_ENUMADAPTERS
+{
+    ULONG              NumAdapters;
+    D3DKMT_ADAPTERINFO Adapters[MAX_ENUM_ADAPTERS];
+} D3DKMT_ENUMADAPTERS;
+
 typedef struct _D3DKMT_ENUMADAPTERS2
 {
   ULONG               NumAdapters;
   D3DKMT_ADAPTERINFO *pAdapters;
 } D3DKMT_ENUMADAPTERS2;
 
+typedef struct _D3DKMT_CREATEKEYEDMUTEX
+{
+    UINT64 InitialValue;
+    D3DKMT_HANDLE hSharedHandle;
+    D3DKMT_HANDLE hKeyedMutex;
+} D3DKMT_CREATEKEYEDMUTEX;
+
+typedef struct _D3DDDICB_SIGNALFLAGS
+{
+    union
+    {
+        struct
+        {
+            UINT SignalAtSubmission : 1;
+            UINT EnqueueCpuEvent : 1;
+            UINT AllowFenceRewind : 1;
+            UINT Reserved : 28;
+            UINT DXGK_SIGNAL_FLAG_INTERNAL0 : 1;
+        };
+        UINT Value;
+    };
+} D3DDDICB_SIGNALFLAGS;
+
+typedef struct _D3DKMT_CREATEKEYEDMUTEX2_FLAGS
+{
+    union
+    {
+        struct
+        {
+            UINT NtSecuritySharing : 1;
+            UINT Reserved : 31;
+        };
+        UINT Value;
+    };
+} D3DKMT_CREATEKEYEDMUTEX2_FLAGS;
+
+typedef struct _D3DKMT_CREATEKEYEDMUTEX2
+{
+    UINT64 InitialValue;
+    D3DKMT_HANDLE hSharedHandle;
+    D3DKMT_HANDLE hKeyedMutex;
+    void *pPrivateRuntimeData;
+    UINT PrivateRuntimeDataSize;
+    D3DKMT_CREATEKEYEDMUTEX2_FLAGS Flags;
+} D3DKMT_CREATEKEYEDMUTEX2;
+
+typedef struct _D3DKMT_DESTROYKEYEDMUTEX
+{
+    D3DKMT_HANDLE hKeyedMutex;
+} D3DKMT_DESTROYKEYEDMUTEX;
+
+typedef struct _D3DKMT_OPENKEYEDMUTEX
+{
+    D3DKMT_HANDLE hSharedHandle;
+    D3DKMT_HANDLE hKeyedMutex;
+} D3DKMT_OPENKEYEDMUTEX;
+
+typedef struct _D3DKMT_OPENKEYEDMUTEX2
+{
+    D3DKMT_HANDLE hSharedHandle;
+    D3DKMT_HANDLE hKeyedMutex;
+    void *pPrivateRuntimeData;
+    UINT PrivateRuntimeDataSize;
+} D3DKMT_OPENKEYEDMUTEX2;
+
+typedef struct _D3DKMT_OPENKEYEDMUTEXFROMNTHANDLE
+{
+    HANDLE hNtHandle;
+    D3DKMT_HANDLE hKeyedMutex;
+    void *pPrivateRuntimeData;
+    UINT PrivateRuntimeDataSize;
+} D3DKMT_OPENKEYEDMUTEXFROMNTHANDLE;
+
+typedef ULONGLONG D3DGPU_VIRTUAL_ADDRESS;
+
+#ifndef D3DDDI_SYNCHRONIZATIONOBJECT_FLAGS_EXT
+#define D3DDDI_SYNCHRONIZATIONOBJECT_FLAGS_EXT
+#define D3DDDI_SYNCHRONIZATIONOBJECT_FLAGS_RESERVED0 Reserved0
+#endif
+
+typedef struct _D3DDDI_SYNCHRONIZATIONOBJECT_FLAGS
+{
+    union
+    {
+        struct
+        {
+            UINT Shared : 1;
+            UINT NtSecuritySharing : 1;
+            UINT CrossAdapter : 1;
+            UINT TopOfPipeline : 1;
+            UINT NoSignal : 1;
+            UINT NoWait : 1;
+            UINT NoSignalMaxValueOnTdr : 1;
+            UINT NoGPUAccess : 1;
+            UINT Reserved : 23;
+            UINT D3DDDI_SYNCHRONIZATIONOBJECT_FLAGS_RESERVED0 : 1;
+        };
+        UINT Value;
+    };
+} D3DDDI_SYNCHRONIZATIONOBJECT_FLAGS;
+
+typedef UINT D3DDDI_VIDEO_PRESENT_TARGET_ID;
+
+typedef enum _D3DDDI_SYNCHRONIZATIONOBJECT_TYPE
+{
+    D3DDDI_SYNCHRONIZATION_MUTEX = 1,
+    D3DDDI_SEMAPHORE = 2,
+    D3DDDI_FENCE = 3,
+    D3DDDI_CPU_NOTIFICATION = 4,
+    D3DDDI_MONITORED_FENCE = 5,
+    D3DDDI_PERIODIC_MONITORED_FENCE = 6,
+    D3DDDI_SYNCHRONIZATION_TYPE_LIMIT
+} D3DDDI_SYNCHRONIZATIONOBJECT_TYPE;
+
+typedef struct _D3DDDI_SYNCHRONIZATIONOBJECTINFO
+{
+    D3DDDI_SYNCHRONIZATIONOBJECT_TYPE Type;
+    union
+    {
+        struct
+        {
+            BOOL InitialState;
+        } SynchronizationMutex;
+        struct
+        {
+            UINT MaxCount;
+            UINT InitialCount;
+        } Semaphore;
+        struct
+        {
+            UINT Reserved[16];
+        } Reserved;
+    };
+} D3DDDI_SYNCHRONIZATIONOBJECTINFO;
+
+typedef struct _D3DKMT_CREATESYNCHRONIZATIONOBJECT
+{
+    D3DKMT_HANDLE hDevice;
+    D3DDDI_SYNCHRONIZATIONOBJECTINFO Info;
+    D3DKMT_HANDLE hSyncObject;
+} D3DKMT_CREATESYNCHRONIZATIONOBJECT;
+
+typedef struct _D3DDDI_SYNCHRONIZATIONOBJECTINFO2
+{
+    D3DDDI_SYNCHRONIZATIONOBJECT_TYPE Type;
+    D3DDDI_SYNCHRONIZATIONOBJECT_FLAGS Flags;
+    union
+    {
+        struct
+        {
+            BOOL InitialState;
+        } SynchronizationMutex;
+        struct
+        {
+            UINT MaxCount;
+            UINT InitialCount;
+        } Semaphore;
+        struct
+        {
+            UINT64 FenceValue;
+        } Fence;
+        struct
+        {
+            HANDLE Event;
+        } CPUNotification;
+        struct
+        {
+            UINT64 InitialFenceValue;
+            void *FenceValueCPUVirtualAddress;
+            D3DGPU_VIRTUAL_ADDRESS FenceValueGPUVirtualAddress;
+            UINT EngineAffinity;
+        } MonitoredFence;
+        struct
+        {
+            D3DKMT_HANDLE hAdapter;
+            D3DDDI_VIDEO_PRESENT_TARGET_ID VidPnTargetId;
+            UINT64 Time;
+            void *FenceValueCPUVirtualAddress;
+            D3DGPU_VIRTUAL_ADDRESS FenceValueGPUVirtualAddress;
+            UINT EngineAffinity;
+        } PeriodicMonitoredFence;
+        struct
+        {
+            UINT64 Reserved[8];
+        } Reserved;
+    };
+    D3DKMT_HANDLE SharedHandle;
+} D3DDDI_SYNCHRONIZATIONOBJECTINFO2;
+
+typedef struct _D3DKMT_CREATESYNCHRONIZATIONOBJECT2
+{
+    D3DKMT_HANDLE hDevice;
+    D3DDDI_SYNCHRONIZATIONOBJECTINFO2 Info;
+    D3DKMT_HANDLE hSyncObject;
+} D3DKMT_CREATESYNCHRONIZATIONOBJECT2;
+
+typedef struct _D3DKMT_OPENSYNCOBJECTFROMNTHANDLE
+{
+    HANDLE hNtHandle;
+    D3DKMT_HANDLE hSyncObject;
+} D3DKMT_OPENSYNCOBJECTFROMNTHANDLE;
+
+typedef struct _D3DKMT_OPENSYNCOBJECTFROMNTHANDLE2
+{
+    HANDLE hNtHandle;
+    D3DKMT_HANDLE hDevice;
+    D3DDDI_SYNCHRONIZATIONOBJECT_FLAGS Flags;
+    D3DKMT_HANDLE hSyncObject;
+    union
+    {
+
+        struct
+        {
+            void *FenceValueCPUVirtualAddress;
+            D3DGPU_VIRTUAL_ADDRESS FenceValueGPUVirtualAddress;
+            UINT EngineAffinity;
+        } MonitoredFence;
+        UINT64 Reserved[8];
+    };
+} D3DKMT_OPENSYNCOBJECTFROMNTHANDLE2;
+
+typedef struct _D3DKMT_OPENSYNCHRONIZATIONOBJECT
+{
+    D3DKMT_HANDLE hSharedHandle;
+    D3DKMT_HANDLE hSyncObject;
+    UINT64 Reserved[8];
+} D3DKMT_OPENSYNCHRONIZATIONOBJECT;
+
+typedef struct _D3DKMT_OPENSYNCOBJECTNTHANDLEFROMNAME
+{
+    DWORD dwDesiredAccess;
+    OBJECT_ATTRIBUTES *pObjAttrib;
+    HANDLE hNtHandle;
+} D3DKMT_OPENSYNCOBJECTNTHANDLEFROMNAME;
+
+typedef struct _D3DKMT_DESTROYSYNCHRONIZATIONOBJECT
+{
+    D3DKMT_HANDLE hSyncObject;
+} D3DKMT_DESTROYSYNCHRONIZATIONOBJECT;
+
+typedef struct _D3DKMT_CREATESTANDARDALLOCATIONFLAGS
+{
+    union
+    {
+        struct
+        {
+            UINT Reserved : 32;
+        };
+        UINT Value;
+    };
+} D3DKMT_CREATESTANDARDALLOCATIONFLAGS;
+
+typedef enum _D3DKMT_STANDARDALLOCATIONTYPE
+{
+    D3DKMT_STANDARDALLOCATIONTYPE_EXISTINGHEAP = 1,
+} D3DKMT_STANDARDALLOCATIONTYPE;
+
+typedef struct _D3DKMT_STANDARDALLOCATION_EXISTINGHEAP
+{
+    SIZE_T Size;
+} D3DKMT_STANDARDALLOCATION_EXISTINGHEAP;
+
+typedef struct _D3DKMT_CREATESTANDARDALLOCATION
+{
+    D3DKMT_STANDARDALLOCATIONTYPE Type;
+    union
+    {
+        D3DKMT_STANDARDALLOCATION_EXISTINGHEAP ExistingHeapData;
+    };
+    D3DKMT_CREATESTANDARDALLOCATIONFLAGS Flags;
+} D3DKMT_CREATESTANDARDALLOCATION;
+
+typedef struct _D3DDDI_ALLOCATIONINFO
+{
+    D3DKMT_HANDLE hAllocation;
+    const void *pSystemMem;
+    void *pPrivateDriverData;
+    UINT PrivateDriverDataSize;
+    D3DDDI_VIDEO_PRESENT_SOURCE_ID VidPnSourceId;
+    union
+    {
+        struct
+        {
+            UINT Primary : 1;
+            UINT Stereo : 1;
+            UINT Reserved : 30;
+        };
+        UINT Value;
+    } Flags;
+} D3DDDI_ALLOCATIONINFO;
+
+typedef struct _D3DDDI_ALLOCATIONINFO2
+{
+    D3DKMT_HANDLE hAllocation;
+    union
+    {
+        HANDLE hSection;
+        const void *pSystemMem;
+    };
+    VOID *pPrivateDriverData;
+    UINT PrivateDriverDataSize;
+    D3DDDI_VIDEO_PRESENT_SOURCE_ID VidPnSourceId;
+    union
+    {
+        struct
+        {
+            UINT Primary : 1;
+            UINT Stereo : 1;
+            UINT OverridePriority : 1;
+            UINT Reserved : 29;
+        };
+        UINT Value;
+    } Flags;
+    D3DGPU_VIRTUAL_ADDRESS GpuVirtualAddress;
+    union
+    {
+        UINT Priority;
+        ULONG_PTR Unused;
+    };
+    ULONG_PTR Reserved[5];
+} D3DDDI_ALLOCATIONINFO2;
+
+typedef struct _D3DKMT_CREATEALLOCATIONFLAGS
+{
+    UINT CreateResource : 1;
+    UINT CreateShared : 1;
+    UINT NonSecure : 1;
+    UINT CreateProtected : 1;
+    UINT RestrictSharedAccess : 1;
+    UINT ExistingSysMem : 1;
+    UINT NtSecuritySharing : 1;
+    UINT ReadOnly : 1;
+    UINT CreateWriteCombined : 1;
+    UINT CreateCached : 1;
+    UINT SwapChainBackBuffer : 1;
+    UINT CrossAdapter : 1;
+    UINT OpenCrossAdapter : 1;
+    UINT PartialSharedCreation : 1;
+    UINT Zeroed : 1;
+    UINT WriteWatch : 1;
+    UINT StandardAllocation : 1;
+    UINT ExistingSection : 1;
+    UINT AllowNotZeroed : 1;
+    UINT PhysicallyContiguous : 1;
+    UINT Reserved : 12;
+} D3DKMT_CREATEALLOCATIONFLAGS;
+
+typedef struct _D3DKMT_CREATEALLOCATION
+{
+    D3DKMT_HANDLE hDevice;
+    D3DKMT_HANDLE hResource;
+    D3DKMT_HANDLE hGlobalShare;
+    const void *pPrivateRuntimeData;
+    UINT PrivateRuntimeDataSize;
+    union
+    {
+        D3DKMT_CREATESTANDARDALLOCATION *pStandardAllocation;
+        const void *pPrivateDriverData;
+    };
+    UINT PrivateDriverDataSize;
+    UINT NumAllocations;
+    union
+    {
+        D3DDDI_ALLOCATIONINFO *pAllocationInfo;
+        D3DDDI_ALLOCATIONINFO2 *pAllocationInfo2;
+    };
+    D3DKMT_CREATEALLOCATIONFLAGS Flags;
+    HANDLE hPrivateRuntimeResourceHandle;
+} D3DKMT_CREATEALLOCATION;
+
+typedef struct _D3DKMT_DESTROYALLOCATION
+{
+    D3DKMT_HANDLE hDevice;
+    D3DKMT_HANDLE hResource;
+    const D3DKMT_HANDLE *phAllocationList;
+    UINT AllocationCount;
+} D3DKMT_DESTROYALLOCATION;
+
+typedef struct _D3DDDICB_DESTROYALLOCATION2FLAGS
+{
+    union
+    {
+        struct
+        {
+            UINT AssumeNotInUse : 1;
+            UINT SynchronousDestroy : 1;
+            UINT Reserved : 29;
+            UINT SystemUseOnly : 1;
+        };
+        UINT Value;
+    };
+} D3DDDICB_DESTROYALLOCATION2FLAGS;
+
+typedef struct _D3DKMT_DESTROYALLOCATION2
+{
+    D3DKMT_HANDLE hDevice;
+    D3DKMT_HANDLE hResource;
+    const D3DKMT_HANDLE *phAllocationList;
+    UINT AllocationCount;
+    D3DDDICB_DESTROYALLOCATION2FLAGS Flags;
+} D3DKMT_DESTROYALLOCATION2;
+
+typedef struct _D3DDDI_OPENALLOCATIONINFO
+{
+    D3DKMT_HANDLE hAllocation;
+    const void *pPrivateDriverData;
+    UINT PrivateDriverDataSize;
+} D3DDDI_OPENALLOCATIONINFO;
+
+typedef struct _D3DDDI_OPENALLOCATIONINFO2
+{
+    D3DKMT_HANDLE hAllocation;
+    const void *pPrivateDriverData;
+    UINT PrivateDriverDataSize;
+    D3DGPU_VIRTUAL_ADDRESS GpuVirtualAddress;
+    ULONG_PTR Reserved[6];
+} D3DDDI_OPENALLOCATIONINFO2;
+
+typedef struct _D3DKMT_OPENRESOURCE
+{
+    D3DKMT_HANDLE hDevice;
+    D3DKMT_HANDLE hGlobalShare;
+    UINT NumAllocations;
+    union
+    {
+        D3DDDI_OPENALLOCATIONINFO *pOpenAllocationInfo;
+        D3DDDI_OPENALLOCATIONINFO2 *pOpenAllocationInfo2;
+    };
+    void *pPrivateRuntimeData;
+    UINT PrivateRuntimeDataSize;
+    void *pResourcePrivateDriverData;
+    UINT ResourcePrivateDriverDataSize;
+    void *pTotalPrivateDriverDataBuffer;
+    UINT TotalPrivateDriverDataBufferSize;
+    D3DKMT_HANDLE hResource;
+} D3DKMT_OPENRESOURCE;
+
+typedef struct _D3DKMT_OPENRESOURCEFROMNTHANDLE
+{
+    D3DKMT_HANDLE hDevice;
+    HANDLE hNtHandle;
+    UINT NumAllocations;
+    D3DDDI_OPENALLOCATIONINFO2 *pOpenAllocationInfo2;
+    UINT PrivateRuntimeDataSize;
+    void *pPrivateRuntimeData;
+    UINT ResourcePrivateDriverDataSize;
+    void *pResourcePrivateDriverData;
+    UINT TotalPrivateDriverDataBufferSize;
+    void *pTotalPrivateDriverDataBuffer;
+    D3DKMT_HANDLE hResource;
+    D3DKMT_HANDLE hKeyedMutex;
+    void *pKeyedMutexPrivateRuntimeData;
+    UINT KeyedMutexPrivateRuntimeDataSize;
+    D3DKMT_HANDLE hSyncObject;
+} D3DKMT_OPENRESOURCEFROMNTHANDLE;
+
+typedef struct _D3DKMT_QUERYRESOURCEINFO
+{
+    D3DKMT_HANDLE hDevice;
+    D3DKMT_HANDLE hGlobalShare;
+    void *pPrivateRuntimeData;
+    UINT PrivateRuntimeDataSize;
+    UINT TotalPrivateDriverDataSize;
+    UINT ResourcePrivateDriverDataSize;
+    UINT NumAllocations;
+} D3DKMT_QUERYRESOURCEINFO;
+
+typedef struct _D3DKMT_QUERYRESOURCEINFOFROMNTHANDLE
+{
+    D3DKMT_HANDLE hDevice;
+    HANDLE hNtHandle;
+    void *pPrivateRuntimeData;
+    UINT PrivateRuntimeDataSize;
+    UINT TotalPrivateDriverDataSize;
+    UINT ResourcePrivateDriverDataSize;
+    UINT NumAllocations;
+} D3DKMT_QUERYRESOURCEINFOFROMNTHANDLE;
+
 #ifdef __cplusplus
 extern "C"
 {
@@ -794,18 +1281,43 @@ extern "C"
 
 NTSTATUS WINAPI D3DKMTCheckVidPnExclusiveOwnership(const D3DKMT_CHECKVIDPNEXCLUSIVEOWNERSHIP *desc);
 NTSTATUS WINAPI D3DKMTCloseAdapter(const D3DKMT_CLOSEADAPTER *desc);
-NTSTATUS WINAPI D3DKMTCreateDevice(D3DKMT_CREATEDEVICE *desc);
+NTSTATUS WINAPI D3DKMTCreateAllocation( D3DKMT_CREATEALLOCATION *params );
+NTSTATUS WINAPI D3DKMTCreateAllocation2( D3DKMT_CREATEALLOCATION *params );
 NTSTATUS WINAPI D3DKMTCreateDCFromMemory(D3DKMT_CREATEDCFROMMEMORY *desc);
+NTSTATUS WINAPI D3DKMTCreateDevice(D3DKMT_CREATEDEVICE *desc);
+NTSTATUS WINAPI D3DKMTCreateKeyedMutex( D3DKMT_CREATEKEYEDMUTEX *params );
+NTSTATUS WINAPI D3DKMTCreateKeyedMutex2( D3DKMT_CREATEKEYEDMUTEX2 *params );
+NTSTATUS WINAPI D3DKMTCreateSynchronizationObject( D3DKMT_CREATESYNCHRONIZATIONOBJECT *params );
+NTSTATUS WINAPI D3DKMTCreateSynchronizationObject2( D3DKMT_CREATESYNCHRONIZATIONOBJECT2 *params );
+NTSTATUS WINAPI D3DKMTDestroyAllocation( const D3DKMT_DESTROYALLOCATION *params );
+NTSTATUS WINAPI D3DKMTDestroyAllocation2( const D3DKMT_DESTROYALLOCATION2 *params );
 NTSTATUS WINAPI D3DKMTDestroyDCFromMemory(const D3DKMT_DESTROYDCFROMMEMORY *desc);
 NTSTATUS WINAPI D3DKMTDestroyDevice(const D3DKMT_DESTROYDEVICE *desc);
+NTSTATUS WINAPI D3DKMTDestroyKeyedMutex( const D3DKMT_DESTROYKEYEDMUTEX *params );
+NTSTATUS WINAPI D3DKMTDestroySynchronizationObject( const D3DKMT_DESTROYSYNCHRONIZATIONOBJECT *params );
+NTSTATUS WINAPI D3DKMTEnumAdapters2(D3DKMT_ENUMADAPTERS2 *desc);
 NTSTATUS WINAPI D3DKMTEscape( const D3DKMT_ESCAPE *desc );
 NTSTATUS WINAPI D3DKMTOpenAdapterFromGdiDisplayName(D3DKMT_OPENADAPTERFROMGDIDISPLAYNAME *desc);
 NTSTATUS WINAPI D3DKMTOpenAdapterFromHdc( D3DKMT_OPENADAPTERFROMHDC *desc );
 NTSTATUS WINAPI D3DKMTOpenAdapterFromLuid( D3DKMT_OPENADAPTERFROMLUID * desc );
+NTSTATUS WINAPI D3DKMTOpenKeyedMutex( D3DKMT_OPENKEYEDMUTEX *params );
+NTSTATUS WINAPI D3DKMTOpenKeyedMutex2( D3DKMT_OPENKEYEDMUTEX2 *params );
+NTSTATUS WINAPI D3DKMTOpenKeyedMutexFromNtHandle( D3DKMT_OPENKEYEDMUTEXFROMNTHANDLE *params );
+NTSTATUS WINAPI D3DKMTOpenResource( D3DKMT_OPENRESOURCE *params );
+NTSTATUS WINAPI D3DKMTOpenResource2( D3DKMT_OPENRESOURCE *params );
+NTSTATUS WINAPI D3DKMTOpenResourceFromNtHandle( D3DKMT_OPENRESOURCEFROMNTHANDLE *params );
+NTSTATUS WINAPI D3DKMTOpenSynchronizationObject( D3DKMT_OPENSYNCHRONIZATIONOBJECT *params );
+NTSTATUS WINAPI D3DKMTOpenSyncObjectFromNtHandle( D3DKMT_OPENSYNCOBJECTFROMNTHANDLE *params );
+NTSTATUS WINAPI D3DKMTOpenSyncObjectFromNtHandle2( D3DKMT_OPENSYNCOBJECTFROMNTHANDLE2 *params );
+NTSTATUS WINAPI D3DKMTOpenSyncObjectNtHandleFromName( D3DKMT_OPENSYNCOBJECTNTHANDLEFROMNAME *params );
+NTSTATUS WINAPI D3DKMTQueryAdapterInfo(D3DKMT_QUERYADAPTERINFO *desc);
+NTSTATUS WINAPI D3DKMTQueryResourceInfo( D3DKMT_QUERYRESOURCEINFO *params );
+NTSTATUS WINAPI D3DKMTQueryResourceInfoFromNtHandle( D3DKMT_QUERYRESOURCEINFOFROMNTHANDLE *params );
 NTSTATUS WINAPI D3DKMTQueryStatistics(D3DKMT_QUERYSTATISTICS *stats);
 NTSTATUS WINAPI D3DKMTQueryVideoMemoryInfo(D3DKMT_QUERYVIDEOMEMORYINFO *desc);
 NTSTATUS WINAPI D3DKMTSetQueuedLimit(D3DKMT_SETQUEUEDLIMIT *desc);
 NTSTATUS WINAPI D3DKMTSetVidPnSourceOwner(const D3DKMT_SETVIDPNSOURCEOWNER *desc);
+NTSTATUS WINAPI D3DKMTShareObjects( UINT count, const D3DKMT_HANDLE *handles, OBJECT_ATTRIBUTES *attr, UINT access, HANDLE *handle );
 
 #ifdef __cplusplus
 }
