@@ -51,7 +51,7 @@ struct irp_call
     struct device_file    *file;          /* file containing this irp */
     struct thread         *thread;        /* thread that queued the irp */
     struct async          *async;         /* pending async op */
-    irp_params_t           params;        /* irp parameters */
+    union irp_params       params;        /* irp parameters */
     struct iosb           *iosb;          /* I/O status block */
     int                    canceled;      /* the call was canceled */
     client_ptr_t           user_ptr;      /* client side pointer */
@@ -362,7 +362,8 @@ static void irp_call_destroy( struct object *obj )
     if (irp->thread) release_object( irp->thread );
 }
 
-static struct irp_call *create_irp( struct device_file *file, const irp_params_t *params, struct async *async )
+static struct irp_call *create_irp( struct device_file *file, const union irp_params *params,
+                                    struct async *async )
 {
     struct irp_call *irp;
 
@@ -440,6 +441,7 @@ static struct object *device_open_file( struct object *obj, unsigned int access,
     struct device *device = (struct device *)obj;
     struct device_file *file;
     struct unicode_str nt_name;
+    WCHAR *fullname;
 
     if (!(file = alloc_object( &device_file_ops ))) return NULL;
 
@@ -450,11 +452,16 @@ static struct object *device_open_file( struct object *obj, unsigned int access,
     list_add_tail( &device->files, &file->entry );
     if (device->unix_path)
     {
-        mode_t mode = 0666;
-        access = file->obj.ops->map_access( &file->obj, access );
-        nt_name.str = device->obj.ops->get_full_name( &device->obj, &nt_name.len );
-        file->fd = open_fd( NULL, device->unix_path, nt_name, O_NONBLOCK, &mode, access, sharing, options );
-        if (file->fd) set_fd_user( file->fd, &device_file_fd_ops, &file->obj );
+        if ((fullname = device->obj.ops->get_full_name( &device->obj, &nt_name.len )))
+        {
+            mode_t mode = 0666;
+            access = file->obj.ops->map_access( &file->obj, access );
+            nt_name.str = fullname;
+            file->fd = open_fd( NULL, device->unix_path, nt_name, O_NONBLOCK, &mode, access, sharing, options );
+            if (file->fd) set_fd_user( file->fd, &device_file_fd_ops, &file->obj );
+            free( fullname );
+        }
+        else file->fd = NULL;
     }
     else file->fd = alloc_pseudo_fd( &device_file_fd_ops, &file->obj, options );
 
@@ -469,7 +476,7 @@ static struct object *device_open_file( struct object *obj, unsigned int access,
     if (device->manager)
     {
         struct irp_call *irp;
-        irp_params_t params;
+        union irp_params params;
 
         memset( &params, 0, sizeof(params) );
         params.create.type    = IRP_CALL_CREATE;
@@ -526,7 +533,7 @@ static int device_file_close_handle( struct object *obj, struct process *process
     if (!file->closed && file->device->manager && obj->handle_count == 1)  /* last handle */
     {
         struct irp_call *irp;
-        irp_params_t params;
+        union irp_params params;
 
         file->closed = 1;
         memset( &params, 0, sizeof(params) );
@@ -556,7 +563,7 @@ static void device_file_destroy( struct object *obj )
     release_object( file->device );
 }
 
-static int fill_irp_params( struct device_manager *manager, struct irp_call *irp, irp_params_t *params )
+static int fill_irp_params( struct device_manager *manager, struct irp_call *irp, union irp_params *params )
 {
     switch (irp->params.type)
     {
@@ -609,7 +616,7 @@ static void free_irp_params( struct irp_call *irp )
 }
 
 /* queue an irp to the device */
-static void queue_irp( struct device_file *file, const irp_params_t *params, struct async *async )
+static void queue_irp( struct device_file *file, const union irp_params *params, struct async *async )
 {
     struct irp_call *irp = create_irp( file, params, async );
     if (!irp) return;
@@ -629,7 +636,7 @@ static enum server_fd_type device_file_get_fd_type( struct fd *fd )
 static void device_file_get_volume_info( struct fd *fd, struct async *async, unsigned int info_class )
 {
     struct device_file *file = get_fd_user( fd );
-    irp_params_t params;
+    union irp_params params;
 
     memset( &params, 0, sizeof(params) );
     params.volume.type = IRP_CALL_VOLUME;
@@ -640,7 +647,7 @@ static void device_file_get_volume_info( struct fd *fd, struct async *async, uns
 static void device_file_read( struct fd *fd, struct async *async, file_pos_t pos )
 {
     struct device_file *file = get_fd_user( fd );
-    irp_params_t params;
+    union irp_params params;
 
     memset( &params, 0, sizeof(params) );
     params.read.type = IRP_CALL_READ;
@@ -652,7 +659,7 @@ static void device_file_read( struct fd *fd, struct async *async, file_pos_t pos
 static void device_file_write( struct fd *fd, struct async *async, file_pos_t pos )
 {
     struct device_file *file = get_fd_user( fd );
-    irp_params_t params;
+    union irp_params params;
 
     memset( &params, 0, sizeof(params) );
     params.write.type = IRP_CALL_WRITE;
@@ -664,7 +671,7 @@ static void device_file_write( struct fd *fd, struct async *async, file_pos_t po
 static void device_file_flush( struct fd *fd, struct async *async )
 {
     struct device_file *file = get_fd_user( fd );
-    irp_params_t params;
+    union irp_params params;
 
     memset( &params, 0, sizeof(params) );
     params.flush.type = IRP_CALL_FLUSH;
@@ -674,7 +681,7 @@ static void device_file_flush( struct fd *fd, struct async *async )
 static void device_file_ioctl( struct fd *fd, ioctl_code_t code, struct async *async )
 {
     struct device_file *file = get_fd_user( fd );
-    irp_params_t params;
+    union irp_params params;
 
     memset( &params, 0, sizeof(params) );
     params.ioctl.type = IRP_CALL_IOCTL;
@@ -685,7 +692,7 @@ static void device_file_ioctl( struct fd *fd, ioctl_code_t code, struct async *a
 static void cancel_irp_call( struct irp_call *irp )
 {
     struct irp_call *cancel_irp;
-    irp_params_t params;
+    union irp_params params;
 
     irp->canceled = 1;
     if (!irp->user_ptr || !irp->file || !irp->file->device->manager) return;
@@ -887,7 +894,7 @@ void free_kernel_objects( struct object *obj )
     {
         struct kernel_object *kernel_object = LIST_ENTRY( ptr, struct kernel_object, list_entry );
         struct irp_call *irp;
-        irp_params_t params;
+        union irp_params params;
 
         assert( !kernel_object->owned );
 
