@@ -136,27 +136,6 @@ static const struct object_ops desktop_ops =
     desktop_destroy               /* destroy */
 };
 
-#if defined(__i386__) || defined(__x86_64__)
-#define __SHARED_INCREMENT_SEQ( x ) ++(x)
-#else
-#define __SHARED_INCREMENT_SEQ( x ) __atomic_add_fetch( &(x), 1, __ATOMIC_RELEASE )
-#endif
-
-#define SHARED_WRITE_BEGIN( object, type )                           \
-    do {                                                             \
-        const type *__shared = (object)->shared;                     \
-        type *shared = (type *)__shared;                             \
-        unsigned int __seq = __SHARED_INCREMENT_SEQ( shared->seq );  \
-        assert( (__seq & 1) != 0 );                                  \
-        do
-
-#define SHARED_WRITE_END                                             \
-        while(0);                                                    \
-        __seq = __SHARED_INCREMENT_SEQ( shared->seq ) - __seq;       \
-        assert( __seq == 1 );                                        \
-    } while(0);
-
-
 /* create a winstation object */
 static struct winstation *create_winstation( struct object *root, const struct unicode_str *name,
                                              unsigned int attr, unsigned int flags )
@@ -169,8 +148,12 @@ static struct winstation *create_winstation( struct object *root, const struct u
         {
             /* initialize it if it didn't already exist */
             winstation->flags = flags;
+            winstation->input_desktop = NULL;
             winstation->clipboard = NULL;
             winstation->atom_table = NULL;
+            winstation->monitors = NULL;
+            winstation->monitor_count = 0;
+            winstation->monitor_serial = 1;
             list_add_tail( &winstation_list, &winstation->entry );
             list_init( &winstation->desktops );
             if (!(winstation->desktop_names = create_namespace( 7 )))
@@ -227,6 +210,7 @@ static void winstation_destroy( struct object *obj )
     if (winstation->clipboard) release_object( winstation->clipboard );
     if (winstation->atom_table) release_object( winstation->atom_table );
     free( winstation->desktop_names );
+    free( winstation->monitors );
 }
 
 /* retrieve the process window station, checking the handle access rights */
@@ -236,73 +220,129 @@ struct winstation *get_process_winstation( struct process *process, unsigned int
                                                 access, &winstation_ops );
 }
 
+/* retrieve the visible winstation */
+struct winstation *get_visible_winstation(void)
+{
+    struct winstation *winstation;
+    LIST_FOR_EACH_ENTRY( winstation, &winstation_list, struct winstation, entry )
+        if (winstation->flags & WSF_VISIBLE) return winstation;
+    return NULL;
+}
+
+/* retrieve the winstation current input desktop */
+struct desktop *get_input_desktop( struct winstation *winstation )
+{
+    struct desktop *desktop;
+    if (!(desktop = winstation->input_desktop)) return NULL;
+    return (struct desktop *)grab_object( desktop );
+}
+
+/* changes the winstation current input desktop and update its input time */
+int set_input_desktop( struct winstation *winstation, struct desktop *new_desktop )
+{
+    struct desktop *old_desktop = winstation->input_desktop;
+    struct thread *thread;
+
+    if (!(winstation->flags & WSF_VISIBLE)) return 0;
+    if (new_desktop) new_desktop->input_time = current_time;
+    if (old_desktop == new_desktop) return 1;
+
+    if (old_desktop)
+    {
+        /* disconnect every process of the old input desktop from rawinput */
+        LIST_FOR_EACH_ENTRY( thread, &old_desktop->threads, struct thread, desktop_entry )
+            set_rawinput_process( thread->process, 0 );
+    }
+
+    if ((winstation->input_desktop = new_desktop))
+    {
+        /* connect every process of the new input desktop to rawinput */
+        LIST_FOR_EACH_ENTRY( thread, &new_desktop->threads, struct thread, desktop_entry )
+            set_rawinput_process( thread->process, 1 );
+    }
+
+    return 1;
+}
+
 /* retrieve a pointer to a desktop object */
 struct desktop *get_desktop_obj( struct process *process, obj_handle_t handle, unsigned int access )
 {
     return (struct desktop *)get_handle_obj( process, handle, access, &desktop_ops );
 }
 
-static int init_desktop_mapping( struct desktop *desktop, const struct unicode_str *name )
-{
-    struct object *dir;
-
-    desktop->shared = NULL;
-    desktop->shared_mapping = NULL;
-
-    if (!(dir = create_desktop_map_directory( desktop->winstation ))) return 0;
-    if ((desktop->shared_mapping = create_shared_mapping( dir, name, sizeof(struct desktop_shared_memory),
-                                                          0, NULL, (void **)&desktop->shared )))
-    {
-        memset( (void *)desktop->shared, 0, sizeof(*desktop->shared) );
-        ((desktop_shm_t *)desktop->shared)->update_serial = 1;
-    }
-    release_object( dir );
-
-    return !!desktop->shared;
-}
-
 /* create a desktop object */
 static struct desktop *create_desktop( const struct unicode_str *name, unsigned int attr,
                                        unsigned int flags, struct winstation *winstation )
 {
-    struct desktop *desktop;
+    struct desktop *desktop, *current_desktop;
 
     if ((desktop = create_named_object( &winstation->obj, &desktop_ops, name, attr, NULL )))
     {
         if (get_error() != STATUS_OBJECT_NAME_EXISTS)
         {
             /* initialize it if it didn't already exist */
-            desktop->flags = flags;
+
+            /* inherit DF_WINE_*_DESKTOP flags if none of them are specified */
+            if (!(flags & (DF_WINE_ROOT_DESKTOP | DF_WINE_VIRTUAL_DESKTOP))
+                && (current_desktop = get_thread_desktop( current, 0 )))
+            {
+                flags |= current_desktop->shared->flags & (DF_WINE_VIRTUAL_DESKTOP | DF_WINE_ROOT_DESKTOP);
+                release_object( current_desktop );
+            }
+
             desktop->winstation = (struct winstation *)grab_object( winstation );
             desktop->top_window = NULL;
             desktop->msg_window = NULL;
+            desktop->shell_window = NULL;
+            desktop->shell_listview = NULL;
+            desktop->progman_window = NULL;
+            desktop->taskman_window = NULL;
             desktop->global_hooks = NULL;
             desktop->close_timeout = NULL;
             desktop->close_timeout_val = 0;
             desktop->foreground_input = NULL;
             desktop->users = 0;
+            list_init( &desktop->threads );
+            desktop->clip_flags = 0;
             desktop->cursor_win = 0;
-            desktop->cursor_clip_flags = 0;
-            desktop->last_press_alt = 0;
+            desktop->alt_pressed = 0;
+            memset( &desktop->key_repeat, 0, sizeof(desktop->key_repeat) );
             list_add_tail( &winstation->desktops, &desktop->entry );
             list_init( &desktop->hotkeys );
-            list_init( &desktop->touches );
-            if (!init_desktop_mapping( desktop, name ))
+            list_init( &desktop->pointers );
+
+            if (!(desktop->shared = alloc_shared_object()))
             {
                 release_object( desktop );
                 return NULL;
             }
+
+            SHARED_WRITE_BEGIN( desktop->shared, desktop_shm_t )
+            {
+                shared->flags = flags;
+                shared->cursor.x = 0;
+                shared->cursor.y = 0;
+                shared->cursor.last_change = 0;
+                shared->cursor.clip.left = 0;
+                shared->cursor.clip.top = 0;
+                shared->cursor.clip.right = 0;
+                shared->cursor.clip.bottom = 0;
+                memset( (void *)shared->keystate, 0, sizeof(shared->keystate) );
+                shared->keystate_serial = 1;
+                shared->monitor_serial = winstation->monitor_serial;
+            }
+            SHARED_WRITE_END;
         }
         else
         {
-            desktop->flags |= (flags & DF_WINE_CREATE_DESKTOP);
+            SHARED_WRITE_BEGIN( desktop->shared, desktop_shm_t )
+            {
+                shared->flags |= flags & (DF_WINE_VIRTUAL_DESKTOP | DF_WINE_ROOT_DESKTOP);
+            }
+            SHARED_WRITE_END;
+
             clear_error();
         }
-        SHARED_WRITE_BEGIN( desktop, desktop_shm_t )
-        {
-            shared->flags = desktop->flags;
-        }
-        SHARED_WRITE_END
     }
     return desktop;
 }
@@ -312,7 +352,7 @@ static void desktop_dump( struct object *obj, int verbose )
     struct desktop *desktop = (struct desktop *)obj;
 
     fprintf( stderr, "Desktop flags=%x winstation=%p top_win=%p hooks=%p\n",
-             desktop->flags, desktop->winstation, desktop->top_window, desktop->global_hooks );
+             desktop->shared->flags, desktop->winstation, desktop->top_window, desktop->global_hooks );
 }
 
 static int desktop_link_name( struct object *obj, struct object_name *name, struct object *parent )
@@ -347,17 +387,27 @@ static int desktop_close_handle( struct object *obj, struct process *process, ob
 static void desktop_destroy( struct object *obj )
 {
     struct desktop *desktop = (struct desktop *)obj;
+    struct winstation *winstation = desktop->winstation;
+
+    list_remove( &desktop->entry );
+
+    if (desktop == winstation->input_desktop)
+    {
+        struct desktop *other, *found = NULL;
+        LIST_FOR_EACH_ENTRY(other, &winstation->desktops, struct desktop, entry)
+            if (!found || other->input_time > found->input_time) found = other;
+        set_input_desktop( winstation, found );
+    }
 
     free_hotkeys( desktop, 0 );
-    free_touches( desktop, 0 );
+    free_pointers( desktop );
     if (desktop->top_window) free_window_handle( desktop->top_window );
     if (desktop->msg_window) free_window_handle( desktop->msg_window );
     if (desktop->global_hooks) release_object( desktop->global_hooks );
     if (desktop->close_timeout) remove_timeout_user( desktop->close_timeout );
-    list_remove( &desktop->entry );
-    if (desktop->shared_mapping) release_object( desktop->shared_mapping );
-    desktop->shared_mapping = NULL;
+    if (desktop->key_repeat.timeout) remove_timeout_user( desktop->key_repeat.timeout );
     release_object( desktop->winstation );
+    if (desktop->shared) free_shared_object( desktop->shared );
 }
 
 /* retrieve the thread desktop, checking the handle access rights */
@@ -372,25 +422,34 @@ static void close_desktop_timeout( void *private )
 
     desktop->close_timeout = NULL;
     unlink_named_object( &desktop->obj );  /* make sure no other process can open it */
-    unlink_named_object( desktop->shared_mapping );
     post_desktop_message( desktop, WM_CLOSE, 0, 0 );  /* and signal the owner to quit */
 }
 
 /* add a user of the desktop and cancel the close timeout */
-static void add_desktop_user( struct desktop *desktop )
+static void add_desktop_thread( struct desktop *desktop, struct thread *thread )
 {
-    desktop->users++;
-    if (desktop->close_timeout)
+    list_add_tail( &desktop->threads, &thread->desktop_entry );
+    add_desktop_hook_count( desktop, thread, 1 );
+
+    if (!thread->process->is_system)
     {
-        remove_timeout_user( desktop->close_timeout );
-        desktop->close_timeout = NULL;
+        desktop->users++;
+        if (desktop->close_timeout)
+        {
+            remove_timeout_user( desktop->close_timeout );
+            desktop->close_timeout = NULL;
+        }
     }
+
+    /* if thread process is now connected to the input desktop, let it receive rawinput */
+    if (desktop == desktop->winstation->input_desktop) set_rawinput_process( thread->process, 1 );
 }
 
 /* remove a user of the desktop and start the close timeout if necessary */
-static void remove_desktop_user( struct desktop *desktop )
+static void remove_desktop_user( struct desktop *desktop, struct thread *thread )
 {
     struct process *process;
+
     assert( desktop->users > 0 );
     desktop->users--;
 
@@ -399,13 +458,29 @@ static void remove_desktop_user( struct desktop *desktop )
         desktop->close_timeout = add_timeout_user( desktop->close_timeout_val, close_desktop_timeout, desktop );
 }
 
+/* remove a thread from the list of threads attached to a desktop */
+static void remove_desktop_thread( struct desktop *desktop, struct thread *thread )
+{
+    add_desktop_hook_count( desktop, thread, -1 );
+    list_remove( &thread->desktop_entry );
+
+    if (!thread->process->is_system) remove_desktop_user( desktop, thread );
+
+    if (desktop == desktop->winstation->input_desktop)
+    {
+        /* thread process might still be connected the input desktop through another thread, update the full list */
+        LIST_FOR_EACH_ENTRY( thread, &desktop->threads, struct thread, desktop_entry )
+            set_rawinput_process( thread->process, 1 );
+    }
+}
+
 /* set the thread default desktop handle */
 void set_thread_default_desktop( struct thread *thread, struct desktop *desktop, obj_handle_t handle )
 {
     if (thread->desktop) return;  /* nothing to do */
 
     thread->desktop = handle;
-    if (!thread->process->is_system) add_desktop_user( desktop );
+    add_desktop_thread( desktop, thread );
 }
 
 /* set the process default desktop handle */
@@ -424,18 +499,37 @@ void set_process_default_desktop( struct process *process, struct desktop *deskt
 }
 
 /* connect a process to its window station */
-void connect_process_winstation( struct process *process, struct thread *parent_thread,
-                                 struct process *parent_process )
+void connect_process_winstation( struct process *process, struct unicode_str *desktop_path,
+                                 struct thread *parent_thread, struct process *parent_process )
 {
+    struct unicode_str desktop_name = *desktop_path, winstation_name = {0};
+    const int attributes = OBJ_CASE_INSENSITIVE | OBJ_OPENIF;
     struct winstation *winstation = NULL;
     struct desktop *desktop = NULL;
+    const WCHAR *wch, *end;
     obj_handle_t handle;
 
+    for (wch = desktop_name.str, end = wch + desktop_name.len / sizeof(WCHAR); wch != end; wch++)
+    {
+        if (*wch == '\\')
+        {
+            winstation_name.str = desktop_name.str;
+            winstation_name.len = (wch - winstation_name.str) * sizeof(WCHAR);
+            desktop_name.str = wch + 1;
+            desktop_name.len = (end - desktop_name.str) * sizeof(WCHAR);
+            break;
+        }
+    }
+
     /* check for an inherited winstation handle (don't ask...) */
     if ((handle = find_inherited_handle( process, &winstation_ops )))
     {
         winstation = (struct winstation *)get_handle_obj( process, handle, 0, &winstation_ops );
     }
+    else if (winstation_name.len && (winstation = open_named_object( NULL, &winstation_ops, &winstation_name, attributes )))
+    {
+        handle = alloc_handle( process, winstation, STANDARD_RIGHTS_REQUIRED | WINSTA_ALL_ACCESS, 0 );
+    }
     else if (parent_process->winstation)
     {
         handle = duplicate_handle( parent_process, parent_process->winstation,
@@ -450,6 +544,10 @@ void connect_process_winstation( struct process *process, struct thread *parent_
         desktop = get_desktop_obj( process, handle, 0 );
         if (!desktop || desktop->winstation != winstation) goto done;
     }
+    else if (desktop_name.len && (desktop = open_named_object( &winstation->obj, &desktop_ops, &desktop_name, attributes )))
+    {
+        handle = alloc_handle( process, desktop, STANDARD_RIGHTS_REQUIRED | DESKTOP_ALL_ACCESS, 0 );
+    }
     else
     {
         if (parent_thread && parent_thread->desktop)
@@ -492,14 +590,12 @@ void release_thread_desktop( struct thread *thread, int close )
 
     if (!(handle = thread->desktop)) return;
 
-    if (!thread->process->is_system)
+    if (!(desktop = get_desktop_obj( thread->process, handle, 0 ))) clear_error();  /* ignore errors */
+    else
     {
-        if (!(desktop = get_desktop_obj( thread->process, handle, 0 ))) clear_error();  /* ignore errors */
-        else
-        {
-            remove_desktop_user( desktop );
-            release_object( desktop );
-        }
+        if (close) remove_desktop_thread( desktop, thread );
+        else remove_desktop_user( desktop, thread );
+        release_object( desktop );
     }
 
     if (close)
@@ -551,6 +647,42 @@ DECL_HANDLER(close_winstation)
 }
 
 
+/* set the process current window station monitors */
+DECL_HANDLER(set_winstation_monitors)
+{
+    struct winstation *winstation;
+    struct desktop *desktop;
+    unsigned int size = get_req_data_size();
+
+    if (!(winstation = (struct winstation *)get_handle_obj( current->process, current->process->winstation,
+                                                           0, &winstation_ops )))
+        return;
+
+    if (req->increment || winstation->monitor_count != size / sizeof(*winstation->monitors) ||
+        !winstation->monitors || memcmp( winstation->monitors, get_req_data(), size ))
+        winstation->monitor_serial++;
+
+    free( winstation->monitors );
+    winstation->monitors = NULL;
+    winstation->monitor_count = 0;
+
+    LIST_FOR_EACH_ENTRY(desktop, &winstation->desktops, struct desktop, entry)
+    {
+        SHARED_WRITE_BEGIN( desktop->shared, desktop_shm_t )
+        {
+            shared->monitor_serial = winstation->monitor_serial;
+        }
+        SHARED_WRITE_END;
+    }
+
+    if (size && (winstation->monitors = memdup( get_req_data(), size )))
+        winstation->monitor_count = size / sizeof(*winstation->monitors);
+
+    reply->serial = winstation->monitor_serial;
+    release_object( winstation );
+}
+
+
 /* get the process current window station */
 DECL_HANDLER(get_process_winstation)
 {
@@ -589,6 +721,7 @@ DECL_HANDLER(create_desktop)
     {
         if ((desktop = create_desktop( &name, req->attributes, req->flags, winstation )))
         {
+            if (!winstation->input_desktop) set_input_desktop( winstation, desktop );
             reply->handle = alloc_handle( current->process, desktop, req->access, req->attributes );
             release_object( desktop );
         }
@@ -636,7 +769,7 @@ DECL_HANDLER(open_input_desktop)
         return;
     }
 
-    if ((desktop = get_desktop_obj( current->process, current->process->desktop, 0 )))
+    if ((desktop = get_input_desktop( winstation )))
     {
         reply->handle = alloc_handle( current->process, desktop, req->access, req->attributes );
         release_object( desktop );
@@ -644,6 +777,24 @@ DECL_HANDLER(open_input_desktop)
     release_object( winstation );
 }
 
+/* changes the current input desktop */
+DECL_HANDLER(set_input_desktop)
+{
+    /* FIXME: check access rights */
+    struct winstation *winstation;
+    struct desktop *desktop;
+
+    if (!(winstation = get_process_winstation( current->process, 0 ))) return;
+
+    if ((desktop = (struct desktop *)get_handle_obj( current->process, req->handle, 0, &desktop_ops )))
+    {
+        if (!set_input_desktop( winstation, desktop )) set_error( STATUS_ILLEGAL_FUNCTION );
+        release_object( desktop );
+    }
+
+    release_object( winstation );
+}
+
 /* close a desktop */
 DECL_HANDLER(close_desktop)
 {
@@ -662,10 +813,19 @@ DECL_HANDLER(close_desktop)
 /* get the thread current desktop */
 DECL_HANDLER(get_thread_desktop)
 {
+    struct desktop *desktop;
     struct thread *thread;
 
     if (!(thread = get_thread_from_id( req->tid ))) return;
     reply->handle = thread->desktop;
+
+    if (!(desktop = get_thread_desktop( thread, 0 ))) clear_error();
+    else
+    {
+        if (desktop->shared) reply->locator = get_shared_object_locator( desktop->shared );
+        release_object( desktop );
+    }
+
     release_object( thread );
 }
 
@@ -703,11 +863,12 @@ DECL_HANDLER(set_thread_desktop)
     else
     {
         current->desktop = req->handle;  /* FIXME: should we close the old one? */
-        if (!current->process->is_system && old_desktop != new_desktop)
+        if (old_desktop != new_desktop)
         {
-            add_desktop_user( new_desktop );
-            if (old_desktop) remove_desktop_user( old_desktop );
+            if (old_desktop) remove_desktop_thread( old_desktop, current );
+            add_desktop_thread( new_desktop, current );
         }
+        reply->locator = get_shared_object_locator( new_desktop->shared );
     }
 
     if (!current->process->desktop)
@@ -733,15 +894,14 @@ DECL_HANDLER(set_user_object_info)
     {
         struct desktop *desktop = (struct desktop *)obj;
         reply->is_desktop = 1;
-        reply->old_obj_flags = desktop->flags;
+        reply->old_obj_flags = desktop->shared->flags;
         if (req->flags & SET_USER_OBJECT_SET_FLAGS)
         {
-            desktop->flags = req->obj_flags;
-            SHARED_WRITE_BEGIN( desktop, desktop_shm_t )
+            SHARED_WRITE_BEGIN( desktop->shared, desktop_shm_t )
             {
-                shared->flags = desktop->flags;
+                shared->flags = req->obj_flags;
             }
-            SHARED_WRITE_END
+            SHARED_WRITE_END;
         }
         if (req->flags & SET_USER_OBJECT_SET_CLOSE_TIMEOUT) desktop->close_timeout_val = req->close_timeout;
     }
@@ -771,9 +931,9 @@ DECL_HANDLER(set_user_object_info)
         len = winstation_len + desktop_len + sizeof(WCHAR);
         if ((full_name = mem_alloc( len )))
         {
-            memcpy( full_name, winstation_name, winstation_len );
-            full_name[winstation_len / sizeof(WCHAR)] = '\\';
-            memcpy( full_name + winstation_len / sizeof(WCHAR) + 1, desktop_name, desktop_len );
+            WCHAR *ptr = mem_append( full_name, winstation_name, winstation_len );
+            *ptr++ = '\\';
+            mem_append( ptr, desktop_name, desktop_len );
             set_reply_data_ptr( full_name, min( len, get_reply_max_size() ));
         }
     }
@@ -786,56 +946,68 @@ DECL_HANDLER(set_user_object_info)
 }
 
 
-/* enumerate window stations */
+/* enumerate window stations and desktops */
 DECL_HANDLER(enum_winstation)
 {
-    unsigned int index = 0;
-    struct winstation *winsta;
-    const WCHAR *name;
-    data_size_t len;
-
-    LIST_FOR_EACH_ENTRY( winsta, &winstation_list, struct winstation, entry )
-    {
-        unsigned int access = WINSTA_ENUMERATE;
-        if (req->index > index++) continue;
-        if (!check_object_access( NULL, &winsta->obj, &access )) continue;
-        clear_error();
-        reply->next = index;
-        if ((name = get_object_name( &winsta->obj, &len )))
-            set_reply_data( name, min( len, get_reply_max_size() ));
-        return;
-    }
-    set_error( STATUS_NO_MORE_ENTRIES );
-}
-
-
-/* enumerate desktops */
-DECL_HANDLER(enum_desktop)
-{
+    data_size_t size = get_reply_max_size() / sizeof(WCHAR) * sizeof(WCHAR);
+    struct object_name *name;
     struct winstation *winstation;
-    struct desktop *desktop;
-    unsigned int index = 0;
-    const WCHAR *name;
-    data_size_t len;
+    WCHAR *data = NULL, *ptr;
 
-    if (!(winstation = (struct winstation *)get_handle_obj( current->process, req->winstation,
-                                                            WINSTA_ENUMDESKTOPS, &winstation_ops )))
-        return;
+    if (size && !(data = mem_alloc( size ))) return;
+    ptr = data;
 
-    LIST_FOR_EACH_ENTRY( desktop, &winstation->desktops, struct desktop, entry )
+    if (req->handle)
     {
-        unsigned int access = DESKTOP_ENUMERATE;
-        if (req->index > index++) continue;
-        if (!desktop->obj.name) continue;
-        if (!check_object_access( NULL, &desktop->obj, &access )) continue;
-        if ((name = get_object_name( &desktop->obj, &len )))
-            set_reply_data( name, min( len, get_reply_max_size() ));
+        struct desktop *desktop;
+
+        if (!(winstation = (struct winstation *)get_handle_obj( current->process, req->handle,
+                                                                WINSTA_ENUMDESKTOPS, &winstation_ops )))
+        {
+            free( data );
+            return;
+        }
+
+        LIST_FOR_EACH_ENTRY( desktop, &winstation->desktops, struct desktop, entry )
+        {
+            unsigned int access = DESKTOP_ENUMERATE;
+            if (!(name = desktop->obj.name)) continue;
+            if (!check_object_access( NULL, &desktop->obj, &access )) continue;
+            reply->count++;
+            reply->total += name->len + sizeof(WCHAR);
+            if (reply->total <= size)
+            {
+                ptr = mem_append( ptr, name->name, name->len );
+                *ptr++ = 0;
+            }
+        }
         release_object( winstation );
-        clear_error();
-        reply->next = index;
-        return;
+    }
+    else
+    {
+        LIST_FOR_EACH_ENTRY( winstation, &winstation_list, struct winstation, entry )
+        {
+            unsigned int access = WINSTA_ENUMERATE;
+            if (!(name = winstation->obj.name)) continue;
+            if (!check_object_access( NULL, &winstation->obj, &access )) continue;
+            reply->count++;
+            reply->total += name->len + sizeof(WCHAR);
+            if (reply->total <= size)
+            {
+                ptr = mem_append( ptr, name->name, name->len );
+                *ptr++ = 0;
+            }
+        }
     }
 
-    release_object( winstation );
-    set_error( STATUS_NO_MORE_ENTRIES );
+    if (reply->total <= size)
+    {
+        set_reply_data_ptr( data, reply->total );
+        clear_error();
+    }
+    else
+    {
+        set_reply_data_ptr( data, (ptr - data) * sizeof(WCHAR) );
+        set_error( STATUS_BUFFER_TOO_SMALL );
+    }
 }
